using System;
using JPGL.Common;
using JPGL.Common.Primitives;
using JPGL.Math;

namespace DrivingGame
{
	/// <summary>
	/// Summary description for Car.
	/// </summary>
	public class Car : SceneNode
	{
		//-----------------------------------// All wheel scenenodes must have 1 child!
		protected SceneNode m_Body;
		protected SceneNode m_WheelBackLeft;
		protected SceneNode m_WheelBackRight;
		protected SceneNode m_WheelFrontLeft;
		protected SceneNode m_WheelFrontRight;
		//-----------------------------------//

		//-----------------------------------//
		protected Ode.Body          m_CarBody;
		protected Ode.Body[]        m_WheelBody;
		protected Ode.Hinge2Joint[] m_WheelJoints;
		//-----------------------------------//

		//-----------------------------------//
		protected Ode.World m_OdeWorld;
		protected Ode.Space m_OdeSpace;
		//-----------------------------------//

		//-----------------------------------//
		protected Vector3 m_Offset;          // Positioning offset
		protected Vector3 m_BodyPos;         // Initial Body Position        
		protected Vector3 m_BodySize;        // Initial Body Size
		protected float   m_WheelRadius;     // Initial Wheel Radius
		protected float   m_COGYOffset;      // Center Of Gravity offset in the y
		//-----------------------------------//

		//-----------------------------------//
		protected Texture m_CarTexture;
		//-----------------------------------//

		//-----------------------------------//
		protected float m_Density;
		protected float m_EnginePower;
		protected float m_SteeringPower;
		protected float m_TopSpeed;
		//-----------------------------------//

		//-----------------------------------//
		protected static float DIFFERENTIAL = 0.2f;
		protected static float SPEEDDIFFFACTOR = 10;
		protected static float FORCEDIFFFACTOR = 1.5f;
		//-----------------------------------//

		#region Public Accessor Properties
		//-----------------------------------//
		public Vector3 Offset
		{
			get { return m_Offset;  }
			set { m_Offset = value; }
		}
		//-----------------------------------//
		public float EnginePower
		{
			get { return m_EnginePower; }
		}
		//-----------------------------------//
		public float SteeringPower
		{
			get { return m_SteeringPower; }
		}
		//-----------------------------------//
		public float TopSpeed
		{
			get { return m_TopSpeed; }
		}
		//-----------------------------------//
		public SceneNode Body
		{
			get { return m_Body; }
		}
		//-----------------------------------//
		public SceneNode WheelBackLeft
		{
			get { return m_WheelBackLeft; }
		}
		//-----------------------------------//
		public SceneNode WheelBackRight
		{
			get { return m_WheelBackRight; }
		}
		//-----------------------------------//
		public SceneNode WheelFrontLeft
		{
			get { return m_WheelFrontLeft; }
		}
		//-----------------------------------//
		public SceneNode WheelFrontRight
		{
			get { return m_WheelFrontRight; }
		}
		//-----------------------------------//
		//public Ode.Body CarBody
		public Ode.Body CarBody
		{
			get { return m_CarBody; }
		}
		//-----------------------------------//
		public Ode.Body[] WheelBody
		{
			get { return m_WheelBody; }
		}
		//-----------------------------------//
		public Ode.Hinge2Joint[] WheelJoints
		{
			get { return m_WheelJoints; }
		}
		//-----------------------------------//
		public float CurrentSpeed
		{
			get { return System.Math.Abs(m_WheelJoints[3].Angle2Rate*m_WheelRadius); }
		}
		//-----------------------------------//
		public Vector3 BodySize
		{
			get { return m_BodySize; }
		}
		//-----------------------------------//
		#endregion


		//-------------------------------------------------------------------------//
		public Car(Ode.World w, Ode.Space s) : this(w, s, new Vector3())
		{

		}

		//-------------------------------------------------------------------------//
		public Car(Ode.World w, Ode.Space s, Vector3 offset) : base(null, new Transform())
		{
			m_OdeWorld = w;
			m_OdeSpace = s;

			m_Density       = 10;
			m_EnginePower   = 50;
			m_SteeringPower = 100;
			m_TopSpeed      = 15; // as defined, a little over 140MPH

			m_COGYOffset = 0;
			m_Offset = offset;
		}

		//-------------------------------------------------------------------------//
		public void Destroy()
		{
			if (m_OdeSpace != null)
			{
				for (int i=0; i<4; i++) 
				{
					m_OdeSpace.RemoveGeom(m_WheelBody[i].Geoms.get_Geom(0));
				}
				m_OdeSpace.RemoveGeom(m_CarBody.Geoms.get_Geom(0));
			}

			if (m_CarTexture != null)
				JPGL.OpenGL.GraphicsSystem.UnloadTexture(m_CarTexture);
		}

		//-------------------------------------------------------------------------//
		// Flip the car in the y
		public virtual void Flip()
		{
			m_Offset = T.T;
			Reset();
		}

		//-------------------------------------------------------------------------//
		public virtual void Reset()
		{
			for (int i=0; i<4; i++) 
			{
				m_OdeSpace.RemoveGeom(m_WheelBody[i].Geoms.get_Geom(0));
			}
			m_OdeSpace.RemoveGeom(m_CarBody.Geoms.get_Geom(0));

			BuildCarDynamics(m_BodyPos, m_BodySize, m_WheelRadius);
		}

		//-------------------------------------------------------------------------//
		public virtual void UpdateForces(float timestep, int steer, int speed, bool handbrake)
		{
			#region Power-Differential Drive System
			float frdiff=1, fldiff=1, brdiff=1, bldiff=1;

			// Turning left, reduce power to the left wheels,
			// increase power to the right.
			if (steer < -0.1f)
			{
				fldiff -= DIFFERENTIAL;
				bldiff -= DIFFERENTIAL;
				frdiff += DIFFERENTIAL;
				brdiff += DIFFERENTIAL;
			}
			// Turning right, reduce power to the right wheels,
			// increase power to the left.
			else if (steer > 0.1f)
			{
				frdiff -= DIFFERENTIAL;
				brdiff -= DIFFERENTIAL;
				fldiff += DIFFERENTIAL;
				bldiff += DIFFERENTIAL;
			}

			// Front Left
			m_WheelJoints[0].SetParam(Ode.JointParam.Vel2, 
				(SPEEDDIFFFACTOR*fldiff)*speed*m_TopSpeed);
			m_WheelJoints[0].SetParam(Ode.JointParam.FMax2,
				(FORCEDIFFFACTOR*fldiff)*m_EnginePower);
			// Front Right
			m_WheelJoints[1].SetParam(Ode.JointParam.Vel2, 
				(SPEEDDIFFFACTOR*frdiff)*speed*m_TopSpeed);
			m_WheelJoints[1].SetParam(Ode.JointParam.FMax2, 
				(FORCEDIFFFACTOR*frdiff)*m_EnginePower);
			// Back Left
			m_WheelJoints[2].SetParam(Ode.JointParam.Vel2, 
				(handbrake ? 0 : 1.0f)*(SPEEDDIFFFACTOR*bldiff)*speed*m_TopSpeed);
			m_WheelJoints[2].SetParam(Ode.JointParam.FMax2, 
				(handbrake ? 500f : 1.0f)*(FORCEDIFFFACTOR*bldiff)*m_EnginePower);
			// Back Right
			m_WheelJoints[3].SetParam(Ode.JointParam.Vel2, 
				(handbrake ? 0 : 1.0f)*(SPEEDDIFFFACTOR*brdiff)*speed*m_TopSpeed);
			m_WheelJoints[3].SetParam(Ode.JointParam.FMax2, 
				(handbrake ? 500f : 1.0f)*(FORCEDIFFFACTOR*brdiff)*m_EnginePower);
			#endregion

			#region Front Wheel Steering
			for (int i=0; i<2; i++)
			{
				// Not actively steering and moving, then move the wheels
				// back to no rotation.
				if (steer == 0 && 
					(m_WheelJoints[i].Angle2Rate > 0.00001f ||
					 m_WheelJoints[i].Angle2Rate < -0.00001f))
				{
					m_WheelJoints[i].SetParam(Ode.JointParam.Vel, -m_WheelJoints[i].Angle1);
					m_WheelJoints[i].SetParam(Ode.JointParam.FMax, 200);
				}
				else
				{
					m_WheelJoints[i].SetParam(Ode.JointParam.Vel, steer);
					m_WheelJoints[i].SetParam(Ode.JointParam.FMax, m_SteeringPower);
				}
				m_WheelJoints[i].SetParam(Ode.JointParam.LoStop, -0.5f);
				m_WheelJoints[i].SetParam(Ode.JointParam.HiStop, 0.5f);
				m_WheelJoints[i].SetParam(Ode.JointParam.FudgeFactor, 0.001f);
			}
			#endregion
		}

		//-------------------------------------------------------------------------//
		public virtual void Animate(float timestep)
		{
			m_Body.T.RQ.W = m_CarBody.Quaternion.W;
			m_Body.T.RQ.X = m_CarBody.Quaternion.X;
			m_Body.T.RQ.Y = m_CarBody.Quaternion.Y;
			m_Body.T.RQ.Z = m_CarBody.Quaternion.Z;

			Quaternion rot = m_Body.T.RQ.Inverse;

			((SceneNode)m_WheelFrontLeft.Children[0]).T.RQ = rot * new Quaternion(
				m_WheelBody[0].Quaternion.W, m_WheelBody[0].Quaternion.X,
				m_WheelBody[0].Quaternion.Y, m_WheelBody[0].Quaternion.Z);

			((SceneNode)m_WheelFrontRight.Children[0]).T.RQ = 
				((SceneNode)m_WheelFrontLeft.Children[0]).T.RQ;
//			((SceneNode)m_WheelFrontRight.Children[0]).T.RQ = rot * new Quaternion(
//				m_WheelBody[1].Quaternion.W, m_WheelBody[1].Quaternion.X,
//				m_WheelBody[1].Quaternion.Y, m_WheelBody[1].Quaternion.Z);

			((SceneNode)m_WheelBackLeft.Children[0]).T.RQ = rot * new Quaternion(
				m_WheelBody[2].Quaternion.W, m_WheelBody[2].Quaternion.X,
				m_WheelBody[2].Quaternion.Y, m_WheelBody[2].Quaternion.Z);

			((SceneNode)m_WheelBackRight.Children[0]).T.RQ =
				((SceneNode)m_WheelBackLeft.Children[0]).T.RQ;
//			((SceneNode)m_WheelBackRight.Children[0]).T.RQ = rot * new Quaternion(
//				m_WheelBody[3].Quaternion.W, m_WheelBody[3].Quaternion.X,
//				m_WheelBody[3].Quaternion.Y, m_WheelBody[3].Quaternion.Z);

			T.T.X = m_CarBody.Position.X - m_BodyPos.X;
			T.T.Y = m_CarBody.Position.Y - m_BodyPos.Y;
			T.T.Z = m_CarBody.Position.Z - m_BodyPos.Z;
		}

		//-------------------------------------------------------------------------//
		// Given Geometry IDs for parts, build model structure
		protected void BuildCarModel(int body, int wbr, int wbl, int wfl, int wfr)
		{
			m_Body            = new SceneNode("Body", new LoadedPrimitive(body), new Transform());
			m_WheelBackRight  = new SceneNode("Wheel BR Adjust", null, new Transform());
			m_WheelBackRight.AddChild(new SceneNode("Wheel BR", new LoadedPrimitive(wbr), new Transform()));
			m_WheelBackLeft   = new SceneNode("Wheel BL Adjust", null, new Transform());
			m_WheelBackLeft.AddChild(new SceneNode("Wheel BL", new LoadedPrimitive(wbl), new Transform()));
			m_WheelFrontLeft  = new SceneNode("Wheel FL Adjust", null, new Transform());
			m_WheelFrontLeft.AddChild(new SceneNode("Wheel FL", new LoadedPrimitive(wfl), new Transform()));
			m_WheelFrontRight = new SceneNode("Wheel FR Adjust", null, new Transform());
			m_WheelFrontRight.AddChild(new SceneNode("Wheel FR", new LoadedPrimitive(wfr), new Transform()));
		}

		//-------------------------------------------------------------------------//
		protected void SetCarTexture(string filename)
		{
			m_CarTexture = JPGL.OpenGL.GraphicsSystem.LoadTexture(filename);
			m_Body.P.Texture = m_CarTexture;
			((SceneNode)m_WheelBackLeft.Children[0]).P.Texture = m_CarTexture;
			((SceneNode)m_WheelBackRight.Children[0]).P.Texture = m_CarTexture;
			((SceneNode)m_WheelFrontLeft.Children[0]).P.Texture = m_CarTexture;
			((SceneNode)m_WheelFrontRight.Children[0]).P.Texture = m_CarTexture;
		}

		//-------------------------------------------------------------------------//
		protected void BuildSNHeirarchy()
		{
			this.AddChild(m_Body);
			Body.AddChild(m_WheelBackLeft);
			Body.AddChild(m_WheelBackRight);
			Body.AddChild(m_WheelFrontLeft);
			Body.AddChild(m_WheelFrontRight);
		}

		//-------------------------------------------------------------------------//
		protected void BuildCarDynamics(Vector3 bpos, Vector3 bsize, float r)
		{
			#region Setup Body
			m_Body.T.P         = bpos;
			m_CarBody          = new Ode.Body(m_OdeWorld);
			m_CarBody.Position = new Ode.Vector3(bpos.X+m_Offset.X, bpos.Y+m_Offset.Y, bpos.Z+m_Offset.Z);
			m_CarBody.Geoms.Add(new Ode.BoxGeom(m_OdeSpace, bsize.X, bsize.Y, bsize.Z));

//			int vcount = m_Body.P.GetGeometry().VertexCount / 3;
//			Ode.Vector3[] vectors = new Ode.Vector3[vcount];
//			short[] indices = new short[vcount];
//			for (int i=0; i<vcount; i++)
//			{
//				indices[i] = (short)i;
//				vectors[i] = new Ode.Vector3(
//					m_Body.P.GetGeometry().Vertices[(i*3)+0],
//					m_Body.P.GetGeometry().Vertices[(i*3)+1],
//					m_Body.P.GetGeometry().Vertices[(i*3)+2]);
//			}
//
//			Ode.TriMeshData tridata = new Ode.TriMeshData();
//			tridata.Build(vectors, indices);
//
//			Ode.TriMeshGeom triGeom = new Ode.TriMeshGeom(m_OdeSpace, tridata);
//			triGeom.Position = new Ode.Vector3(bpos.X, bpos.Y, bpos.Z);
//			m_CarBody.Geoms.Add(triGeom);

			Ode.Mass m = new Ode.Mass();			
			m.SetBox(m_Density, bsize.X, bsize.Y, bsize.Z);
			m.Translate(0, m_COGYOffset, 0);
			m_CarBody.SetMass(m);
			#endregion
			
			m_WheelBody = new Ode.Body[4];
			Ode.Geom wGeom;

			// Rotation for cylinder geoms
			Quaternion q = new Quaternion(new Vector3(0, 1, 0), 90);
			Ode.Quaternion qode = new Ode.Quaternion();
			qode.X = q.X; qode.Y = q.Y; qode.Z = q.Z; qode.W = q.W;

			#region Setup Front Left Wheel
			m_WheelBody[0] = new Ode.Body(m_OdeWorld);
			m_WheelBody[0].Position = new Ode.Vector3(
				((SceneNode)m_WheelFrontLeft.Children[0]).T.T.X+m_Offset.X, 
				((SceneNode)m_WheelFrontLeft.Children[0]).T.T.Y+m_Offset.Y, 
				((SceneNode)m_WheelFrontLeft.Children[0]).T.T.Z+m_Offset.Z);
			wGeom = new Ode.GeomTransform(m_OdeSpace);//new Ode.SphereGeom(m_OdeSpace, r);
			((Ode.GeomTransform)wGeom).WrappedGeom = new Ode.CappedCylinderGeom(r, 0.25f);
			((Ode.GeomTransform)wGeom).WrappedGeom.Quaternion = qode;
			((Ode.GeomTransform)wGeom).InformationMode = 1;
			m_WheelBody[0].Geoms.Add(wGeom);
			#endregion

			#region Setup Front Right Wheel
			m_WheelBody[1] = new Ode.Body(m_OdeWorld);
			m_WheelBody[1].Position = new Ode.Vector3(
				((SceneNode)m_WheelFrontRight.Children[0]).T.T.X+m_Offset.X, 
				((SceneNode)m_WheelFrontRight.Children[0]).T.T.Y+m_Offset.Y, 
				((SceneNode)m_WheelFrontRight.Children[0]).T.T.Z+m_Offset.Z);
			wGeom = new Ode.GeomTransform(m_OdeSpace);//new Ode.SphereGeom(m_OdeSpace, r);
			((Ode.GeomTransform)wGeom).WrappedGeom = new Ode.CappedCylinderGeom(r, 0.25f);
			((Ode.GeomTransform)wGeom).WrappedGeom.Quaternion = qode;
			((Ode.GeomTransform)wGeom).InformationMode = 1;
			m_WheelBody[1].Geoms.Add(wGeom);
			#endregion

			#region Setup Back Left Wheel
			m_WheelBody[2] = new Ode.Body(m_OdeWorld);
			m_WheelBody[2].Position = new Ode.Vector3(
				((SceneNode)m_WheelBackLeft.Children[0]).T.T.X+m_Offset.X, 
				((SceneNode)m_WheelBackLeft.Children[0]).T.T.Y+m_Offset.Y, 
				((SceneNode)m_WheelBackLeft.Children[0]).T.T.Z+m_Offset.Z);
			wGeom = new Ode.GeomTransform(m_OdeSpace);//new Ode.SphereGeom(m_OdeSpace, r);
			((Ode.GeomTransform)wGeom).WrappedGeom = new Ode.CappedCylinderGeom(r, 0.25f);
			((Ode.GeomTransform)wGeom).WrappedGeom.Quaternion = qode;
			((Ode.GeomTransform)wGeom).InformationMode = 1;
			m_WheelBody[2].Geoms.Add(wGeom);
			#endregion

			#region Setup Back Right Wheel
			m_WheelBody[3] = new Ode.Body(m_OdeWorld);
			m_WheelBody[3].Position = new Ode.Vector3(
				((SceneNode)m_WheelBackRight.Children[0]).T.T.X+m_Offset.X, 
				((SceneNode)m_WheelBackRight.Children[0]).T.T.Y+m_Offset.Y, 
				((SceneNode)m_WheelBackRight.Children[0]).T.T.Z+m_Offset.Z);
			wGeom = new Ode.GeomTransform(m_OdeSpace);//new Ode.SphereGeom(m_OdeSpace, r);
			((Ode.GeomTransform)wGeom).WrappedGeom = new Ode.CappedCylinderGeom(r, 0.25f);
			((Ode.GeomTransform)wGeom).WrappedGeom.Quaternion = qode;
			((Ode.GeomTransform)wGeom).InformationMode = 1;
			m_WheelBody[3].Geoms.Add(wGeom);
			
			#endregion

			#region Setup Wheel Joints and Masses
			m_WheelJoints = new Ode.Hinge2Joint[4];
			for (int i=0; i<4; i++) 
			{
				// Joint
				m_WheelJoints[i] = new Ode.Hinge2Joint(m_OdeWorld);
				m_WheelJoints[i].Attach(m_CarBody, m_WheelBody[i]);
				Ode.Vector3 wheelpos = m_WheelBody[i].Position;
				m_WheelJoints[i].Anchor = wheelpos;
				m_WheelJoints[i].Axis1 = new Ode.Vector3(0,1,0);
				m_WheelJoints[i].Axis2 = new Ode.Vector3(-1,0,0);

				// Suspension
				m_WheelJoints[i].SetParam(Ode.JointParam.SuspensionERP, 0.8f);
				m_WheelJoints[i].SetParam(Ode.JointParam.SuspensionCFM, 0.001f);

				// Rear Wheels dont turn
				if (i >= 2)
				{
					m_WheelJoints[i].SetParam(Ode.JointParam.LoStop, 0);
					m_WheelJoints[i].SetParam(Ode.JointParam.HiStop, 0);
				}
			}
			#endregion

		}
	}
}
