using System;
using System.Collections;

using JPGL.Common;
using JPGL.Common.Colliders;
using JPGL.Common.Lighting;
using JPGL.Common.Primitives;
using JPGL.Common.SceneNodes;
using JPGL.Math;


namespace DrivingGame
{
	/// <summary>
	/// Summary description for WorldModel.
	/// </summary>
	public class WorldModel
	{
		//-----------------------------------//
		private SceneNode m_SceneRoot;       // Root SceneNode
		private Skybox    m_Skybox;
		//-----------------------------------//

		//-----------------------------------//
		private Random m_RandGen;
		private bool   m_SimRunning;
		//-----------------------------------//

		//-----------------------------------// Lights
		private Light m_Headlight1;          // spot light
		private Light m_Headlight2;          // spot light
		private Light m_Light3;              // Directional light
		private Light m_Light4;              // Point light
		//-----------------------------------//

		#region Public Property Accessors
		//-----------------------------------//
		public bool SimRunning
		{
			get { return m_SimRunning;  }
			set { m_SimRunning = value; }
		}
		//-----------------------------------//
		public SceneNode Scene
		{
			get { return m_SceneRoot; }
		}
		//-----------------------------------//
		public Light Light1
		{
			get { return m_Headlight1;  }
			set { m_Headlight1 = value; }
		}
		//-----------------------------------//
		public Light Light2
		{
			get { return m_Headlight2;  }
			set { m_Headlight2 = value; }
		}
		//-----------------------------------//
		public Light Light3
		{
			get { return m_Light3;  }
			set { m_Light3 = value; }
		}
		//-----------------------------------//
		public Light Light4
		{
			get { return m_Light4;  }
			set { m_Light4 = value; }
		}
		//-----------------------------------//
		public Skybox Skybox
		{
			get { return m_Skybox;  }
			set { m_Skybox = value; }
		}
		//-----------------------------------//
		#endregion

		//-----------------------------------//
		private Ode.World          m_OdeWorld; // The world our bodies are in
		private Ode.Space          m_OdeSpace; // The space our Geoms collide in
		private Ode.CollideHandler m_OdeCollideHandler;
		private Ode.JointGroup     m_OdeContactGroup;
		//-----------------------------------//

		//-----------------------------------//
		private Car   m_Car1;
		private Car   m_Car2;
		private int   m_Car1Steer = 0;
		private int   m_Car1Speed = 0;
		private bool  m_Car1HandbrakeOn = false;
		private bool  m_Car2Chasing = false;
		public enum CarTypes
		{
			F1,HUMVEE,STOCK,JEEP,DELOREAN,SMALLCAR,H2
		}
		//-----------------------------------//

		//-----------------------------------//
		private ArrayList m_Boxes;
		private ArrayList m_Fence;
		private Windmill  m_Windmill;
		//-----------------------------------//

		#region Driving Control Public Accessor Properties
		//-----------------------------------//
		public Car Car1
		{
			get { return m_Car1;  }
			set { m_Car1 = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public Car Car2
		{
			get { return m_Car2;  }
			set { m_Car2 = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public enum Steering { LEFT=-1, NONE=0, RIGHT=1 }
		//-----------------------------------//
		public Steering CarSteering
		{
			set { m_Car1Steer = (int)value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public enum Speed { REVERSE=-1, NONE=0, FORWARD=1 }
		//-----------------------------------//
		public Speed CarSpeed
		{
			set { m_Car1Speed = (int)value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public bool Handbrake
		{
			set { m_Car1HandbrakeOn = value; }
		}
		//-----------------------------------//	
		#endregion

		//-----------------------------------//
		private ParticleSystem m_PartSys;
		public bool            m_PartSysOn = false;
		//-----------------------------------//

		//-----------------------------------//
		public static string EchoString = null;
		//-----------------------------------//


		//-------------------------------------------------------------------------//
		public WorldModel()
		{
			m_RandGen = new Random();

			// Scene Root node
			m_SceneRoot = new SceneNode("Scene Root", null, new Transform());

			#region ODE Initialization
			m_OdeWorld = new Ode.World();
			m_OdeWorld.SetGravity(0, -9.8f, 0);
			m_OdeWorld.CFM = 0.0001f;

			m_OdeSpace = new Ode.HashSpace();

			// Walls
			Ode.PlaneGeom p = new Ode.PlaneGeom(m_OdeSpace, 1, 0, 0, -75);
			p = new Ode.PlaneGeom(m_OdeSpace,-1, 0, 0, -75);
			p = new Ode.PlaneGeom(m_OdeSpace, 0, 0, 1, -75);
			p = new Ode.PlaneGeom(m_OdeSpace, 0, 0,-1, -75);

			// Collision Callback and Contact Group
			m_OdeCollideHandler = new Ode.CollideHandler(this.CollideCallback);
			m_OdeContactGroup   = new Ode.JointGroup();
			#endregion
		
			#region Light Setup
			m_Headlight1 = new SpotLight();
			m_Headlight1.DiffuseColor = System.Drawing.Color.LightYellow;
			m_Headlight2 = new SpotLight();
			m_Headlight2.DiffuseColor = System.Drawing.Color.LightYellow;

			m_Light3 = new DirectionalLight();
			m_Light3.Position = new Vector3(-5, 5, 0);

			m_Light4 = new PointLight();
			m_Light4.Position = new Vector3(10, 10, -10);
			#endregion

			#region Skybox, Ground Plane, and Scenery Setup
			m_Skybox = new Skybox();

			Ground g = new Ground(m_OdeSpace);
			m_SceneRoot.AddChild(g);

			House h = new House(m_OdeSpace);
			m_SceneRoot.AddChild(h);

			m_Windmill = new Windmill(m_OdeSpace);
			m_SceneRoot.AddChild(m_Windmill);

			// Boxes
			m_Boxes = new ArrayList(8);
			for (int i=0; i<=3; i++)
			{
				Box b = new Box(i+41.5f, 0.5f, 22, m_OdeWorld, m_OdeSpace);
				m_SceneRoot.AddChild(b);
				m_Boxes.Add(b);
				b = new Box(i+41.5f, 1.5f, 22, m_OdeWorld, m_OdeSpace);
				m_SceneRoot.AddChild(b);
				m_Boxes.Add(b);
			}

			// Fences
			m_Fence = new ArrayList();
			for (int i=0; i<8; i++)
			{
				if (i == 3 || i == 4) continue;
				Fence f = new Fence(-2+2*i, 0, -26, 0, m_OdeWorld, m_OdeSpace);
				m_SceneRoot.AddChild(f);
				m_Fence.Add(f);
			}
			for (int i=0; i<8; i++)
			{
				Fence f = new Fence(-2+2*i, 0, -42, 0, m_OdeWorld, m_OdeSpace);
				m_SceneRoot.AddChild(f);
				m_Fence.Add(f);
			}
			for (int i=0; i<8; i++)
			{
				Fence f = new Fence(-3, 0, -27-i*2, 90, m_OdeWorld, m_OdeSpace);
				m_SceneRoot.AddChild(f);
				m_Fence.Add(f);
			}
			for (int i=0; i<8; i++)
			{
				Fence f = new Fence(13, 0, -27-i*2, 90, m_OdeWorld, m_OdeSpace);
				m_SceneRoot.AddChild(f);
				m_Fence.Add(f);
			}
			
			// Low Ramps
			Vector3 xaxis = new Vector3(1, 0, 0);
			Ramp ramp = new Ramp(0, 0, 8, 4, 6, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, 10));
			m_SceneRoot.AddChild(ramp);
			ramp = new Ramp(4, 0, 8, 4, 6, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, 10));
			m_SceneRoot.AddChild(ramp);
			ramp = new Ramp(0, 0, -6, 4, 6, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, -10));
			m_SceneRoot.AddChild(ramp);
			ramp = new Ramp(4, 0, -6, 4, 6, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, -10));
			m_SceneRoot.AddChild(ramp);

			// High Ramps
			ramp = new Ramp(0, 0, 6, 4, 6, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, 30));
			m_SceneRoot.AddChild(ramp);
			ramp = new Ramp(0, 0, -4, 4, 6, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, -30));
			m_SceneRoot.AddChild(ramp);

			// Long Ramps
			float yoff = 10.0f * (float)System.Math.Sin(System.Math.PI/18.0);
			ramp = new Ramp(-40, yoff-0.5f, 20, 4, 20, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, 10));
			m_SceneRoot.AddChild(ramp);

			yoff = 10.0f * (float)System.Math.Sin(System.Math.PI/18.0);
			ramp = new Ramp(40, yoff-0.5f, -10, 4, 20, m_OdeSpace);
			ramp.SetRotation(new Quaternion(xaxis, -10));
			m_SceneRoot.AddChild(ramp);

			// Twisty ramp
			Vector3 zaxis = new Vector3(0, 0, 1);
			ramp = new Ramp(
				10, 2.0f*(float)System.Math.Sin(System.Math.PI/18.0)-0.5f, 40, 
				4, 4, m_OdeSpace);
			ramp.SetRotation(new Quaternion(zaxis, 10));
			m_SceneRoot.AddChild(ramp);

			for (int i=30; i<=180; i+=20)
			{
				double thetalast = ((double)(i-20) * System.Math.PI / 180.0);
				double thetarad  = ((double)i * System.Math.PI / 180.0);
				ramp = new Ramp(
					ramp.T.T.X+2.0f*(float)System.Math.Cos(thetalast)
					+2.0f*(float)System.Math.Cos(thetarad), 
					ramp.T.T.Y+2.0f*(float)System.Math.Sin(thetalast)
					+2.0f*(float)System.Math.Sin(thetarad),
					ramp.T.T.Z, 
					4, 4, m_OdeSpace);

				ramp.SetRotation(new Quaternion(zaxis, i));
				m_SceneRoot.AddChild(ramp);	

			}
			#endregion

			#region Initialize Cars
			m_Car2 = new Car_Police(m_OdeWorld, m_OdeSpace, new Vector3(-70, 0, 0));
			m_SceneRoot.AddChild(m_Car2);

			SetCar(CarTypes.STOCK);
			#endregion

			m_PartSys = new ParticleSystem(100);
			ParticleSystem.Texture = JPGL.OpenGL.GraphicsSystem.LoadTexture("particle.bmp");


			m_SimRunning = true;
		}

		//-------------------------------------------------------------------------//
		public void SetCar(CarTypes c)
		{
			Vector3 startpos = new Vector3(-15, 0, 20);

			if (m_Car1 != null)
			{
				startpos = m_Car1.T.T;

				m_SceneRoot.RemoveChild(m_Car1);
				m_Car1.Destroy();
			}

			if      (c == CarTypes.F1)
				m_Car1 = new Car_F1(m_OdeWorld, m_OdeSpace, startpos);
			else if (c == CarTypes.HUMVEE)
				m_Car1 = new Car_Humvee(m_OdeWorld, m_OdeSpace, startpos);
			else if (c == CarTypes.JEEP)
				m_Car1 = new Car_Small4by4(m_OdeWorld, m_OdeSpace, startpos);
			else if (c == CarTypes.STOCK)
				m_Car1 = new Car_Stock(m_OdeWorld, m_OdeSpace, startpos);
			else if (c == CarTypes.DELOREAN)
				m_Car1 = new Car_Delorean(m_OdeWorld, m_OdeSpace, startpos);
			else if (c == CarTypes.SMALLCAR)
				m_Car1 = new Car_Small(m_OdeWorld, m_OdeSpace, startpos);
			else if (c == CarTypes.H2)
				m_Car1 = new Car_H2(m_OdeWorld, m_OdeSpace, startpos);

			m_SceneRoot.AddChild(m_Car1);
		}

		//-------------------------------------------------------------------------//
		public void DrawWorld(PrimitiveRenderer r)
		{
			// Lighting
			m_Headlight1.Display(r);
			m_Headlight2.Display(r);
			m_Light3.Display(r);
			m_Light4.Display(r);

			// Draw the world
			m_SceneRoot.DrawScene(r);

			// Render the particle system
			if (m_PartSysOn)
				r.RenderParticleSystem(m_PartSys);
		}

		//-------------------------------------------------------------------------//
		public void ResetWorld()
		{
			// Boxes
			for (int i=0; i<=3; i++)
			{
				((Box)m_Boxes[i]).Reset(i+41.5f, 0.5f, 22);
				((Box)m_Boxes[i+4]).Reset(i+41.5f, 1.5f, 22);
			}

			foreach (Box b in m_Boxes) 
				b.Animate(0);

			// Fences
			for (int i=0; i<8; i++)
			{
				if (i == 3 || i == 4) continue;
				if (i > 4)
					((Fence)m_Fence[i-2]).Reset(-2+2*i, 0, -26, 0);
				else
					((Fence)m_Fence[i]).Reset(-2+2*i, 0, -26, 0);
			}
			for (int i=0; i<8; i++)
				((Fence)m_Fence[i+6]).Reset(-2+2*i, 0, -42, 0);
			for (int i=0; i<8; i++)
				((Fence)m_Fence[i+14]).Reset(-3, 0, -27-i*2, 90);
			for (int i=0; i<8; i++)
				((Fence)m_Fence[i+22]).Reset(13, 0, -27-i*2, 90);

			foreach (Fence f in m_Fence)
				f.Animate(0);

			// Reset Cars
			m_SceneRoot.RemoveChild(m_Car2);
			m_Car2.Destroy();
			m_Car2 = new Car_Police(m_OdeWorld, m_OdeSpace, new Vector3(-70, 0, 0));
			m_SceneRoot.AddChild(m_Car2);
			m_Car2.Animate(0);

			m_SceneRoot.RemoveChild(m_Car1);
			m_Car1.Destroy();
			m_Car1 = new Car_Stock(m_OdeWorld, m_OdeSpace, new Vector3(-15, 0, 20));
			m_SceneRoot.AddChild(m_Car1);
			m_Car1.Animate(0);

			m_Car2Chasing = false;

			m_PartSysOn = false;
		}

		//-------------------------------------------------------------------------//
		public void UpdateWorld(float timestep, JPGL.Common.Viewports.Viewport v)
		{
			// Update Dynamics Simulation
			if (m_SimRunning)
			{
				float t = timestep;
				if (t < 0.017f)
				{
					UpdateWorldForces(t);
					m_OdeContactGroup.Empty();
					m_OdeSpace.Collide(m_OdeCollideHandler);
					m_OdeWorld.Step(t);	
				}
				else 
				{
					while (t > 0.017f)
					{
						UpdateWorldForces(0.017f);
						m_OdeContactGroup.Empty();
						m_OdeSpace.Collide(m_OdeCollideHandler);
						m_OdeWorld.Step(0.017f);

						t -= 0.017f;
					}
					// Disregard whatever is left in t
				}				
			}
			
			// Animations
			m_Car1.Animate(timestep);
			m_Car2.Animate(timestep);

			foreach (Box b in m_Boxes) 
				b.Animate(timestep);

			foreach (Fence f in m_Fence)
				f.Animate(timestep);

			m_Windmill.Animate(timestep);

			if (m_PartSysOn)
			{
				m_PartSys.UpdateSystem(timestep);
				m_PartSys.OriginX = m_Car1.T.T.X + m_Car1.Body.T.P.X;
				m_PartSys.OriginY = m_Car1.T.T.Y + m_Car1.Body.T.P.Y;
				m_PartSys.OriginZ = m_Car1.T.T.Z + m_Car1.Body.T.P.Z;
			}

			#region Update Car Headlights
			Vector3 pos; v.ObjectToWorld(m_Car1.Body, m_Car1.Body.T.P + new Vector3(0.5f, 0, 1), out pos);
			Vector3 pos2; v.ObjectToWorld(m_Car1.Body, m_Car1.Body.T.P + new Vector3(-0.5f, 0, 1), out pos2);

			// Get look at pos
			Vector3 look; v.ObjectToWorld(m_Car1.Body, m_Car1.Body.T.P + new Vector3(0.5f, 0, 2), out look);
			Vector3 look2; v.ObjectToWorld(m_Car1.Body, m_Car1.Body.T.P + new Vector3(-0.5f, 0, 2), out look2);

			m_Headlight1.Position  = pos;
			m_Headlight1.Direction = (look - pos).Normalize();

			m_Headlight2.Position  = pos;
			m_Headlight2.Direction = (look2 - pos2).Normalize();
			#endregion
		}

		//-------------------------------------------------------------------------//
		private void UpdateWorldForces(float timestep)
		{
			#region Make Car2 chase Car1
			if (!m_Car2Chasing && m_Car1.CurrentSpeed > 20.0f &&
				(m_Car1.T.T - m_Car2.T.T).Length() < 30.0f)
			{
				m_Car2Chasing = true;
				EchoString = "Here comes the law!";
			}
			if (m_Car2.CarBody.Rotation.m22 < -0.0001f)
			{
				// Reset car 2 if flipped
				m_Car2.Flip();
			}
			else if (m_Car2Chasing && m_RandGen.Next(10) > 8)
			{
				// 10% chance of car2 getting new force
				Vector3 heading = new Vector3(
					m_Car2.CarBody.Rotation.m31, 
					m_Car2.CarBody.Rotation.m32, 
					m_Car2.CarBody.Rotation.m33);

				Vector3 tocar1 = (m_Car1.T.T - m_Car2.T.T).Normalize();

				double costh = heading * tocar1;	
				double theta = System.Math.Acos(costh);
				Vector3 cross = heading.CrossProduct(tocar1);

				int steerdir = 0;
				if (cross.Y > 0.0001f)
					steerdir = -1;
				else if (cross.Y < -0.0001f)
					steerdir = 1;

				int drivedir = 1;
				// Greater than 90 degrees, hit reverse
				if (theta > System.Math.PI/2.0)
				{
					drivedir = -1;
					steerdir = -steerdir;
				}
				
				m_Car2.UpdateForces(timestep, steerdir, drivedir*(int)m_Car2.TopSpeed, false);
			}
			#endregion

			// Drive car 1
			m_Car1.UpdateForces(timestep, m_Car1Steer, m_Car1Speed, m_Car1HandbrakeOn);
		}

		//-------------------------------------------------------------------------//
		protected virtual void CollideCallback(Ode.Geom o1, Ode.Geom o2)
		{
			// don't collide things that aren't dynamic
			if (o1.Body == null && o2.Body == null)
				return;
			// or things that are connected
			else if(o1.Body != null && o2.Body != null && o1.Body.IsConnectedTo(o2.Body))
				return;

			Ode.ContactGeom[] cgeoms = o1.Collide(o2, 3);
			if (cgeoms.Length > 0)
			{
				Ode.Contact[] contacts = Ode.Contact.FromContactGeomArray(cgeoms);
            
				for (int i = 0; i < contacts.Length; i++)
				{
					contacts[i].Surface.mode   = Ode.SurfaceMode.Bounce; 
					contacts[i].Surface.mu     = 500;
					contacts[i].Surface.bounce = 0.3f;
					contacts[i].Surface.bounce_vel = 0.2f;

					Ode.ContactJoint cj = new Ode.ContactJoint(m_OdeWorld, m_OdeContactGroup, contacts[i]);
					cj.Attach(contacts[i].Geom.Geom1.Body, contacts[i].Geom.Geom2.Body);
				}
			}
		}
	}
}
