using System;
using JPGL.Common.Primitives;
using JPGL.Math;


namespace JPGL.Common
{
	#region Fun with .Net Attributes
	/// <summary>
	/// A Brief Tangent:
	/// .Net has some very interesting features - such as user
	/// defined custom attributes. Here's a sample...
	/// 
	/// - This allows us to specify metadeta regarding
	///   anything from a class to an assembly to a 
	///   field, which can then be accessed programatically.
	///   
	/// - The Angle class itself has a compile time attribute
	///   that will throw a compiler error if the Angle 
	///   attribute is applied to anything but a Field or
	///   Paramater.
	///   
	/// - The Angle class constructor forces us to give the
	///   attribute a positional property - that is, when
	///   applying the attribute, we must give it the 
	///   paramater the constructor wants. Alternatively, 
	///   we could make the constructor take no paramaters
	///   and apply the attribute as [Angle (Units = Angle.UNITS.DEGREES)]
	///   to set the named "Units" attribute.
	/// </summary>
	[System.AttributeUsage (AttributeTargets.Field | AttributeTargets.Parameter)]
	public class Angle : System.Attribute
	{
		// Private fields
		private UNITS m_Units;
		private static float fPI = (float)System.Math.PI;

		// Different types of units
		public enum UNITS { DEGREES, RADIANS };
		
		#region Public Accessor Properties
		public UNITS Units
		{
			get { return m_Units; }
		}
		#endregion

		public Angle (UNITS units)
		{
			m_Units = units;
		}
	}
	#endregion

	/// <summary>
	/// Camera stores variables related to a camera (position, look at, up), and
	/// allows for movement. An event is provided so that objects may register to
	/// be notified when the camera moves.
	/// </summary>
	public class Camera
	{
		//-----------------------------------// 
		[Angle (Angle.UNITS.DEGREES)]
		private float   m_FOV;
		//-----------------------------------// 

		private static int CAMERAMESH = Geometry.LoadOBJ("camera.obj");

		//-----------------------------------// Event handler handles.
		public event CameraHandler CameraUpdate;
		//-----------------------------------//

		//-----------------------------------// Camera Variables
		private Vector3 m_EyePosition;
		private Vector3 m_LookAtPosition;
		private Vector3 m_EyeUpDirection;
		private bool    m_ShowCamera;
		private float   m_Near;
		private float   m_Far;
		//-----------------------------------//

		#region Camera Public Properties
		//-----------------------------------//
		public Vector3 EyePosition
		{
			get { return m_EyePosition; }
			set
			{
				m_EyePosition = value;
				this.OnCameraUpdate(new CameraEventArgs());
			}
		}
		//-----------------------------------//
		public Vector3 LookAtPosition
		{
			get { return m_LookAtPosition; }
			set
			{
				m_LookAtPosition = value;
				this.OnCameraUpdate(new CameraEventArgs());
			}
		}
		//-----------------------------------//
		public Vector3 EyeUpDirection
		{
			get { return m_EyeUpDirection; }
			set
			{
				m_EyeUpDirection = value;
				this.OnCameraUpdate(new CameraEventArgs());
			}
		}
		//-----------------------------------//
		public float FOV
		{
			get { return m_FOV; }
			set
			{
				m_FOV = value;
				this.OnCameraUpdate(new CameraEventArgs());
			}
		}
		//-----------------------------------//
		public bool ShowCamera
		{
			get { return m_ShowCamera;  }
			set { m_ShowCamera = value; }
		}
		//-----------------------------------//
		public float Far
		{
			get { return m_Far;  }
			set { m_Far = value; }
		}
		//-----------------------------------//
		public float Near
		{
			get { return m_Near;  }
			set { m_Near = value; }
		}
		//-----------------------------------//
		#endregion


		//-------------------------------------------------------------------------//
		public Camera()
		{
			m_EyePosition    = new Vector3(0, 10, 18);
			m_LookAtPosition = new Vector3(0,  0,  0);
			m_EyeUpDirection = new Vector3(0,  1,  0);
			m_FOV            = 45;
			m_ShowCamera     = true;
			m_Near           = 1.0f;
			m_Far            = 100.0f;
		}

		//-------------------------------------------------------------------------//
		public void SetCamera(Vector3 eye, Vector3 look, Vector3 up, float fov)
		{
			m_EyePosition    = eye;
			m_LookAtPosition = look;
			m_EyeUpDirection = up;
			m_FOV            = fov;

			this.OnCameraUpdate(new CameraEventArgs());
		}

		//-------------------------------------------------------------------------//
		// 3rd Person Rotation:  p = m_LookAtPosition
		// 1st Person Rotation:  p = m_EyePosition
		// Rotate Up/Down:    axis = <1, 0, 0>
		// Rotate Left/Right: axis = <0, 1, 0>
		public void RotateAroundPoint(Vector3 p, float angle, Vector3 axis)
		{
			Quaternion qRot, qPos, qNewPos;

			// Create the rotation quaternion based on the axis we are rotating on.
			qRot = new Quaternion(axis, angle);

			// Create the pos quaternion.  This will be the direction of the center point and position.
			qPos = new Quaternion(0, m_EyePosition.X - p.X, m_EyePosition.Y - p.Y, m_EyePosition.Z - p.Z);

			// Create the resulting quaternion by multiplying the rotation quat by the pos quat
			// then multiplying that by the conjugate of the rotation quat.
			qNewPos = ((qRot * qPos) * qRot.Conjugate);

			// Update the position information by adding the position to the resulting quaternion.
			m_EyePosition.X = m_LookAtPosition.X + qNewPos.X;
			m_EyePosition.Y = m_LookAtPosition.Y + qNewPos.Y;
			m_EyePosition.Z = m_LookAtPosition.Z + qNewPos.Z;

			this.OnCameraUpdate(new CameraEventArgs());
		}

		//-------------------------------------------------------------------------//
		// Standard eye position based zooming
		public void Zoom(float delta)
		{
			Vector3 dir = (m_LookAtPosition - m_EyePosition).Normalize();
			m_EyePosition += dir * delta;

			this.OnCameraUpdate(new CameraEventArgs());
		}

		//-------------------------------------------------------------------------//
		public void Strafe(Vector3 delta)
		{
			m_EyePosition    += delta;
			m_LookAtPosition += delta;

			this.OnCameraUpdate(new CameraEventArgs());
		}

		//-------------------------------------------------------------------------//
		public void DrawCamera(PrimitiveRenderer r)
		{
			if (!m_ShowCamera) return;

			// Draw a camera mesh at the eye position
			SceneNode c = new SceneNode(new LoadedPrimitive(CAMERAMESH), new Transform());
			c.P.Lighted = false;
			c.P.Filled = false;
			c.P.Color = System.Drawing.Color.Black;

			c.T.T = m_EyePosition;
			Vector3 direction = (m_EyePosition - m_LookAtPosition).Normalize();
			c.T.RQ = new Quaternion(
				new Vector3(0, 1, 0).CrossProduct(direction), 
				(float)System.Math.Acos(new Vector3(0, 1, 0) * direction)  * (180.0f/(float)System.Math.PI));

			c.DrawScene(r);

			// Draw a sphere at the look at position
			SceneNode s = new SceneNode(new Sphere(), new Transform());
			s.P.MoveTo(new Vector3(0, 0, 0));
			s.P.SizeTo(new Vector3(0, 1, 0));
			s.P.Lighted = false;
			s.P.Filled  = false;
			s.P.Color = System.Drawing.Color.Black;

			s.T.T = m_LookAtPosition;
			s.T.S = new Vector3(0.5f, 0.5f, 0.5f);
			s.DrawScene(r);

			// Draw a line between the two
			Line3d l = new Line3d();
			l.Lighted = false;
			l.Shaded  = false;
			l.Color   = System.Drawing.Color.Black;
			l.SetEnds(
				m_EyePosition.X, m_EyePosition.Y, m_EyePosition.Z,
				m_LookAtPosition.X, m_LookAtPosition.Y, m_LookAtPosition.Z);
			r.DrawPrimitive(l);

		}

		//-------------------------------------------------------------------------//
		protected virtual void OnCameraUpdate(CameraEventArgs e)
		{
			if (CameraUpdate != null)
				CameraUpdate(this, e);
		}
	}

	/// <summary>
	/// Event Handler Argument & Delegate for Camera Event
	/// </summary>
	public delegate void CameraHandler(object sender, CameraEventArgs e);
	public class CameraEventArgs: EventArgs
	{
		public CameraEventArgs()
		{
			
		}
	}
}
