using System;
using System.Collections;
using System.Drawing;

using JPGL.Math;


namespace JPGL.Common.Lighting
{
	/// <summary>
	///  A Light
	/// </summary>
	public abstract class Light
	{
		#region Static Light Type Definitions
		//-----------------------------------//
		public static int LIGHT_POINT       = 0;
		public static int LIGHT_DIRECTIONAL = 1;
		public static int LIGHT_SPOT        = 2;
		//-----------------------------------//
		#endregion

		//-----------------------------------//
		public abstract int LightType
		{
			get;
		}
		//-----------------------------------//


		//-----------------------------------//
		protected bool m_On;                 // Is the light on?
		protected bool m_Show;               // Display the light?
		//-----------------------------------//

		//-----------------------------------// Colors/intensities
		protected Color m_DiffuseColor;
		protected Color m_AmbientColor;
		protected Color m_SpecularColor;
		//-----------------------------------//

		//-----------------------------------//
		protected float m_Range;            
		protected float m_Falloff;
		//-----------------------------------//

		//-----------------------------------//
		protected float m_Attenuation0;
		protected float m_Attenuation1;
		protected float m_Attenuation2;
		//-----------------------------------//

		//-----------------------------------//
		protected Vector3 m_Position;
		protected Vector3 m_Direction;
		//-----------------------------------//

		//-----------------------------------//
		protected float m_Theta;             // Inner cone radius
		protected float m_Phi;               // Outer cone radius
		//-----------------------------------//


		#region Public Property Accessors
		//-----------------------------------//
		public bool On
		{
			get { return m_On;  }
			set { m_On = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public bool Show
		{
			get { return m_Show;  }
			set { m_Show = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public Color DiffuseColor
		{
			get { return m_DiffuseColor;  }
			set { m_DiffuseColor = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public Color AmbientColor
		{
			get { return m_AmbientColor;  }
			set { m_AmbientColor = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public Color SpecularColor
		{
			get { return m_SpecularColor;  }
			set { m_SpecularColor = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Range
		{
			get { return m_Range;  }
			set { m_Range = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Falloff
		{
			get { return m_Falloff;  }
			set { m_Falloff = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Attenuation0
		{
			get { return m_Attenuation0;  }
			set { m_Attenuation0 = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Attenuation1
		{
			get { return m_Attenuation1;  }
			set { m_Attenuation1 = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Attenuation2
		{
			get { return m_Attenuation2;  }
			set { m_Attenuation2 = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Phi
		{
			get { return m_Phi;  }
			set { m_Phi = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public float Theta
		{
			get { return m_Theta;  }
			set { m_Theta = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public Vector3 Position
		{
			get { return m_Position;  }
			set { m_Position = value; }
		}
		//-----------------------------------//
		
		//-----------------------------------//
		public Vector3 Direction
		{
			get { return m_Direction;  }
			set 
			{ 
				m_Direction = value.Normalize(); 
			}
		}
		//-----------------------------------//
		#endregion


		//-------------------------------------------------------------------------//
		public Light()
		{
			m_On   = true;
			m_Show = false;

			m_DiffuseColor  = Color.FromArgb(255, 204, 25, 25);
			m_AmbientColor  = Color.FromArgb(255, 25, 25, 25);
			m_SpecularColor = Color.FromArgb(255, 204, 25, 25);

			m_Range   = 200.0f;
			m_Falloff = 20.0f;

			m_Attenuation0 = 1.0f;
			m_Attenuation1 = 0.0f;
			m_Attenuation2 = 0.0f;

			m_Position  = new Vector3();
			m_Direction = new Vector3(0, 1, 0);

			m_Theta    = (float)System.Math.PI / 4.0f;
			m_Phi      = (float)System.Math.PI / 3.0f;
		}

		//-------------------------------------------------------------------------//
		public void Display(JPGL.Common.Primitives.PrimitiveRenderer r)
		{
			if (m_Show) DrawLight(r);
			r.AddLight(this);
		}

		//-------------------------------------------------------------------------//
		public abstract void DrawLight(JPGL.Common.Primitives.PrimitiveRenderer r);

	}

	/// <summary>
	/// LightList is a list of Lights.
	/// </summary>
	public class LightList : IEnumerable
	{
		//-----------------------------------//
		private ArrayList m_Lights;          //
		//-----------------------------------//

		//-----------------------------------//
		public Light this[int index]
		{
			get { return (Light)m_Lights[index]; }
			set { m_Lights[index] = value;       }
		}
		//-----------------------------------//
		

		//-------------------------------------------------------------------------//
		public LightList()
		{
			m_Lights = new ArrayList();
		}

		//-------------------------------------------------------------------------//
		public void AddLight(Light l)
		{
			m_Lights.Add(l);
		}

		//-------------------------------------------------------------------------//
		public void Clear()
		{
			m_Lights.Clear();
		}


		//-------------------------------------------------------------------------//
		// Enumeration functions (so we can use foreach ... in LightList)
		public IEnumerator GetEnumerator()
		{
			return new LightListEnumerator(this);
		} 

		/// <summary>
		/// Private Enumerator so we can use foreach on our lists.
		/// </summary>
		private class LightListEnumerator : IEnumerator
		{
			//-----------------------------------//
			private LightList m_List;            // Reference to list
			private int m_Position;              // Position used for enumeration
			//-----------------------------------//

			public LightListEnumerator(LightList list)
			{
				m_List     = list;
				m_Position = -1;
			}

			//-------------------------------------------------------------------------//
			public bool MoveNext()
			{
				m_Position++;
				if (m_Position >= m_List.m_Lights.Count)
				{
					return false;
				}

				return true;
			} 

			//-------------------------------------------------------------------------//
			public void Reset()
			{
				m_Position = -1;
			} 

			//-------------------------------------------------------------------------//
			public object Current
			{
				get
				{
					return (m_List.m_Lights[m_Position]);
				}
			}
		}
	}
}
