using System;
using System.Drawing;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// DrawablePrimitive is a Primitive that can has attributes
	/// related to displaying itself on the screen. It also adds
	/// an interface for getting DrawablePrimitive geometry.
	/// 
	/// The following public properties are provided:
	///  Visibility - Is the Primitive visible?
	///  Color - The color of the Primitive.
	///  PointSize - Point Size used for drawing (in point mode).
	///  
	/// Implementing classes must implement these functions:
	///  CalculateVertices - Calculate the geometry of the Primitive
	///  GetDrawMode - How do the Vertices in the geometry relate?
	///  
	/// Implementing classes may override these properties:
	///  Shaded - Is the Primitive shaded?
	///  Filled - Is the Primitive filled?
	///  Wireframe - If the Primitive isn't filled, is it drawn in wireframe (or points)?
	/// </summary>
	public abstract class DrawablePrimitive : Primitive
	{
		#region Display Data Members and Properties
		//-----------------------------------//
		protected bool      m_Visibility;    // Visibility Attribute
		public virtual bool Visibility       //
		{
			get { return m_Visibility;  }
			set { m_Visibility = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected float     m_PointSize;     // PointSize Attribute
		public float        PointSize        //
		{
			get { return m_PointSize;  }
			set { m_PointSize = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public virtual bool Shaded           // DrawablePrimitives are not shadable
		{                                    // by default. Implementing classes
			get { return false; }            // may override if they wish to be
			set { }                          // shadeable.
		}
		//-----------------------------------//

		//-----------------------------------//
		public virtual bool Filled           // By default, nothing is fillable
		{                                
			get { return false; }         
			set {}
		}
		//-----------------------------------//

		//-----------------------------------//
		protected bool      m_Wireframe;     // If filled is false, this goes into
		public virtual bool Wireframe        // effect. Wireframe = false means points.
		{                                    // All DrawablePrimitives, fillable or not
			get { return m_Wireframe;  }     // can toggle between wire and point modes.
			set { m_Wireframe = value; }
		}
		//-----------------------------------//
		#endregion

		#region Color Data Members and Properties
		//-----------------------------------//
		protected Color     m_Color;	     // Color Attribute
		public    Color     Color            //
		{
			get { return m_Color;  }
			set 
			{ 
				m_Color = value; 
				GenColorArray();	
			}
		}
		//-----------------------------------//
		protected float[] m_ColorArray;
		public float[] ColorArray
		{
			get { return m_ColorArray; }
		}
		//-----------------------------------//

		//-----------------------------------// alpha blended?
		protected bool m_Blended;
		public virtual bool Blended
		{
			get { return m_Blended;  }
			set { m_Blended = value; }
		}
		//-----------------------------------//
		#endregion

		#region Texture Data Members and Properties
		//-----------------------------------// Is texture on, if yes, m_Texture
		protected bool m_TextureOn;          // is index, though texture on and
		//-----------------------------------// index -1 is the same as texture off
		public virtual bool TextureOn
		{
			get { return (m_TextureOn && m_Texture != null && m_Texture.Index != -1);  }
			set { m_TextureOn = value; }
		}
		//-----------------------------------//

		//-----------------------------------// Index of texture, -1 for no texture
		protected Texture m_Texture;
		//-----------------------------------//
		public virtual Texture Texture
		{
			get { return m_Texture;  }
			set { m_Texture = value; m_TextureOn = true; }
		}
		//-----------------------------------//
		#endregion

		#region Illumination Data Members and Properties
		//-----------------------------------// illuminated
		protected bool m_Lighted;
		public virtual bool Lighted
		{
			get { return m_Lighted;  }
			set { m_Lighted = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected Material m_Material;
		public Material Material
		{
			get { return m_Material;  }
			set { m_Material = value; }
		}
		//-----------------------------------//
		#endregion

		//-------------------------------------------------------------------------//
		// Default Constructor - Set default values of attributes
		public DrawablePrimitive() : base()
		{
			m_Visibility    = true;
			m_Color         = Color.White;
			GenColorArray();
			m_PointSize     = 2.0f;
			m_Wireframe     = true;
			m_Texture       = null;
			m_TextureOn     = false;
			m_Blended       = false;
			m_Lighted       = true;
			m_Material      = new Material();
		}

		//-------------------------------------------------------------------------//
		// Copy Constructor 
		public DrawablePrimitive(DrawablePrimitive p) : base(p)
		{
			m_Visibility    = p.Visibility;
			m_Color         = p.Color;
			GenColorArray();
			m_PointSize     = p.PointSize;
			m_Wireframe     = p.Wireframe;
			m_Texture       = p.Texture;
			m_TextureOn     = p.TextureOn;
			m_Lighted       = p.Lighted;
			m_Blended       = p.Blended;
			m_Material      = new Material(p.Material);
		}

		//-------------------------------------------------------------------------//
		// Get the basic unit geometry the primitive uses.
		public abstract Geometry GetGeometry();

		//-------------------------------------------------------------------------//
		public virtual void DrawExtra(PrimitiveRenderer r)
		{
			// An implementing class may override this function, which 
			// is called after the primitive is drawn. This allows the
			// primitive to draw something extra that is not normally
			// a part of its geometry.
		}

		//-------------------------------------------------------------------------//
		protected virtual void GenColorArray()
		{
			if (GetGeometry() == null) return;

			int size = GetGeometry().VertexCount;
			if (size <= 0) return;

			if (m_ColorArray == null || m_ColorArray.Length != size*4)
				m_ColorArray = new float[size*4];

			float red   = m_Color.R / 255.0f;
			float green = m_Color.G / 255.0f;
			float blue  = m_Color.B / 255.0f;
			float alpha = m_Color.A / 255.0f;

			if (Shaded)
			{
				m_ColorArray[0] = System.Drawing.Color.Blue.R / 255.0f;
				m_ColorArray[1] = System.Drawing.Color.Blue.G / 255.0f;
				m_ColorArray[2] = System.Drawing.Color.Blue.B / 255.0f;
				m_ColorArray[3] = System.Drawing.Color.Blue.A / 255.0f;
			}
			else
			{
				m_ColorArray[0] = red;
				m_ColorArray[1] = green;
				m_ColorArray[2] = blue;
				m_ColorArray[3] = alpha;
			}

			for (int i=1; i<size; i++)
			{
				m_ColorArray[i*4] = red; 
				m_ColorArray[(i*4)+1] = green; 
				m_ColorArray[(i*4)+2] = blue;
				m_ColorArray[(i*4)+3] = alpha;
			}
		}
	}
}
