using System;
using System.Drawing;

namespace JPGL.Common.Primitives
{
	/// <summary>
	/// PrimitiveGenerator is an interface that provides a single 
	/// function which returns a new DrawablePrimitive. Thus, a 
	/// PrimitaveGenerator can create any kind of primitive that
	/// has an implementing Generator.
	/// </summary>
	public abstract class DrawablePrimitiveGenerator
	{
		private float m_PointSize;
		public float  PointSize
		{
			set { m_PointSize = value; }
		}
		private Color m_Color;
		public Color  Color
		{
			set { m_Color = value; }
		}
		private bool  m_Shaded;
		public bool   Shaded
		{
			set { m_Shaded = value; }
		}
		private bool  m_Filled;
		public bool   Filled
		{
			set { m_Filled = value; }
		}
		private bool  m_Wireframe;
		public bool   Wireframe
		{
			set { m_Wireframe = value; }
		}
		private bool  m_ShowColliders;
		public bool   ShowColliders
		{
			set { m_ShowColliders = value; }
		}

		//-------------------------------------------------------------------------//
		public DrawablePrimitiveGenerator()
		{
			m_PointSize     = 1.0f;
			m_Color         = System.Drawing.Color.White;
			m_Shaded        = true;
			m_Filled        = true;
			m_Wireframe     = true;
			m_ShowColliders = false;
		}
		//-------------------------------------------------------------------------//
		// Create new generator with same attribute values as the given generator.
		public DrawablePrimitiveGenerator(DrawablePrimitiveGenerator old) : this()
		{
			if (old == null) return;

			m_PointSize     = old.m_PointSize;
			m_Color         = old.m_Color;
			m_Shaded        = old.m_Shaded;
			m_Filled        = old.m_Filled;
			m_Wireframe     = old.m_Wireframe;
			m_ShowColliders = old.m_ShowColliders;

		}

		//-------------------------------------------------------------------------//
		public DrawablePrimitive GetNewPrimitive()
		{
			// Get the new primitive
			DrawablePrimitive p = GeneratePrimitive();

			// Set attributes of new primitive
			p.PointSize = m_PointSize;
			p.Color     = m_Color;
			p.Shaded    = m_Shaded;
			p.Filled    = m_Filled;
			p.Wireframe = m_Wireframe;
			p.ShowColliders = m_ShowColliders;
			
			return p;
		}


		//-------------------------------------------------------------------------//
		// Implementing generators must generate a new primitive with this function.
		protected abstract DrawablePrimitive GeneratePrimitive();
	}


	//-------------------------------------------------------------------------//
	public class RandomGenerator : DrawablePrimitiveGenerator
	{
		private Random m_RandGen;

		public RandomGenerator() : base() 
		{
			m_RandGen = new Random();
		}
		public RandomGenerator(DrawablePrimitiveGenerator o) : base(o) 
		{
			m_RandGen = new Random();
		}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			// Get primitive type randomly
			DrawablePrimitive p;
			int ptype = m_RandGen.Next(0, 5);
			if (ptype == 0)      p = new Circle();
			else if (ptype == 1) p = new Line();
			else if (ptype == 2) p = new Triangle();
			else if (ptype == 3) p = new Point();
			else if (ptype == 4) p = new Rectangle();
			else                 p = new Circle4Collider();

			// Set primitive attribs randomly
			PointSize = (float)m_RandGen.Next(1, 10);
			Color     = System.Drawing.Color.FromArgb(m_RandGen.Next(0,255), m_RandGen.Next(0,255), m_RandGen.Next(0,255));
			Shaded    = (m_RandGen.Next(0,10) > 5 ? true : false);
			Filled    = (m_RandGen.Next(0,10) > 5 ? true : false);
			Wireframe = (m_RandGen.Next(0,10) > 5 ? true : false);
			ShowColliders = false;

			return p;
		}
	}


	//-------------------------------------------------------------------------//
	public class CircleGenerator : DrawablePrimitiveGenerator	
	{
		public CircleGenerator() : base() {}
		public CircleGenerator(DrawablePrimitiveGenerator o) : base(o) {}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			return new Circle();
		}
	}

	//-------------------------------------------------------------------------//
	public class LineGenerator : DrawablePrimitiveGenerator	
	{
		public LineGenerator() : base() {}
		public LineGenerator(DrawablePrimitiveGenerator o) : base(o) {}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			return new Line();
		}
	}

	//-------------------------------------------------------------------------//
	public class TriangleGenerator : DrawablePrimitiveGenerator	
	{
		public TriangleGenerator() : base() {}
		public TriangleGenerator(DrawablePrimitiveGenerator o) : base(o) {}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			return new Triangle();
		}
	}

	//-------------------------------------------------------------------------//
	public class PointGenerator : DrawablePrimitiveGenerator	
	{
		public PointGenerator() : base() {}
		public PointGenerator(DrawablePrimitiveGenerator o) : base(o) {}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			return new Point();
		}
	}

	//-------------------------------------------------------------------------//
	public class RectangleGenerator : DrawablePrimitiveGenerator	
	{
		public RectangleGenerator() : base() {}
		public RectangleGenerator(DrawablePrimitiveGenerator o) : base(o) {}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			return new Rectangle();
		}
	}

	//-------------------------------------------------------------------------//
	public class Circle4ColliderGenerator : DrawablePrimitiveGenerator	
	{
		public Circle4ColliderGenerator() : base() {}
		public Circle4ColliderGenerator(DrawablePrimitiveGenerator o) : base(o) {}
		protected override DrawablePrimitive GeneratePrimitive()
		{
			return new Circle4Collider();
		}
	}
}
