using System;
using System.Collections;
using System.IO;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// Geometry objects are instance interfaces to static functionality. The static
	/// functionality is the generation of vertices for unit sized geometry. The list
	/// of vertices for any given type of geometry is only ever generated once (upon
	/// demand), and all subsequent shapes that which to use this geometry will 
	/// reference it through an instance of a Geometry object. Thus every Circle
	/// is drawn using the exact same float array of vertices, and etc.
	/// </summary>
	public struct LoadedGoemetry
	{
		public float[] vertices;
		public float[] texels;
		public float[] normals;
		public int     nVertices;
	}
	public class Geometry
	{
		//-----------------------------------//
		private int    m_GeometryType;       // This Geometry object's type
		public int     GeometryID
		{
			get { return m_GeometryType; }
		}
		//-----------------------------------//

		//-----------------------------------// Is this Geometry's data in 
		protected bool m_Indexed;            // face index form? 
		public  bool Indexed
		{
			get { return m_Indexed;  }
			set { m_Indexed = value; }
		}
		//-----------------------------------//

		//-----------------------------------// Any cached versions of this
		protected bool m_Dirty;              // Geometry may be out of date
		public  bool Dirty
		{
			get { return m_Dirty;  }
			set { m_Dirty = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected bool m_AutoCache;
		public    bool AutoCache
		{
			get { return m_AutoCache;  }
			set { m_AutoCache = value; }
		}
		//-----------------------------------//

		#region Data Accessor Properties
		//-----------------------------------// Get the vertices of this object
		public virtual float[] Vertices
		{
			get 
			{ 
				if (m_GeometryType == Geometry.CIRCLE)   return m_CircleVertices;
				if (m_GeometryType == Geometry.TRIANGLE) return m_TriangleVertices;
				if (m_GeometryType == Geometry.LINE)     return m_LineVertices;
				if (m_GeometryType == Geometry.POINT)    return m_PointVertices;
				if (m_GeometryType == Geometry.SQUARE)   return m_SquareVertices;
				if (m_GeometryType == Geometry.SPHERE)   return m_SphereVertices;
				if (m_GeometryType == Geometry.CUBE)     return m_CubeVertices;
				if (m_GeometryType >= LOADED)
				{
					return ((LoadedGoemetry)LoadedObj[(m_GeometryType-10)]).vertices;
				}
				return null; 
			}
			set {}
		}
		//-----------------------------------//

		//-----------------------------------// Get the vertices of this object
		public virtual float[] Texels
		{
			get 
			{ 
				if (m_GeometryType == Geometry.CIRCLE)   return m_CircleTexels;
				if (m_GeometryType == Geometry.TRIANGLE) return m_TriangleTexels;
				if (m_GeometryType == Geometry.LINE)     return m_LineTexels;
				if (m_GeometryType == Geometry.POINT)    return m_PointTexels;
				if (m_GeometryType == Geometry.SQUARE)   return m_SquareTexels;
				if (m_GeometryType == Geometry.SPHERE)   return m_SphereTexels;
				if (m_GeometryType == Geometry.CUBE)     return m_CubeTexels;
				if (m_GeometryType >= LOADED)
				{
					return ((LoadedGoemetry)LoadedObj[(m_GeometryType-LOADED)]).texels;
				}
				return null; 
			}
			set {}
		}
		//-----------------------------------//

		//-----------------------------------// Get the vertices of this object
		public virtual float[] Normals
		{
			get 
			{ 
				if (m_GeometryType == Geometry.CIRCLE)   return m_CircleNormals;
				if (m_GeometryType == Geometry.TRIANGLE) return m_TriangleNormals;
				if (m_GeometryType == Geometry.LINE)     return m_LineNormals;
				if (m_GeometryType == Geometry.POINT)    return m_PointNormals;
				if (m_GeometryType == Geometry.SQUARE)   return m_SquareNormals;
				if (m_GeometryType == Geometry.SPHERE)   return m_SphereNormals;
				if (m_GeometryType == Geometry.CUBE)     return m_CubeNormals;
				if (m_GeometryType >= LOADED)
				{
					return ((LoadedGoemetry)LoadedObj[(m_GeometryType-LOADED)]).normals;
				}
				return null; 
			}
			set {}
		}
		//-----------------------------------//

		//-----------------------------------// Get number of vertices
		public virtual int VertexCount
		{
			get 
			{
				if (m_GeometryType == Geometry.CIRCLE)   return m_CircleVCount;
				if (m_GeometryType == Geometry.TRIANGLE) return m_TriangleVCount;
				if (m_GeometryType == Geometry.LINE)     return m_LineVCount;
				if (m_GeometryType == Geometry.POINT)    return m_PointVCount;
				if (m_GeometryType == Geometry.SQUARE)   return m_SquareVCount;
				if (m_GeometryType == Geometry.SPHERE)   return m_SphereVCount;
				if (m_GeometryType == Geometry.CUBE)     return m_CubeVCount;
				if (m_GeometryType >= LOADED)
				{
					return ((LoadedGoemetry)LoadedObj[(m_GeometryType-LOADED)]).nVertices;
				}
				return -1; 
			}
			set {}
		}
		//-----------------------------------//

		//-----------------------------------// Get the drawmode of this object
		public virtual int DrawMode
		{
			get 
			{
				if (m_GeometryType == Geometry.CIRCLE)   return m_CircleMode;
				if (m_GeometryType == Geometry.TRIANGLE) return m_TriangleMode;
				if (m_GeometryType == Geometry.LINE)     return m_LineMode;
				if (m_GeometryType == Geometry.POINT)    return m_PointMode;
				if (m_GeometryType == Geometry.SQUARE)   return m_SquareMode;
				if (m_GeometryType == Geometry.SPHERE)   return m_SphereMode;
				if (m_GeometryType == Geometry.CUBE)     return m_CubeMode;
				if (m_GeometryType >= LOADED)
				{
					return PrimitiveRenderer.TRIANGLES;
				}
				return -1; 
			}
			set {}
		}
		//-----------------------------------//
		#endregion

		//-------------------------------------------------------------------------//
		// Given a Geometry Type, create a new geometry object.
		protected Geometry(int type)
		{
			m_GeometryType = type;
			m_Dirty        = false;
			m_Indexed      = false;
			m_AutoCache    = true;
		}


		//-----------------------------------// Basic Unit Size Geometry Types
		public static int CIRCLE   = 0;      // 
		public static int TRIANGLE = 1;      //
		public static int LINE     = 2;      //
		public static int POINT    = 3;      //
		public static int SQUARE   = 4;      //
		public static int SPHERE   = 5;      //
		public static int CUBE     = 6;      //
		//-----------------------------------//

		//-----------------------------------//
		private static int LOADED = 10;      // Offset for loaded types
		//-----------------------------------//

		//-----------------------------------//
		private static ArrayList LoadedObj = new ArrayList();
		//-----------------------------------//

		//-------------------------------------------------------------------------//
		public static Geometry GetLoadedGeometry(int index)
		{
			return new Geometry(index);
		}


		#region CircleGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_CircleVertices;
		private static float[] m_CircleTexels;
		private static float[] m_CircleNormals;
		private static int     m_CircleVCount;
		private static int     m_CircleMode;
		public static Geometry CircleGeometry
		{
			get
			{
				if (m_CircleVertices == null)
				{
					int numPts = 100;
					m_CircleVertices = new float[(numPts+2)*3];
					m_CircleTexels   = new float[(numPts+2)*2];
					m_CircleNormals  = new float[(numPts+2)*3];
					
					float delta = (float)((2.0f * System.Math.PI) / numPts);
					m_CircleVertices[0] = 0.0f;
					m_CircleVertices[1] = 0.0f;
					m_CircleVertices[2] = 0.0f;

					for (int i=1; i<=(numPts); i++)
					{
						m_CircleVertices[i*3]   = (float)(System.Math.Sin((i-1)*delta));
						m_CircleVertices[i*3+1] = (float)(System.Math.Cos((i-1)*delta));
						m_CircleVertices[i*3+2] = 0.0f;
					}
					m_CircleVertices[((numPts+1)*3)]   = m_CircleVertices[3];
					m_CircleVertices[((numPts+1)*3)+1] = m_CircleVertices[4];
					m_CircleVertices[((numPts+1)*3)+2] = m_CircleVertices[5];

					for (int i=0; i<=(numPts+1); i++)
					{
						m_CircleTexels[(i*2)+0] = m_CircleVertices[(i*3)+0];
						m_CircleTexels[(i*2)+1] = m_CircleVertices[(i*3)+1];

						m_CircleNormals[(i*3)+0] = 0;
						m_CircleNormals[(i*3)+1] = 0;
						m_CircleNormals[(i*3)+2] = 1;
					}

					
					m_CircleVCount = numPts + 2;
					m_CircleMode   = PrimitiveRenderer.TRIANGLE_FAN;
				}				
				return new Geometry(Geometry.CIRCLE);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region SphereGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_SphereVertices;
		private static float[] m_SphereTexels;
		private static float[] m_SphereNormals;
		private static int     m_SphereVCount;
		private static int     m_SphereMode;
		public static Geometry SphereGeometry
		{
			get
			{
				if (m_SphereVertices == null)
				{
					m_SphereVCount   = 703*2;
					m_SphereMode     = PrimitiveRenderer.TRIANGLE_STRIP;
					m_SphereVertices = new float[(703*2)*3];
					m_SphereTexels   = new float[(703*2)*2];
					m_SphereNormals  = new float[(703*2)*3];

					double c  = System.Math.PI / 180; // For degrees -> rads conversion
					int index = 0; int j=18;
					for (double eps = -90.0; eps <= 90.0; eps += 10.0)
					{
						int i = 0;
						for (double theta = -180.0; theta <= 180.0; theta += 10.0) 
						{
							m_SphereVertices[(index*6)+0] = (float)(System.Math.Cos(c*theta) * System.Math.Cos(c*eps));
							m_SphereVertices[(index*6)+1] = (float)(System.Math.Sin(c*eps));
							m_SphereVertices[(index*6)+2] = (float)(System.Math.Sin(c*theta) * System.Math.Cos(c*eps));

							m_SphereVertices[(index*6)+3] = (float)(System.Math.Cos(c*theta) * System.Math.Cos(c*(eps+10.0)));
							m_SphereVertices[(index*6)+4] = (float)(System.Math.Sin(c*(eps+10.0)));
							m_SphereVertices[(index*6)+5] = (float)(System.Math.Sin(c*theta) * System.Math.Cos(c*(eps+10.0)));

							m_SphereTexels[(index*4)+0] = 1.0f - ((float)i)/36.0f;
							m_SphereTexels[(index*4)+1] = 1.0f - (((float)j))/18.0f;

							m_SphereTexels[(index*4)+2] = 1.0f - ((float)i)/36.0f;
							m_SphereTexels[(index*4)+3] = 1.0f - (((float)j)-1.0f)/18.0f;

							m_SphereNormals[(index*6)+0] = 0;
							m_SphereNormals[(index*6)+1] = 0;
							m_SphereNormals[(index*6)+2] = 1;

							m_SphereNormals[(index*6)+3] = 0;
							m_SphereNormals[(index*6)+4] = 0;
							m_SphereNormals[(index*6)+5] = 1;

							index++; i++;
						}
						j--;
					}
				}				
				return new Geometry(Geometry.SPHERE);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region TriangleGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_TriangleVertices;
		private static float[] m_TriangleTexels;
		private static float[] m_TriangleNormals;
		private static int     m_TriangleVCount;
		private static int     m_TriangleMode;
		public static Geometry TriangleGeometry
		{
			get
			{
				if (m_TriangleVertices == null)
				{
					m_TriangleVertices = new float[3*3];
					m_TriangleTexels   = new float[3*2];
					m_TriangleNormals  = new float[3*3];

					// Root 3 over 4
					float r3o4 = ((float)System.Math.Sqrt(3)) / 4.0f;

					m_TriangleVertices[0] = 0.0f; m_TriangleTexels[0] = 0;
					m_TriangleVertices[1] = r3o4; m_TriangleTexels[1] = 1;
					m_TriangleVertices[2] = 0.0f;

					m_TriangleVertices[3] = 0.5f; m_TriangleTexels[2] = 0;
					m_TriangleVertices[4] = -1.0f*r3o4;
					m_TriangleVertices[5] = 0.0f; m_TriangleTexels[3] = 0;

					m_TriangleVertices[6] = -0.5f; m_TriangleTexels[4] = 1;
					m_TriangleVertices[7] = -1.0f*r3o4;
					m_TriangleVertices[8] = 0.0f; m_TriangleTexels[5] = 0;

					for (int i=0; i<3; i++)
					{
						m_TriangleNormals[(i*3)+0] = 0;
						m_TriangleNormals[(i*3)+1] = 0;
						m_TriangleNormals[(i*3)+2] = 1;
					}
					m_TriangleVCount = 3;
					m_TriangleMode = PrimitiveRenderer.TRIANGLES;
				}
				return new Geometry(Geometry.TRIANGLE);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region LineGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_LineVertices;
		private static float[] m_LineTexels;
		private static float[] m_LineNormals;
		private static int     m_LineVCount;
		private static int     m_LineMode;
		public static Geometry LineGeometry
		{
			get
			{
				if (m_LineVertices == null)
				{
					m_LineVertices = new float[2*3];
					m_LineTexels   = new float[2*3];
					m_LineNormals  = new float[2*3];

					// One end of the line
					m_LineVertices[0] = -0.5f;
					m_LineVertices[1] = 0.0f;
					m_LineVertices[2] = 0.0f;

					// The other end of the line
					m_LineVertices[3] = 0.5f;
					m_LineVertices[4] = 0.0f;
					m_LineVertices[5] = 0.0f;

					for (int i=0; i<2; i++)
					{
						m_LineNormals[(i*3)+0] = 0;
						m_LineNormals[(i*3)+1] = 0;
						m_LineNormals[(i*3)+2] = 1;
					}

					m_LineVCount = 2;
					m_LineMode   = PrimitiveRenderer.LINES;

					
				}
				return new Geometry(Geometry.LINE);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region PointGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_PointVertices;
		private static float[] m_PointTexels;
		private static float[] m_PointNormals;
		private static int     m_PointVCount;
		private static int     m_PointMode;
		public static Geometry PointGeometry
		{
			get
			{
				if (m_PointVertices == null)
				{
					m_PointVertices = new float[3];
					m_PointTexels   = new float[2];
					m_PointNormals  = new float[3];

					m_PointVertices[0] = 0.0f; m_PointTexels[0] = 0;
					m_PointVertices[1] = 0.0f; m_PointTexels[1] = 1;
					m_PointVertices[2] = 0.0f;

					m_PointNormals[0] = 0;
					m_PointNormals[1] = 0;
					m_PointNormals[2] = 1;

					m_PointVCount = 1;
					m_PointMode = PrimitiveRenderer.POINTS;
				}
				return new Geometry(Geometry.POINT);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region SquareGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_SquareVertices;
		private static float[] m_SquareTexels;
		private static float[] m_SquareNormals;
		private static int     m_SquareVCount;
		private static int     m_SquareMode;
		public static Geometry SquareGeometry
		{
			get
			{
				if (m_SquareVertices == null)
				{
					m_SquareVertices = new float[4*3];
					m_SquareTexels   = new float[4*2];
					m_SquareNormals  = new float[4*3];

					// Corner 1
					m_SquareVertices[0] = -0.5f; m_SquareTexels[0] = 0;
					m_SquareVertices[1] = 0.5f;  m_SquareTexels[1] = 1;
					m_SquareVertices[2] = 0;
					// Corner 2
					m_SquareVertices[3] = -0.5f; m_SquareTexels[2] = 0;
					m_SquareVertices[4] = -0.5f; m_SquareTexels[3] = 0;
					m_SquareVertices[5] = 0;
					// Corner 3
					m_SquareVertices[6] = 0.5f; m_SquareTexels[4] = 1;
					m_SquareVertices[7] = 0.5f;  m_SquareTexels[5] = 1;
					m_SquareVertices[8] = 0;
					// Corner 4
					m_SquareVertices[9]  = 0.5f;  m_SquareTexels[6] = 1;
					m_SquareVertices[10] = -0.5f; m_SquareTexels[7] = 0;
					m_SquareVertices[11] = 0;

					for (int i=0; i<4; i++)
					{
						m_SquareNormals[(i*3)+0] = 0;
						m_SquareNormals[(i*3)+1] = 0;
						m_SquareNormals[(i*3)+2] = 1;
					}

					m_SquareVCount = 4;
					m_SquareMode = PrimitiveRenderer.TRIANGLE_STRIP;
				}
				return new Geometry(Geometry.SQUARE);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region CubeGeometry
		//-------------------------------------------------------------------------//
		private static float[] m_CubeVertices;
		private static float[] m_CubeTexels;
		private static float[] m_CubeNormals;
		private static int     m_CubeVCount;
		private static int     m_CubeMode;
		public static Geometry CubeGeometry
		{
			get
			{
				if (m_CubeVertices == null)
				{
					m_CubeVertices = new float[24*3];
					m_CubeTexels   = new float[24*2];
					m_CubeNormals  = new float[24*3];

					#region Front Face
					m_CubeVertices[0] = -0.5f; m_CubeTexels[0] = 0;
					m_CubeVertices[1] =  0.5f; m_CubeTexels[1] = 1;
					m_CubeVertices[2] =  0.5f;  

					m_CubeVertices[3] = -0.5f; m_CubeTexels[2] = 0;
					m_CubeVertices[4] = -0.5f; m_CubeTexels[3] = 0;
					m_CubeVertices[5] =  0.5f;

					m_CubeVertices[6] =  0.5f; m_CubeTexels[4] = 1;
					m_CubeVertices[7] = -0.5f; m_CubeTexels[5] = 0;
					m_CubeVertices[8] =  0.5f;

					m_CubeVertices[9]  =  0.5f; m_CubeTexels[6] = 1;
					m_CubeVertices[10] =  0.5f; m_CubeTexels[7] = 1;
					m_CubeVertices[11] =  0.5f;

					for (int i=0; i<4; i++)
					{
						m_CubeNormals[(i*3)+0] = 0;
						m_CubeNormals[(i*3)+1] = 0;
						m_CubeNormals[(i*3)+2] = 1;
					}
					#endregion

					#region Back Face
					m_CubeVertices[12] = -0.5f; m_CubeTexels[8] = 0;
					m_CubeVertices[13] =  0.5f; m_CubeTexels[9] = 1;
					m_CubeVertices[14] = -0.5f;  

					m_CubeVertices[15] = -0.5f; m_CubeTexels[10] = 0;
					m_CubeVertices[16] = -0.5f; m_CubeTexels[11] = 0;
					m_CubeVertices[17] = -0.5f;

					m_CubeVertices[18] =  0.5f; m_CubeTexels[12] = 1;
					m_CubeVertices[19] = -0.5f; m_CubeTexels[13] = 0;
					m_CubeVertices[20] = -0.5f;

					m_CubeVertices[21] =  0.5f; m_CubeTexels[14] = 1;
					m_CubeVertices[22] =  0.5f; m_CubeTexels[15] = 1;
					m_CubeVertices[23] = -0.5f;

					for (int i=4; i<8; i++)
					{
						m_CubeNormals[(i*3)+0] = 0;
						m_CubeNormals[(i*3)+1] = 0;
						m_CubeNormals[(i*3)+2] = -1;
					}
					#endregion

					#region Top Face
					m_CubeVertices[24] = -0.5f; m_CubeTexels[16] = 0;
					m_CubeVertices[25] =  0.5f; m_CubeTexels[17] = 0;
					m_CubeVertices[26] = -0.5f;  

					m_CubeVertices[27] =  0.5f; m_CubeTexels[18] = 1;
					m_CubeVertices[28] =  0.5f; m_CubeTexels[19] = 0;
					m_CubeVertices[29] = -0.5f;

					m_CubeVertices[30] =  0.5f; m_CubeTexels[20] = 1;
					m_CubeVertices[31] =  0.5f; m_CubeTexels[21] = 1;
					m_CubeVertices[32] =  0.5f;

					m_CubeVertices[33] = -0.5f; m_CubeTexels[22] = 0;
					m_CubeVertices[34] =  0.5f; m_CubeTexels[23] = 1;
					m_CubeVertices[35] =  0.5f;

					for (int i=8; i<12; i++)
					{
						m_CubeNormals[(i*3)+0] = 0;
						m_CubeNormals[(i*3)+1] = 1;
						m_CubeNormals[(i*3)+2] = 0;
					}
					#endregion

					#region Bottom Face
					m_CubeVertices[36] = -0.5f; m_CubeTexels[24] = 0;
					m_CubeVertices[37] = -0.5f; m_CubeTexels[25] = 0;
					m_CubeVertices[38] = -0.5f;  

					m_CubeVertices[39] =  0.5f; m_CubeTexels[26] = 1;
					m_CubeVertices[40] = -0.5f; m_CubeTexels[27] = 0;
					m_CubeVertices[41] = -0.5f;

					m_CubeVertices[42] =  0.5f; m_CubeTexels[28] = 1;
					m_CubeVertices[43] = -0.5f; m_CubeTexels[29] = 1;
					m_CubeVertices[44] =  0.5f;

					m_CubeVertices[45] = -0.5f; m_CubeTexels[30] = 0;
					m_CubeVertices[46] = -0.5f; m_CubeTexels[31] = 1;
					m_CubeVertices[47] =  0.5f;

					for (int i=12; i<16; i++)
					{
						m_CubeNormals[(i*3)+0] = 0;
						m_CubeNormals[(i*3)+1] = -1;
						m_CubeNormals[(i*3)+2] = 0;
					}
					#endregion

					#region Right Face
					m_CubeVertices[48] =  0.5f; m_CubeTexels[32] = 0;
					m_CubeVertices[49] = -0.5f; m_CubeTexels[33] = 0;
					m_CubeVertices[50] =  0.5f;  

					m_CubeVertices[51] =  0.5f; m_CubeTexels[34] = 1;
					m_CubeVertices[52] = -0.5f; m_CubeTexels[35] = 0;
					m_CubeVertices[53] = -0.5f;

					m_CubeVertices[54] =  0.5f; m_CubeTexels[36] = 1;
					m_CubeVertices[55] =  0.5f; m_CubeTexels[37] = 1;
					m_CubeVertices[56] = -0.5f;

					m_CubeVertices[57] =  0.5f; m_CubeTexels[38] = 0;
					m_CubeVertices[58] =  0.5f; m_CubeTexels[39] = 1;
					m_CubeVertices[59] =  0.5f;

					for (int i=16; i<20; i++)
					{
						m_CubeNormals[(i*3)+0] = 1;
						m_CubeNormals[(i*3)+1] = 0;
						m_CubeNormals[(i*3)+2] = 0;
					}
					#endregion

					#region Left Face
					m_CubeVertices[60] = -0.5f; m_CubeTexels[40] = 0;
					m_CubeVertices[61] = -0.5f; m_CubeTexels[41] = 0;
					m_CubeVertices[62] =  0.5f;  

					m_CubeVertices[63] = -0.5f; m_CubeTexels[42] = 1;
					m_CubeVertices[64] = -0.5f; m_CubeTexels[43] = 0;
					m_CubeVertices[65] = -0.5f;

					m_CubeVertices[66] = -0.5f; m_CubeTexels[44] = 1;
					m_CubeVertices[67] =  0.5f; m_CubeTexels[45] = 1;
					m_CubeVertices[68] = -0.5f;

					m_CubeVertices[69] = -0.5f; m_CubeTexels[46] = 0;
					m_CubeVertices[70] =  0.5f; m_CubeTexels[47] = 1;
					m_CubeVertices[71] =  0.5f;

					for (int i=20; i<24; i++)
					{
						m_CubeNormals[(i*3)+0] = -1;
						m_CubeNormals[(i*3)+1] = 0;
						m_CubeNormals[(i*3)+2] = 0;
					}
					#endregion

					m_CubeVCount = 24;
					m_CubeMode = PrimitiveRenderer.QUADS;
				}
				return new Geometry(Geometry.CUBE);
			}
		}
		//-------------------------------------------------------------------------//
		#endregion

		#region OBJ Loader Functions
		//-------------------------------------------------------------------------//
		// Load OBJ file as a single object, no matter what is in it.
		public static int LoadOBJ(string fileName)
		{
			string fileNameWPath;
			if (File.Exists(fileName))
				fileNameWPath = fileName;
			else if (System.IO.File.Exists("../../../Models/" + fileName))
				fileNameWPath = "../../../Models/" + fileName;
			else if (System.IO.File.Exists("Models/" + fileName))
				fileNameWPath = "Models/" + fileName;
			else
				return -1;

			StreamReader file = new StreamReader(File.OpenRead(fileNameWPath));
			
			ArrayList vertices = new ArrayList(100);
			ArrayList finalVertices = new ArrayList(100);

			ArrayList texCoords = new ArrayList(100);
			ArrayList finalTexCoords = new ArrayList(100);

			ArrayList normals = new ArrayList(100);
			ArrayList finalNormals = new ArrayList(100);

			string line;
			while ((line = file.ReadLine()) != null)
			{
				if (line.Length == 0) continue;

				if (line[0] == 'v') // Vertex Related Data
				{
					if (line[1] == ' ')     // Vertex
					{
						string[] tokens = line.Split(' ');
						float[] vertex = {float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3])};
						vertices.Add(vertex);
					}
					else if (line[1] == 't') // Texture Coordinate
					{
						string[] tokens = line.Split(' ');
						float[] texcoord = {float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3])};
						texCoords.Add(texcoord);
					}
					else if (line[1] == 'n') // Normal
					{
						string[] tokens = line.Split(' ');
						float[] normal = {float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3])};
						normals.Add(normal);
					}
					else continue;
				}
				else if (line[0] == 'f') // Face Related Data
				{
					string[] tokens = line.Split(' ');

					finalVertices.Add(vertices[(int.Parse(tokens[1].Split('/')[0]))-1]);
					finalVertices.Add(vertices[(int.Parse(tokens[2].Split('/')[0]))-1]);
					finalVertices.Add(vertices[(int.Parse(tokens[3].Split('/')[0]))-1]);

					if (tokens[1].Split('/').Length > 1  && tokens[1].Split('/')[1].Length > 0)
					{
						finalTexCoords.Add(texCoords[(int.Parse(tokens[1].Split('/')[1]))-1]);
						finalTexCoords.Add(texCoords[(int.Parse(tokens[2].Split('/')[1]))-1]);
						finalTexCoords.Add(texCoords[(int.Parse(tokens[3].Split('/')[1]))-1]);
					}

					if (tokens[1].Split('/').Length > 2 && tokens[1].Split('/')[2].Length > 0)
					{
						finalNormals.Add(normals[(int.Parse(tokens[1].Split('/')[2]))-1]);
						finalNormals.Add(normals[(int.Parse(tokens[2].Split('/')[2]))-1]);
						finalNormals.Add(normals[(int.Parse(tokens[3].Split('/')[2]))-1]);
					}

				}
			}
			
			LoadedGoemetry g = new LoadedGoemetry();
			g.vertices = new float[finalVertices.Count*3];
			g.texels   = new float[finalTexCoords.Count*2];
			g.normals  = new float[finalNormals.Count*3];
			for (int i=0; i<finalVertices.Count; i++)
			{
				g.vertices[(i*3)+0] = ((float[])finalVertices[i])[0];
				g.vertices[(i*3)+1] = ((float[])finalVertices[i])[1];
				g.vertices[(i*3)+2] = ((float[])finalVertices[i])[2];

				if (finalTexCoords.Count > i)
				{
					g.texels[(i*2)+0] = ((float[])finalTexCoords[i])[0];
					g.texels[(i*2)+1] = ((float[])finalTexCoords[i])[1];
				}

				if (finalNormals.Count > i)
				{
					g.normals[(i*3)+0] = ((float[])finalNormals[i])[0];
					g.normals[(i*3)+1] = ((float[])finalNormals[i])[1];
					g.normals[(i*3)+2] = ((float[])finalNormals[i])[2];
				}
			}

			g.nVertices = finalVertices.Count;


			return (LoadedObj.Add(g) + LOADED);
		}

		//-------------------------------------------------------------------------//
		// Load OBJ file as a multiple objects, assuming it has multiple objects in
		// it.
		// Returns the number of objects loaded. GeometryIDs are put in objects array.
		public static int LoadMultiOBJ(string fileName, out ArrayList objects)
		{
			objects = new ArrayList();

			string fileNameWPath;
			if (File.Exists(fileName))
				fileNameWPath = fileName;
			else if (System.IO.File.Exists("../../../Models/" + fileName))
				fileNameWPath = "../../../Models/" + fileName;
			else if (System.IO.File.Exists("Models/" + fileName))
				fileNameWPath = "Models/" + fileName;
			else
				return -1;

			StreamReader file = new StreamReader(File.OpenRead(fileNameWPath));

			// CurrentObject vertices and texcoords
			ArrayList cVertices  = null;
			ArrayList cTexCoords = null;
			ArrayList cNormals   = null;
			
			// ArrayLists that all vertex and texcoord data will be loaded in
			ArrayList vertices  = new ArrayList(100);
			ArrayList texCoords = new ArrayList(100);
			ArrayList normals   = new ArrayList(100);

			string line;
			while ((line = file.ReadLine()) != null)
			{
				if (line.Length == 0) continue;

				if (line[0] == 'v')
				{
					if (line[1] == ' ')      // Vertex
					{
						string[] tokens = line.Split(' ');
						float[] vertex = {float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3])};
						vertices.Add(vertex);
					}
					else if (line[1] == 't') // Texture Coordinate
					{
						string[] tokens = line.Split(' ');
						float[] texcoord = {float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3])};
						texCoords.Add(texcoord);
					}
					else if (line[1] == 'n') // normal Coordinate
					{
						string[] tokens = line.Split(' ');
						float[] normal = {float.Parse(tokens[1]), float.Parse(tokens[2]), float.Parse(tokens[3])};
						normals.Add(normal);
					}
					else continue;
				}
				else if (line[0] == 'f')     // Face
				{
					string[] tokens = line.Split(' ');

					cVertices.Add(vertices[(int.Parse(tokens[1].Split('/')[0]))-1]);
					cVertices.Add(vertices[(int.Parse(tokens[2].Split('/')[0]))-1]);
					cVertices.Add(vertices[(int.Parse(tokens[3].Split('/')[0]))-1]);

					if (tokens[1].Split('/')[1].Length > 0)
					{
						cTexCoords.Add(texCoords[(int.Parse(tokens[1].Split('/')[1]))-1]);
						cTexCoords.Add(texCoords[(int.Parse(tokens[2].Split('/')[1]))-1]);
						cTexCoords.Add(texCoords[(int.Parse(tokens[3].Split('/')[1]))-1]);
					}
					
					if (tokens[1].Split('/').Length > 2 && tokens[1].Split('/')[2].Length > 0)
					{
						cNormals.Add(normals[(int.Parse(tokens[1].Split('/')[2]))-1]);
						cNormals.Add(normals[(int.Parse(tokens[2].Split('/')[2]))-1]);
						cNormals.Add(normals[(int.Parse(tokens[3].Split('/')[2]))-1]);
					}
				}
				else if (line[0] == 'o')   // Object "header"
				{
					// Save the last object, if there is one
					if (cVertices != null)
					{
						LoadedGoemetry g = new LoadedGoemetry();
						g.vertices = new float[cVertices.Count*3];
						g.texels   = new float[cTexCoords.Count*2];
						g.normals  = new float[cNormals.Count*3];
						for (int i=0; i<cVertices.Count; i++)
						{
							g.vertices[(i*3)+0] = ((float[])cVertices[i])[0];
							g.vertices[(i*3)+1] = ((float[])cVertices[i])[1];
							g.vertices[(i*3)+2] = ((float[])cVertices[i])[2];

							if (cTexCoords.Count > 0)
							{
								g.texels[(i*2)+0] = ((float[])cTexCoords[i])[0];
								g.texels[(i*2)+1] = ((float[])cTexCoords[i])[1];
							}
							
							if (cNormals.Count > 0)
							{
								g.normals[(i*3)+0] = ((float[])cNormals[i])[0];
								g.normals[(i*3)+1] = ((float[])cNormals[i])[1];
								g.normals[(i*3)+2] = ((float[])cNormals[i])[2];
							}
						}

						g.nVertices = cVertices.Count;
						objects.Add(LoadedObj.Add(g) + LOADED);
					}

					cVertices  = new ArrayList();
					cTexCoords = new ArrayList();
					cNormals   = new ArrayList();
				}
			}

			// Save the last object, if there is one
			if (cVertices != null)
			{
				LoadedGoemetry g = new LoadedGoemetry();
				g.vertices = new float[cVertices.Count*3];
				g.texels   = new float[cTexCoords.Count*2];
				g.normals  = new float[cNormals.Count*3];
				for (int i=0; i<cVertices.Count; i++)
				{
					g.vertices[(i*3)+0] = ((float[])cVertices[i])[0];
					g.vertices[(i*3)+1] = ((float[])cVertices[i])[1];
					g.vertices[(i*3)+2] = ((float[])cVertices[i])[2];

					if (cTexCoords.Count > 0)
					{
						g.texels[(i*2)+0] = ((float[])cTexCoords[i])[0];
						g.texels[(i*2)+1] = ((float[])cTexCoords[i])[1];
					}

					if (cNormals.Count > 0)
					{
						g.normals[(i*3)+0] = ((float[])cNormals[i])[0];
						g.normals[(i*3)+1] = ((float[])cNormals[i])[1];
						g.normals[(i*3)+2] = ((float[])cNormals[i])[2];
					}

				}

				g.nVertices = cVertices.Count;
				objects.Add(LoadedObj.Add(g) + LOADED);
			}

			return objects.Count;
		}
		#endregion

	}
}
