using System;
using JPGL.Math;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// Summary description for InteractiveMeshGeometry.
	/// </summary>
	public class InteractiveMeshGeometry : IndexedGeometry
	{
		//-----------------------------------//
		protected int   m_NumControlVertices;
		protected int   m_NumFaces;
		protected int   m_NumVertices;
		protected float m_VertexDistance;
		//-----------------------------------//

		#region Public Property Accessors
		//-----------------------------------//
		public int NumControlVertices
		{
			get { return m_NumControlVertices; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public int NumFaces
		{
			get { return m_NumFaces; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public int NumVertices
		{
			get { return m_NumVertices; }
		}
		//-----------------------------------//
		#endregion


		//-------------------------------------------------------------------------//
		protected InteractiveMeshGeometry() : base()
		{
			m_Indexed = true;
		}

		//-------------------------------------------------------------------------//
		public InteractiveMeshGeometry(int n) : this()
		{
			SetSize(n);
			ComputeSurface();
		}

		//-------------------------------------------------------------------------//
		public void ResetSurface(int size)
		{
			SetSize(size);
			ComputeSurface();
			ComputeNormals();
		}

		//-------------------------------------------------------------------------//
		protected void SetSize(int size)
		{
			m_NumControlVertices = size;
			m_NumVertices = m_NumControlVertices * m_NumControlVertices;
			m_NumFaces = (m_NumControlVertices-1) * (m_NumControlVertices-1) * 2;
			m_VertexDistance = 2.0f/((float)m_NumControlVertices-1.0f);
		}

		//-------------------------------------------------------------------------//
		// Build a flat surface on the x-z plane with n by n vertices and side 
		// length of 2 and centered at the y axis.
		protected virtual void ComputeSurface()
		{
			int nthRow = 0, nthColumn = 0;

			m_Vertices = new VertexFormat[m_NumVertices];

			for (int i=0; i<m_NumVertices; i++)
			{
				m_Vertices[i].x = -1.0f + (nthColumn * m_VertexDistance);
				m_Vertices[i].y = 0.0f;
				m_Vertices[i].z =  1.0f - (nthRow * m_VertexDistance);
				nthColumn++;
				if (nthColumn>=m_NumControlVertices) 
				{
					nthRow++;
					nthColumn = 0;
				}

				m_Vertices[i].nX = 0.0f;	// ComputeNormals will do this right.
				m_Vertices[i].nY = 1.0f;
				m_Vertices[i].nZ = 0.0f;

				m_Vertices[i].u = (m_Vertices[i].x / 2.0f) + 0.5f;
				m_Vertices[i].v = (m_Vertices[i].z / 2.0f) + 0.5f;
			}

			//
			// now fill in the strips of triangle (face) index
			// 
			// each strip has two rows of triangle
			//
			// Here are the observations:
			//
			// on nth strip: 
			//		lower row has index: 
			//				nth*m_NumControlVertices, (nth*m_NumControlVertices)+1, (nth+1)*m_NumControlVertices
			//
			//      upper row has inex:
			//				(nth+1)*m_NumControlVertices, (nth*m_NumControlVertices)+1, ((nth+1)*m_NumControlVertices)+1
			//
			// And each row have m_NumControlVertices-1 number of triangles ...
			//
			int indexCount=0;

			m_Indices = new uint[m_NumControlVertices*m_NumControlVertices*6];

			for (int strip = 0; strip < m_NumControlVertices-1; strip++) 
			{
				int lowV0 = strip*m_NumControlVertices;
				int lowV1 = lowV0 + 1;
				int lowV2 = (strip+1)*m_NumControlVertices;

				int hiV0 = lowV0 + m_NumControlVertices;
				int hiV1 = lowV1;
				int hiV2 = hiV0 + 1;

				for (int rowTri=0; rowTri < m_NumControlVertices-1; rowTri++) 
				{
					//
					// lower row ...
					m_Indices[indexCount++] = (uint)lowV0++;
					m_Indices[indexCount++] = (uint)lowV1++;
					m_Indices[indexCount++] = (uint)lowV2++;

					//
					// upper row
					m_Indices[indexCount++] = (uint)hiV0++;
					m_Indices[indexCount++] = (uint)hiV1++;
					m_Indices[indexCount++] = (uint)hiV2++;
				}
			}
		}

		//-------------------------------------------------------------------------//
		// Compute and save the normals for the surface
		public void ComputeNormals()
		{
			Vector3[] normals = new Vector3[m_NumFaces];

			// First we will walk through the surface and compute the normals
			// for each triangle - storing what we get in the normals array
			for (int i=0; i<m_NumFaces; i++) 
			{
				Vector3 v1, v2, v3;
				v1 = new Vector3(m_Vertices[(m_Indices[(i*3)+0])].x, m_Vertices[(m_Indices[(i*3)+0])].y, m_Vertices[(m_Indices[(i*3)+0])].z);
				v2 = new Vector3(m_Vertices[(m_Indices[(i*3)+1])].x, m_Vertices[(m_Indices[(i*3)+1])].y, m_Vertices[(m_Indices[(i*3)+1])].z);
				v3 = new Vector3(m_Vertices[(m_Indices[(i*3)+2])].x, m_Vertices[(m_Indices[(i*3)+2])].y, m_Vertices[(m_Indices[(i*3)+2])].z);

				Vector3 side1  = v2 - v1;
				Vector3 side2  = v3 - v1;
				try 
				{
					// We may be trying to cross two parallel lines
					// If so, just make something up :o
					normals[i] = side1.CrossProduct(side2).Normalize();		
				} catch (System.DivideByZeroException) {
					normals[i] = side1;
				}
			}

			// Now that we have normals for each triangle, we can
			// walk over each vertex in m_Vertices, and set its 
			// normal as the average of the normals of all the
			// triangles using that vertex.

			// Shown below is how the bordering triangles are 
			// referenced for the vertex in the center. Some 
			// special edge cases may have certain bordering 
			// triangles missing.
			//
			//     .---.---.   
			//     | /1|2/3|
			//     .---.---.
			//     |4/5|6/ |
			//     .---.---.
			//
			// Vertex indexing is in form of {j, i}
			for (int i=0; i<m_NumControlVertices; i++)     // Column index
			{
				for (int j=0; j<m_NumControlVertices; j++) // Row index
				{
					int bCount = 0;                    // Number of bordering triangles
					Vector3 sum = new Vector3(0, 0, 0);// Sum of bordering triangle normals

					// Triangle 1 - Above and to the left.
					// Not available at {0,-} or {-,0}
					if (j != 0 && i != 0)
					{
						sum += normals[((j-1)*2*(m_NumControlVertices-1) + (i-1)*2 + 1)];
						bCount++;
					}
		
					// Triangle 2 - Above and to the right - upper.
					// Not available at {0,-} or {-,m_NumControlVertices-1}
					if (j != 0 && i != m_NumControlVertices-1)
					{
						sum += normals[((j-1)*2*(m_NumControlVertices-1) + (i-1)*2 + 2)];
						bCount++;
					}

					// Triangle 3 - Above and to the right - lower.
					// Not available at {0,-} or {-, m_NumControlVertices-1}
					if (j != 0 && i != m_NumControlVertices-1)
					{
						sum += normals[((j-1)*2*(m_NumControlVertices-1) + (i-1)*2 + 3)];
						bCount++;
					}

					// Triangle 4 - Below and to the left - upper.
					// Not available at {-,0} or {m_NumControlVertices-1,-}
					if (i != 0 && j != m_NumControlVertices-1)
					{
						sum += normals[(j*2*(m_NumControlVertices-1) + (i-1)*2 + 0)];
						bCount++;
					}

					// Triangle 5 - Below and to the left - lower.
					// Not available at {-,0} or {m_NumControlVertices-1,-}
					if (i != 0 && j != m_NumControlVertices-1)
					{
						sum += normals[(j*2*(m_NumControlVertices-1) + (i-1)*2 + 1)];
						bCount++;
					}
			
					// Triangle 6 - Below and to the right.
					// Not available at {m_NumControlVertices-1,-} 
					// or {-,m_NumControlVertices-1}
					if (i != m_NumControlVertices-1 && j != m_NumControlVertices-1)
					{
						sum += normals[(j*2*(m_NumControlVertices-1) + (i-1)*2 + 2)];
						bCount++;
					}


					// Now, assuming we got some bordering triangles (we better have
					// at least one!), average out the normal components and save into
					// the vertex. Note that the averaged normal is not necessarily
					// normalized. We will let the gfx API take care of that 
					// (eg: glEnable(GL_NORMALIZE)).
					if (bCount > 0)
					{
						m_Vertices[IndexFromRowColumn(j, i)].nX = sum.X / bCount;
						m_Vertices[IndexFromRowColumn(j, i)].nY = sum.Y / bCount;
						m_Vertices[IndexFromRowColumn(j, i)].nZ = sum.Z / bCount;
					}
				}
			}

		}

		//-------------------------------------------------------------------------//
		// Get the vertex sequential index from the row column index
		public int IndexFromRowColumn(int r, int c)
		{
			return r * m_NumControlVertices + c;
		}

	}
}
