using System;
using JPGL.Math;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// Summary description for LinearSweepGeometry.
	/// </summary>
	public class LoftedGeometry : InteractiveMeshGeometry
	{
		//-----------------------------------//
		private Vector3[] m_StartLine;
		private Vector3[] m_StopLine;
		private int       m_Size;
		//-----------------------------------//

		#region Public Property Accessors
		//-----------------------------------//
		public Vector3[] StartLine
		{
			get { return m_StartLine;  }
			set { m_StartLine = value; }
		}
		public Vector3[] StopLine
		{
			get { return m_StopLine;  }
			set { m_StopLine = value; }
		}
		public int Size
		{
			get { return m_Size;  }
			set { m_Size = value; }
		}
		//-----------------------------------//
		#endregion

		//-------------------------------------------------------------------------//
		public LoftedGeometry(Vector3[] line1, Vector3[] line2, int size) : base()
		{
			m_StartLine = line1;
			m_StopLine  = line2;
			m_Size      = size;

			UpdateSurface();
		}

		//-------------------------------------------------------------------------//
		public void UpdateSurface()
		{
			ResetSurface(m_Size);
		}

		//-------------------------------------------------------------------------//
		// What we want to do here is start at the first line, and linearly 
		// interpolate our way to the second line.
		protected override void ComputeSurface()
		{
			if (m_StartLine == null) return;
			if (m_StopLine == null)  return;

			m_Vertices = new VertexFormat[m_NumVertices];

			float uvdist = 1.0f / (float)(m_NumControlVertices-1);

			// To to the linear sweep, we will walk along the first line
			// and for each point, we will walk from that point to the
			// corresponding point on the second line.			

			// Put line1 in the array as the start
			for (int i=0; i<m_NumControlVertices; i++)
			{
				m_Vertices[i].x = m_StartLine[i].X;
				m_Vertices[i].y = m_StartLine[i].Y;
				m_Vertices[i].z = m_StartLine[i].Z;

				m_Vertices[i].u = uvdist * i;
				m_Vertices[i].v = 0.0f;
			}

			// Put line2 in the array as the end
			for (int i=0; i<m_NumControlVertices; i++)
			{
				m_Vertices[m_NumVertices-m_NumControlVertices+i].x = m_StopLine[i].X;
				m_Vertices[m_NumVertices-m_NumControlVertices+i].y = m_StopLine[i].Y;
				m_Vertices[m_NumVertices-m_NumControlVertices+i].z = m_StopLine[i].Z;

				m_Vertices[i].u = uvdist * i;
				m_Vertices[i].v = 1.0f;
			}


			// Walk from line1 to line2
			for (int col = 0; col<m_NumControlVertices; col++)
			{
				// Get the vector of travel
				Vector3 delta = (m_StopLine[col] - m_StartLine[col]) * (1.0f/(m_NumControlVertices-1));

				for (int row = 1; row<m_NumControlVertices-1; row++)
				{
					// Row i's position = Row i-1's position + delta
					int i   = IndexFromRowColumn(row,   col);
					int rm1 = IndexFromRowColumn(row-1, col);

					m_Vertices[i].x = m_Vertices[rm1].x + delta.X;
					m_Vertices[i].y = m_Vertices[rm1].y + delta.Y;
					m_Vertices[i].z = m_Vertices[rm1].z + delta.Z;

					m_Vertices[i].u = uvdist * col;
					m_Vertices[i].v = 1.0f - (uvdist * row);

				}
			}


			//
			// now fill in the strips of triangle (face) index
			// 
			// each strip has two rows of triangle
			//
			// Here are the observations:
			//
			// on nth strip: 
			//		lower row has index: 
			//				nth*m_NumControlVertices, (nth*m_NumControlVertices)+1, (nth+1)*m_NumControlVertices
			//
			//      upper row has inex:
			//				(nth+1)*m_NumControlVertices, (nth*m_NumControlVertices)+1, ((nth+1)*m_NumControlVertices)+1
			//
			// And each row have m_NumControlVertices-1 number of triangles ...
			//
			int indexCount=0;

			m_Indices = new uint[m_NumControlVertices*m_NumControlVertices*6];

			for (int strip = 0; strip < m_NumControlVertices-1; strip++) 
			{
				int lowV0 = strip*m_NumControlVertices;
				int lowV1 = lowV0 + 1;
				int lowV2 = (strip+1)*m_NumControlVertices;

				int hiV0 = lowV0 + m_NumControlVertices;
				int hiV1 = lowV1;
				int hiV2 = hiV0 + 1;

				for (int rowTri=0; rowTri < m_NumControlVertices-1; rowTri++) 
				{
					//
					// lower row ...
					m_Indices[indexCount++] = (uint)lowV0++;
					m_Indices[indexCount++] = (uint)lowV1++;
					m_Indices[indexCount++] = (uint)lowV2++;

					//
					// upper row
					m_Indices[indexCount++] = (uint)hiV0++;
					m_Indices[indexCount++] = (uint)hiV1++;
					m_Indices[indexCount++] = (uint)hiV2++;
				}
			}

		}

	}
}
