using System;
using JPGL.Common.Colliders;
using JPGL.Math;

namespace JPGL.Common.Primitives
{
	/// <summary>
	/// Summary description for InteractiveMesh.
	/// </summary>
	public class InteractiveMesh : FillablePrimitive
	{
		//-----------------------------------//
		protected InteractiveMeshGeometry m_Geometry;
		private int                       m_SelectedVertex;
		//-----------------------------------//

		//-----------------------------------//
		private bool     m_DrawNormal;
		private bool     m_DrawControlGrid;
		//-----------------------------------//

		#region Public Properties Overriden From DrawablePrimitive
		//-----------------------------------// 
		public override bool Defined
		{
			get
			{
				return (m_Geometry != null);
			}
		}
		//-----------------------------------//
		#endregion

		#region Public Property Accessors
		//-----------------------------------//
		public Vector3 this[int r, int c]
		{
			get
			{
				int i = RowColumnToIndex(r, c);
				return new Vector3(
					m_Geometry.VertexData[i].x,
					m_Geometry.VertexData[i].y,
					m_Geometry.VertexData[i].z);
			}
			set
			{
				int i = RowColumnToIndex(r, c);
				m_Geometry.VertexData[i].x = value.X;
				m_Geometry.VertexData[i].y = value.Y;
				m_Geometry.VertexData[i].z = value.Z;

				// Recompute normals!
				m_Geometry.ComputeNormals();
			}
		}
		//-----------------------------------//

		//-----------------------------------//
		public Vector3 this[int i]
		{
			get
			{
				return new Vector3(
					m_Geometry.VertexData[i].x,
					m_Geometry.VertexData[i].y,
					m_Geometry.VertexData[i].z);
			}
			set
			{
				m_Geometry.VertexData[i].x = value.X;
				m_Geometry.VertexData[i].y = value.Y;
				m_Geometry.VertexData[i].z = value.Z;

				// Recompute normals!
				m_Geometry.ComputeNormals();
			}
		}
		//-----------------------------------//

		//-----------------------------------//
		public bool DrawNormal
		{
			get { return m_DrawNormal;  }
			set { m_DrawNormal = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public bool DrawControlGrid
		{
			get { return m_DrawControlGrid;  }
			set { m_DrawControlGrid = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public int Size
		{
			get { return m_Geometry.NumControlVertices; }
		}
		//-----------------------------------//

		//-----------------------------------//
		public int SelectedVertex
		{
			get { return m_SelectedVertex;  }
			set { m_SelectedVertex = value; }
		}
		//-----------------------------------//
		#endregion

		//-------------------------------------------------------------------------//
		// Protected constructor so inheriting classes can call constructor without
		// the geometry object being created here.
		protected InteractiveMesh() : base()
		{
			m_SelectedVertex  = 0;
			m_DrawNormal      = false;
			m_DrawControlGrid = false;
		}

		//-------------------------------------------------------------------------//
		public InteractiveMesh(int size) : this()
		{
			m_Geometry = new InteractiveMeshGeometry(size);

			GenColorArray();
		}

		//-------------------------------------------------------------------------//
		public int[] IndexToRowColumn(int index)
		{
			return new int[2] { 
								  m_SelectedVertex / m_Geometry.NumControlVertices,
								  m_SelectedVertex % m_Geometry.NumControlVertices 
							  };
		}

		//-------------------------------------------------------------------------//
		public int RowColumnToIndex(int r, int c)
		{
			return m_Geometry.IndexFromRowColumn(r, c);
		}

		//-------------------------------------------------------------------------//
		// Randomly move each vertex along the direction of its normal some amount.
		public void RandomDeform(float deltamax)
		{
			Random r = new Random();
			for (int i=0; i<m_Geometry.NumVertices; i++)
			{
				m_Geometry.VertexData[i].x += ((float)r.NextDouble()) * deltamax * m_Geometry.VertexData[i].nX;
				m_Geometry.VertexData[i].y += ((float)r.NextDouble()) * deltamax * m_Geometry.VertexData[i].nY;
				m_Geometry.VertexData[i].z += ((float)r.NextDouble()) * deltamax * m_Geometry.VertexData[i].nZ;
			}

			m_Geometry.ComputeNormals();
		}

		//-------------------------------------------------------------------------//
		// Recompute surface and normals.
		public void Reset()
		{
			m_Geometry.ResetSurface(m_Geometry.NumControlVertices);

			GenColorArray();
		}

		//-------------------------------------------------------------------------//
		public override void DrawExtra(PrimitiveRenderer r)
		{
			if (m_DrawControlGrid) DrawControlPoints(r);
			if (m_DrawNormal)      DrawNormals(r);
		}

		//-------------------------------------------------------------------------//
		// Collider is a single ColliderCircle with our location and radius.
		public override ColliderList GetColliders()
		{
			return new ColliderList();
		}

		//-------------------------------------------------------------------------//
		// Geometry for a unit circle centered at (0,0).
		public override Geometry GetGeometry()
		{
			return m_Geometry;
		}

		//-------------------------------------------------------------------------//
		public override Transform GetTransform()
		{
			return new Transform(
				new Vector3(1, 1, 1), 
				new Vector3(0, 0, 0), 
				(m_Points[0] != null ? m_Points[0] : new Vector3(0, 0, 0)));
		}

		//-------------------------------------------------------------------------//
		private void DrawControlPoints(PrimitiveRenderer r)
		{
			Point p = new Point();
			p.MoveTo(new Vector3(0, 0, 0));
			p.PointSize = 4;
			p.Lighted = false;
			p.Color = System.Drawing.Color.FromArgb(255, 204, 204, 204);
 
			// Draw a point at each vertex
			for (int i=0; i<m_Geometry.NumVertices; i++)
			{
				p.Point.X = m_Geometry.VertexData[i].x; 
				p.Point.Y = m_Geometry.VertexData[i].y; 
				p.Point.Z = m_Geometry.VertexData[i].z;
				r.DrawPrimitive(p);
			}

			// Highlight selected vertex
			p.Color = System.Drawing.Color.FromArgb(255, 255, 0, 0);
			p.PointSize = 8;
			p.Point.X = m_Geometry.VertexData[m_SelectedVertex].x;
			p.Point.Y = m_Geometry.VertexData[m_SelectedVertex].y;
			p.Point.Z = m_Geometry.VertexData[m_SelectedVertex].z;
			r.DrawPrimitive(p);
		}

		//-------------------------------------------------------------------------//
		private void DrawNormals(PrimitiveRenderer r)
		{
			Line3d l = new Line3d();
			l.Lighted = false;
			l.Color = System.Drawing.Color.FromArgb(255, 255, 255, 255);
			l.Shaded = false;

			for (int i=0; i<m_Geometry.NumVertices; i++)
			{
				l.SetEnds(
					m_Geometry.VertexData[i].x, m_Geometry.VertexData[i].y, m_Geometry.VertexData[i].z, 
					m_Geometry.VertexData[i].x+(m_Geometry.VertexData[i].nX * 0.25f), 
					m_Geometry.VertexData[i].y+(m_Geometry.VertexData[i].nY * 0.25f),
					m_Geometry.VertexData[i].z+(m_Geometry.VertexData[i].nZ * 0.25f));
				r.DrawPrimitive(l);
			}

		}
		//-------------------------------------------------------------------------//
		protected override void GenColorArray()
		{
			if (m_Geometry == null) return;

			for (int i=0; i<m_Geometry.NumVertices; i++)
			{			
				m_Geometry.VertexData[i].r = this.Color.R / 255.0f;
				m_Geometry.VertexData[i].g = this.Color.G / 255.0f;
				m_Geometry.VertexData[i].b = this.Color.B / 255.0f;
				m_Geometry.VertexData[i].a = this.Color.A / 255.0f;
			}
		}

	}
}
