using System;
using JPGL.Math;

namespace JPGL.Common.Primitives
{
	/// <summary>
	/// Summary description for ExtrudePrimitive.
	/// </summary>
	public class LoftPrimitive : InteractiveMesh
	{
		#region Public Property Accessors
		//-----------------------------------//
		public Vector3[] Line1
		{
			get { return ((LoftedGeometry)m_Geometry).StartLine; }
		}
		public Vector3[] Line2
		{
			get { return ((LoftedGeometry)m_Geometry).StopLine; }
		}
		//-----------------------------------//
		#endregion

		public enum PRESETDEST { LINE1, LINE2, BOTH };

		//-------------------------------------------------------------------------//
		public LoftPrimitive(Vector3[] line1, Vector3[] line2, int size) : base()
		{
			m_Geometry = new LoftedGeometry(line1, line2, size);

			GenColorArray();
		}

		//-------------------------------------------------------------------------//
		// Set the given line to be a circle, and 
		public void BuildSurfaceFromCircle(PRESETDEST line)
		{
			int size = m_Geometry.NumControlVertices;
			Vector3[] line1 = ((LoftedGeometry)m_Geometry).StartLine;
			Vector3[] line2 = ((LoftedGeometry)m_Geometry).StopLine;

			float delta = (2.0f * (float)System.Math.PI) / (float)(size-1);
			for (int i=0; i<size; i++)
			{
				float sin = (float)(System.Math.Sin((i)*delta));
				float cos = (float)(System.Math.Cos((i)*delta));
				if (line == PRESETDEST.LINE1 || line == PRESETDEST.BOTH)
				{
					line1[i] = new Vector3(sin, -1, cos);
				}
				if (line == PRESETDEST.LINE2 || line == PRESETDEST.BOTH)
				{
					line2[i] = new Vector3(sin, 1, cos);
				}
			}
			
			BuildSurface(line1, line2, size);
		}

		//-------------------------------------------------------------------------//
		// Set the given line to be a compressed sin curve - squishing -PI->PI 
		// into the range of -1->1 
		public void BuildSurfaceFromSin(PRESETDEST line)
		{
			int size = m_Geometry.NumControlVertices;
			Vector3[] line1 = ((LoftedGeometry)m_Geometry).StartLine;
			Vector3[] line2 = ((LoftedGeometry)m_Geometry).StopLine;

			float deltax = 2.0f  / (float)(size-1);
			float fPI    = (float)System.Math.PI;
			for (int i=0; i<size; i++)
			{
				float x = (i * deltax) - 1.0f;
				float sxpi = (float)(System.Math.Sin(x * fPI));
				if (line == PRESETDEST.LINE1 || line == PRESETDEST.BOTH)
				{
					line1[i] = new Vector3(x, -1, sxpi);
				}
				if (line == PRESETDEST.LINE2 || line == PRESETDEST.BOTH)
				{
					line2[i] = new Vector3(x, 1, sxpi);
				}
			}
			
			BuildSurface(line1, line2, size);
		}

		//-------------------------------------------------------------------------//
		// Set the given line to be a compressed cosine curve - squishing -PI->PI 
		// into the range of -1->1 
		public void BuildSurfaceFromCos(PRESETDEST line)
		{
			int size = m_Geometry.NumControlVertices;
			Vector3[] line1 = ((LoftedGeometry)m_Geometry).StartLine;
			Vector3[] line2 = ((LoftedGeometry)m_Geometry).StopLine;

			float deltax = 2.0f  / (float)(size-1);
			float fPI    = (float)System.Math.PI;
			for (int i=0; i<size; i++)
			{
				float x = (i * deltax) - 1.0f;
				float sxpi = (float)(System.Math.Cos(x * fPI));
				if (line == PRESETDEST.LINE1 || line == PRESETDEST.BOTH)
				{
					line1[i] = new Vector3(x, -1, sxpi);
				}
				if (line == PRESETDEST.LINE2 || line == PRESETDEST.BOTH)
				{
					line2[i] = new Vector3(x, 1, sxpi);
				}
			}
			
			BuildSurface(line1, line2, size);
		}

		//-------------------------------------------------------------------------//
		// Build surface with line1 from (-1, 1, 0) to (1, 1, 0) and line2
		// from (-1, -1, 0) to (1, -1, 0) with size control points in between
		public void BuildSurface(int size)
		{
			if (size == m_Geometry.NumControlVertices) return;

			Vector3[] line1 = new Vector3[size];
			Vector3[] line2 = new Vector3[size];

			float delta = 2.0f / (size-1);
			for (int i=0; i<size; i++)
			{
				line1[i] = new Vector3(-1 + delta*i, -1, 0);
				line2[i] = new Vector3(-1 + delta*i, 1, 0);
			}

			BuildSurface(line1, line2, size);
		}

		//-------------------------------------------------------------------------//
		public void BuildSurface(Vector3[] line1, Vector3[] line2, int size)
		{
			((LoftedGeometry)m_Geometry).StartLine = line1;
			((LoftedGeometry)m_Geometry).StopLine  = line2;
			((LoftedGeometry)m_Geometry).Size      = size;

			((LoftedGeometry)m_Geometry).UpdateSurface();

			GenColorArray();
		}
	}
}
