using System;
using System.Collections;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// A Hardware Accelerated Point Sprite based particle system.
	/// It is assumed that the system will simply not be rendered
	/// if point sprites are not supported - but we will leave that
	/// decision up to the graphics API library.
	/// </summary>
	public class ParticleSystem
	{
		//-----------------------------------//
		private static Texture PointSpriteTexture = null;
		private static Random m_RandGen = new Random();
		//-----------------------------------//

		//-----------------------------------//
		private Particle[] m_Particles;
		private int        m_ParticleCount;
		private float      m_OriginX, 
			               m_OriginY, 
			               m_OriginZ;
		private Transform  m_Transform;
		private float      m_CountDown;
		//-----------------------------------//

		//-----------------------------------//
		public struct Particle
		{
			public float x,y,z;
			public float r,g,b;
			public float vx,vy,vz;
			public float ttl;
			public bool  alive;
		}
		//-----------------------------------//

		#region Public Accessor Properties
		//-----------------------------------//
		public float OriginX
		{
			get { return m_OriginX;  }
			set { m_OriginX = value; }
		}
		//-----------------------------------//
		public float OriginY
		{
			get { return m_OriginY;  }
			set { m_OriginY = value; }
		}
		//-----------------------------------//
		public float OriginZ
		{
			get { return m_OriginZ;  }
			set { m_OriginZ = value; }
		}
		//-----------------------------------//
		public Transform T
		{
			get { return m_Transform;  }
			set { m_Transform = value; }
		}
		//-----------------------------------//
		public static Texture Texture
		{
			get { return PointSpriteTexture;  }
			set { PointSpriteTexture = value; }
		}
		//-----------------------------------//
		public int ParticleCount
		{
			get { return m_ParticleCount; }
		}
		//-----------------------------------//
		public Particle[] Particles
		{
			get { return m_Particles; }
		}
		//-----------------------------------// 
		#endregion


		//-------------------------------------------------------------------------//
		public ParticleSystem(int particlecount)
		{
			m_ParticleCount = particlecount;
			m_Particles  = new Particle[particlecount];

			m_Transform = new Transform();

			SetupSystem(0, 0, 0);
		}

		//-------------------------------------------------------------------------//
		public void SetupSystem(float x, float y, float z)
		{
			m_CountDown = 2;
			m_OriginX = x; m_OriginY = y; m_OriginZ = z;

			// Fireworks
			for(int i = 0; i < m_ParticleCount; i++)
			{
				m_Particles[i].x = m_OriginX; 
				m_Particles[i].y = m_OriginY;
				m_Particles[i].z = m_OriginZ;

				m_Particles[i].vz = (float)(m_RandGen.NextDouble()*2.0-1.0);
				double r = System.Math.Sqrt(1.0 - (m_Particles[i].vz * m_Particles[i].vz));
				double t = (m_RandGen.NextDouble()*2.0*System.Math.PI) - System.Math.PI;
				m_Particles[i].vx = (float)(System.Math.Cos(t) * r);
				m_Particles[i].vy = (float)(System.Math.Sin(t) * r);

				m_Particles[i].r = (float)m_RandGen.NextDouble();
				m_Particles[i].g = (float)m_RandGen.NextDouble();
				m_Particles[i].b = (float)m_RandGen.NextDouble();

				m_Particles[i].ttl   = 2.5f;
				m_Particles[i].alive = false;
			}

			m_Particles[0].alive = true;
		}

		//-------------------------------------------------------------------------//
		public void UpdateSystem(float timestep)
		{
			// Firing the firework in the air
			if (m_CountDown > 0.0001f)
			{
				
				m_Particles[0].x += (float)(m_RandGen.NextDouble()+1.0) * (2.0f-m_CountDown) * timestep;
				m_Particles[0].y += 4 * timestep;
				m_Particles[0].z += (float)(m_RandGen.NextDouble()+1.0) * (2.0f-m_CountDown) * timestep;

				m_CountDown -= timestep;
				if (!(m_CountDown > 0.0001f))
				{
					for(int i = 1; i < m_ParticleCount; i++)
					{
						m_Particles[i].x = m_Particles[0].x; 
						m_Particles[i].y = m_Particles[0].y;
						m_Particles[i].z = m_Particles[0].z;
						m_Particles[i].alive = true;
					}
					m_Particles[0].x = m_Particles[1].x; 
					m_Particles[0].y = m_Particles[1].y;
					m_Particles[0].z = m_Particles[1].z;
				}
				return;
			}

			// Fireworks
			bool anyleft = false;
			for (int i = 0; i < m_ParticleCount; i++)
			{
				if (!m_Particles[i].alive) continue;

				m_Particles[i].x += (m_Particles[i].ttl * m_Particles[i].vx) * 2.0f * timestep;
				m_Particles[i].y += (m_Particles[i].ttl * m_Particles[i].vy) * 2.0f * timestep;
				m_Particles[i].z += (m_Particles[i].ttl * m_Particles[i].vz) * 2.0f * timestep;

				m_Particles[i].ttl -= timestep;
				if (m_Particles[i].ttl < 0)
				{
					m_Particles[i].alive = false;
				}
				else
				{
					anyleft = true;
				}
			}

			// Restart the system if no particles left
			if (!anyleft)
			{
				SetupSystem(m_OriginX, m_OriginY, m_OriginZ);
			}
		}
	}
}
