using System;
using JPGL.Common.Colliders;
using JPGL.Math;

namespace JPGL.Common.Primitives
{
	/// <summary>
	/// Primitive is a public interface and base implementation of a 
	/// Primitive. The public interface provides the following func-
	/// tions:
	///  MoveTo - Move the Primitive to the given location.
	///  SizeTo - Size the Primitive to the given location.
	///  CollidesWith - Check for and do collisions.
	/// And the following properties:
	///  Point - Get the first point in the point list.
	///  Velocity - Get/Set the velocity.
	///  Acceleration - Get/Set the acceleration.
	///  ShowColliders - Whether the colliders will be drawn.
	///  
	/// Implementing classes must define the following functions:
	///  GetTransform - Return a Transform appropriate to the Primitive.
	///  PointInPrimitive - Is a given point in the Primitive?
	///  GetColliders - Return the colliders of the Primitive.
	/// And the following properties:
	///  Defined - True if all points the Primitive needs are set.
	///  
	/// Implementing classes may implement the following (default 
	/// implementations are provided):
	///  SetFirstPoint - Set the first point in the point list
	///  AddPoint      - Add a point to the point list.
	///  
	/// </summary>
	public abstract class Primitive
	{
		//-----------------------------------// Implementing classes must provide
		public abstract bool Defined         // implementation of this attribute
		{                                    // such that a true Defined primitive
			get;                             // can be manipulated (read: has all
		}                                    // the points it needs) without problems.
		//-----------------------------------//

		//-----------------------------------//
		protected Vector3List m_Points;      // List of points
		public Vector3        Point          //
		{
			get { return m_Points[0]; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected bool       m_ShowColliders;// Any Primitive that has colliders
		public virtual bool  ShowColliders   // may be able to draw their  
		{                                    // colliders.
			get { return m_ShowColliders;  }
			set { m_ShowColliders = value; }
		}	
		//-----------------------------------//


		//-------------------------------------------------------------------------//
		public Primitive()
		{
			m_Points        = new Vector3List();
			m_ShowColliders = false;
		}

		//-------------------------------------------------------------------------//
		// Copy constructor
		public Primitive(Primitive old)
		{
			m_Points        = new Vector3List(old.m_Points);
			m_ShowColliders = old.ShowColliders;
		}

		//-------------------------------------------------------------------------//
		// Implementing classes may override setfirstpoint to change moveto behavior.
		public void MoveTo(Vector3 p)
		{
			SetFirstPoint(p);
		}

		//-------------------------------------------------------------------------//
		public void MoveBy(Vector3 d)
		{
			SetFirstPoint(m_Points[0] + d);
		}

		//-------------------------------------------------------------------------//
		// Implementing classes may override addpoint to change sizeto behavior.
		public void SizeTo(Vector3 p)
		{
			AddPoint(p);
		}

		//-------------------------------------------------------------------------//
		// Return the transformation associated with this primitive.
		public abstract Transform GetTransform();

		//-------------------------------------------------------------------------//
		// Implementing classes must return a list of colliders.
		public abstract ColliderList GetColliders();


		//-------------------------------------------------------------------------//
		// By default, setfirstpoint sets the first point...
		protected virtual void SetFirstPoint(Vector3 p) 
		{
			m_Points.Insert(0, p);
		}

		//-------------------------------------------------------------------------//
		// By default, addpoint adds a point...
		protected virtual void AddPoint(Vector3 p)      
		{ 
			m_Points.Add(p);       
		}
	}
}
