using System;
using JPGL.Math;
using JPGL.Common.Lighting;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// PrimitiveRenderer
	/// </summary>
	public abstract class PrimitiveRenderer
	{
		//-------------------------------------------------------------------------//
		// Load the given transformation.
		public void LoadTransform(Transform t)
		{
			ApplyTransformation(t);
		}

		//-------------------------------------------------------------------------//
		// Unload the last transformation.
		public void UnloadTransform()
		{
			UndoLastTransform();
		}

		//-------------------------------------------------------------------------//
		public void AddLight(Light l)
		{
			SetupLight(l);
		}

		//-------------------------------------------------------------------------//
		// Get the DrawablePrimitive's vertices, and draw them.
		public void DrawPrimitive(DrawablePrimitive p)
		{
			if (!p.Defined) return;
			if (!p.Visibility) return;

			// Draw colliders?
			if (p.ShowColliders)
				p.GetColliders().Draw(this);

			// Get and apply the primitive's transform
			LoadTransform(p.GetTransform());

			// Get geometry and draw it
			Geometry g = p.GetGeometry();
			PrepareForPrimitive(g, p);

			if (g.Indexed)
				DrawIndexedGeometry((IndexedGeometry)g);
			else
				DrawGeometry(g, p.ColorArray, p.TextureOn);

			// Let the primitive draw something if it wants to
			p.DrawExtra(this);

			// Unload the Primitive's transformation
			UnloadTransform();
		}

		//-------------------------------------------------------------------------//
		// Protected interface that must be implemented.
		protected abstract void ApplyTransformation(Transform t);
		protected abstract void UndoLastTransform();
		protected abstract void PrepareForPrimitive(Geometry g, DrawablePrimitive p);
		protected abstract void DrawGeometry(Geometry g, float[] c, bool textured);
		protected abstract void DrawIndexedGeometry(IndexedGeometry g);
		protected abstract void SetupLight(Light l);

		//-------------------------------------------------------------------------//
		// Public interface that must be implemented.
		public abstract void RenderParticleSystem(ParticleSystem s);

		//-------------------------------------------------------------------------//
		public static int POINTS         { get { return 0; } }
		public static int LINES          { get { return 1; } }
		public static int LINE_STRIP     { get { return 2; } }
		public static int LINE_LOOP      { get { return 3; } }
		public static int TRIANGLES      { get { return 4; } }
		public static int TRIANGLE_STRIP { get { return 5; } }
		public static int TRIANGLE_FAN   { get { return 6; } }
		public static int QUADS          { get { return 7; } }
		public static int QUAD_STRIP     { get { return 8; } }
		public static int POLYGON        { get { return 9; } }

		//-------------------------------------------------------------------------//
		// Implementing PrimitiveRenderers must define these shading modes.
		public abstract int FLAT   { get; }
		public abstract int SMOOTH { get; }

		//-------------------------------------------------------------------------//
		// Implementing PrimitiveRenderers must define these shading modes.
		public abstract int POINT { get; }
		public abstract int LINE  { get; }
		public abstract int FILL  { get; }
	}
}
