using System;
using JPGL.Common.Colliders;
using JPGL.Math;


namespace JPGL.Common.Primitives
{
	/// <summary>
	/// A Line is defined by 2 points. The first point is the center point on
	/// the line, and the second point is one end of the line. A line is also
	/// Shadable. A Line has 3 circle colliders, one at each endpoint, and a 
	/// third at the center. Each circle collider has a radius equal to 1/5th
	/// the length of the line.
	/// </summary>
	public class Line : ShadablePrimitive
	{
		//-----------------------------------//
		// Adding and subtracting m_Halves to/from the center point
		// on the line will produce the 2 endpoints of the line.
		protected Vector3 m_Halves;
		//-----------------------------------//

		//-----------------------------------// 
		public override bool Defined         // Line is defined when the
		{                                    // midline is defined.
			get
			{
				return (m_Halves != null);
			}
		}
		//-----------------------------------//

		//-------------------------------------------------------------------------//
		public Line() : base()
		{
		}

		//-------------------------------------------------------------------------//
		public Line(Line p) : base(p)
		{
			m_Halves = new Vector3(p.m_Halves);
		}

		//-------------------------------------------------------------------------//
		// When point 2 is added, calculate and save m_Halves segment
		protected override void Point2Added()
		{
			m_Halves = (m_Points[0] - m_Points[1]); 
		}

		//-------------------------------------------------------------------------//
		public override ColliderList GetColliders()
		{
			if (!Defined) return null;

			float radius = m_Halves.Length() * 0.4f;

			ColliderList c = new ColliderList();
			c.AddCollider(new ColliderCircle(m_Points[0], radius));
			c.AddCollider(new ColliderCircle(m_Points[0]+m_Halves, radius));
			c.AddCollider(new ColliderCircle(m_Points[0]-m_Halves, radius));

			return c;
		}

		//-------------------------------------------------------------------------//
		// Geometry is a line on the x axis at origin of length 1 (-0.5,0)->(0.5,0)
		public override Geometry GetGeometry()
		{
			return Geometry.LineGeometry;
		}

		//-------------------------------------------------------------------------//
		public override Transform GetTransform()
		{
			return new Transform(new Vector3(m_Halves.Length()*2.0f, 1.0f, 1.0f), new Vector3(0, 0, 0), m_Points[0]);
		}
	}
}
