using System;
using System.Collections;
using JPGL.Common.Colliders;
using JPGL.Common.Primitives;
using JPGL.Common.SceneNodes;
using JPGL.Common.Viewports;
using JPGL.Math;


namespace JPGL.Common
{
	/// <summary>
	/// SceneNode.
	/// </summary>
	public class SceneNode
	{
		//-----------------------------------//
		protected Transform m_Transform;       // This node's transformation
		//-----------------------------------//
		public Transform T
		{
			get { return m_Transform;  }
			set { m_Transform = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected Vector3 m_ControlPoint;      // Control point of this node
		//-----------------------------------// Set to origin of scene coordinate
		public Vector3 ControlPoint          // system - (0,0) -  by default.
		{
			get { return m_ControlPoint;  }
			set { m_ControlPoint = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected DrawablePrimitive m_Primitive;// Primitive of this node
		//-----------------------------------// 
		public DrawablePrimitive P
		{
			get { return m_Primitive;  }
			set { m_Primitive = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected bool m_DrawFirst;            // Should this node be drawn in front of 
		//-----------------------------------// children? (default=yes/true)
		public bool DrawFirst
		{
			get { return m_DrawFirst;  }
			set { m_DrawFirst = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected bool m_DisplayPivot;         // Should the pivot be shown?
		//-----------------------------------//
		public bool DisplayPivot
		{
			get { return m_DisplayPivot;  }
			set { m_DisplayPivot = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected bool m_DisplayAxis;             // Should the axis be displayed?
		//-----------------------------------//
		public bool DisplayAxis
		{
			get { return m_DisplayAxis;  }
			set { m_DisplayAxis = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected ArrayList m_Children;        // List of child scenenodes
		public ArrayList Children
		{
			get { return m_Children; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected SceneNode m_Parent;          // Parent scene node
		//-----------------------------------//
		public SceneNode Parent
		{
			get { return m_Parent; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected Vector3     m_Velocity;    // Velocity Vector
		public Vector3        Velocity       //
		{
			get { return m_Velocity;  }
			set { m_Velocity = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected Vector3     m_Acceleration;// Acceleration Vector
		public Vector3        Acceleration   //
		{
			get { return m_Acceleration;  }
			set { m_Acceleration = value; }
		}
		//-----------------------------------//

		//-----------------------------------//
		protected string m_NodeName;
		public string NodeName
		{
			get { return m_NodeName;  }
			set { m_NodeName = value; }
		}
		//-----------------------------------//


		//-------------------------------------------------------------------------//
		public SceneNode(DrawablePrimitive p, Transform t)
		{
			m_DrawFirst    = true;
			m_ControlPoint = new Vector3(0, 0, 0);
			m_Primitive    = p;
			m_Transform    = t;
			m_DisplayPivot = false;
			m_DisplayAxis  = false;
			m_Children     = new ArrayList();
			m_Parent       = null;
			m_Velocity     = new Vector3();
			m_Acceleration = new Vector3();
			m_NodeName     = "SceneNode";
		}

		//-------------------------------------------------------------------------//
		public SceneNode(string name, DrawablePrimitive p, Transform t)
		{
			m_DrawFirst    = true;
			m_ControlPoint = new Vector3(0, 0, 0);
			m_Primitive    = p;
			m_Transform    = t;
			m_DisplayPivot = false;
			m_DisplayAxis  = false;
			m_Children     = new ArrayList();
			m_Parent       = null;
			m_Velocity     = new Vector3();
			m_Acceleration = new Vector3();
			m_NodeName     = name;
		}

		//-------------------------------------------------------------------------//
		// 
		public void MoveToOC(float ox, float oy, float oz)
		{
			T.T = new Vector3(ox-ControlPoint.X, oy-ControlPoint.Y, oz-ControlPoint.Z);
		}

		//-------------------------------------------------------------------------//
		// Get the colliders of this scene node's primitive and all the primitives 
		// under this scene node. The viewport given as an argument is used to do
		// conversion from object to world space.
		public virtual ColliderList GetColliders(Viewport v)
		{
			// Get all chilren SN colliders
			ColliderList r = new ColliderList();
			foreach (SceneNode sn in m_Children)
			{
				foreach (Collider c in sn.GetColliders(v))
                    r.AddCollider(c);
			}
			
			// Get our primitive's collider
			if (m_Primitive != null)
			{
				foreach (Collider c in m_Primitive.GetColliders())
				{
					// Take the position to world coordinates
					Vector3 vw; v.ObjectToWorld(this, c.Position, out vw);
					c.Position = vw;
					r.AddCollider(c);
				}
			}

			foreach (Collider c in r)
				c.Size = c.Size * T.S.X;

			return r;
		}

		//-------------------------------------------------------------------------//
		// Draw all the children of this node, then this node itself.
		public virtual void DrawScene(PrimitiveRenderer r)
		{
			// Draw the pivot if necessary
			if (m_DisplayPivot)
			{
				Sphere pivot = new Sphere();
				pivot.MoveTo(T.P);
				pivot.SizeTo(new Vector3(T.P.X+0.075f, T.P.Y, T.P.Z));
				pivot.Filled = false;
				pivot.Lighted = false;
				pivot.Color = System.Drawing.Color.Green;
				r.DrawPrimitive(pivot);
			}

			// Draw the axis if necessary
			if (m_DisplayAxis)
			{
				Arrow xAxis = new Arrow();
				xAxis.T.T = new Vector3(T.P);
				xAxis.T.S = new Vector3(0.25f, 0.5f, 0.25f);
				xAxis.T.RZBy = -90;
				xAxis.Lighted = false;
				xAxis.DrawScene(r);

				Arrow yAxis = new Arrow();
				yAxis.T.T = new Vector3(T.P);
				yAxis.T.S = new Vector3(0.25f, 0.5f, 0.25f);
				yAxis.Color = System.Drawing.Color.Green;
				yAxis.Lighted = false;
				yAxis.DrawScene(r);

				Arrow zAxis = new Arrow();
				zAxis.T.T = new Vector3(T.P);
				zAxis.T.RXBy = 90;
				zAxis.T.S = new Vector3(0.25f, 0.5f, 0.25f);
				zAxis.Color = System.Drawing.Color.Blue;
				zAxis.Lighted = false;
				zAxis.DrawScene(r);
			}

			// Load the transformation
			r.LoadTransform(m_Transform);

			// Draw primitive if it is drawn on top
			if ((m_Primitive != null) && (!m_DrawFirst))
				r.DrawPrimitive(m_Primitive);
			
			// Draw all child scenenodes
			foreach (SceneNode sn in m_Children)
			{
				sn.DrawScene(r);
			}

			// Draw the primitive if it is drawn below
			if ((m_Primitive != null) && (m_DrawFirst))
				r.DrawPrimitive(m_Primitive);

			// Unload the transformation
			r.UnloadTransform();
		}

		//-------------------------------------------------------------------------//
		public void AddChild(SceneNode c)
		{
			m_Children.Add(c);
			c.m_Parent = this;
		}

		//-------------------------------------------------------------------------//
		public void LinkChild(SceneNode c)
		{
			m_Children.Add(c);
		}

		//-------------------------------------------------------------------------//
		public void RemoveChild(SceneNode c)
		{
			m_Children.Remove(c);
		}

		//-------------------------------------------------------------------------//
		// This is complicated, dont use unless you understand, or i get a chance
		// to comment it.
		public SceneNode CloneToChild(SceneNode c)
		{
			if (Parent != null)
			{
				// Make a copy of ourselves with only our transform
				SceneNode copy = new SceneNode(null, new Transform(T));

				// Add child
				copy.AddChild(c);

				return Parent.CloneToChild(copy);
			}
			else
			{
				SceneNode ret = new SceneNode(null, new Transform());
				ret.AddChild(c);
				AddChild(ret);
				return ret;
			}
		}


		//-------------------------------------------------------------------------//
		// Enumeration functions (so we can use foreach)
		public IEnumerator GetEnumerator()
		{
			return new SceneNodeEnumerator(this);
		} 

		/// <summary>
		/// Private Enumerator so we can use foreach on our lists.
		/// </summary>
		private class SceneNodeEnumerator : IEnumerator
		{
			//-----------------------------------//
			private SceneNode m_List;            // Reference to list
			private int m_Position;              // Position used for enumeration
			//-----------------------------------//

			public SceneNodeEnumerator(SceneNode list)
			{
				m_List     = list;
				m_Position = -1;
			}

			//-------------------------------------------------------------------------//
			public bool MoveNext()
			{
				m_Position++;
				if (m_Position >= m_List.m_Children.Count)
				{
					return false;
				}

				return true;
			} 

			//-------------------------------------------------------------------------//
			public void Reset()
			{
				m_Position = -1;
			} 

			//-------------------------------------------------------------------------//
			public object Current
			{
				get
				{
					return (m_List.m_Children[m_Position]);
				}
			}
		}
	}
}
