using System;
using System.Drawing;
using JPGL.Common.Primitives;
using JPGL.Math;

namespace JPGL.Common.Viewports
{
	/// <summary>
	/// Viewport provides the public interface and base implementation for
	/// device specific viewports. Device specific viewports must implement
	/// the abstract functions defined here.
	/// </summary>
	public abstract class Viewport
	{
		//-----------------------------------//
		protected IntPtr   m_WindowContext;  // Handle on the window we are attached to
		protected int	   m_Height;         // The window height (not device size)
		protected int	   m_Width;          // The window width (not device size)
		protected Color    m_BackgroundColor;// Background color of this viewport
		protected Camera   m_Camera;         // This viewport's camera
		private CameraHandler m_CH;          // CameraHandler ref so we can unregister it
		//-----------------------------------//

		#region Window Bound Properties
		//-----------------------------------//
		public int Width                     // Viewport size accessor properties
		{
			get { return m_Width; }
		}
		public int Height
		{
			get { return m_Height; }
		}
		//-----------------------------------//
		#endregion

		#region Camera Properties
		//-----------------------------------//
		public Camera Camera
		{
			get { return m_Camera; }
		}
		//-----------------------------------//
		#endregion

		//-----------------------------------//
		protected abstract PrimitiveRenderer PrimRenderer
		{                                    // Inheriting classes must return
			get;                             // an appropriate Primitive Renderer.
		}                                    // 
		//-----------------------------------//


		//-------------------------------------------------------------------------//
		public Viewport()
		{
			m_Height  = -1;
			m_Width   = -1;

			m_Camera = new Camera();
			m_CH = new CameraHandler(CameraUpdateHandler);
			m_Camera.CameraUpdate += m_CH;

			m_BackgroundColor = Color.LightBlue;
		}

		//-------------------------------------------------------------------------//
		public void SetBackgroundColor(int r, int g, int b)
		{
			m_BackgroundColor = Color.FromArgb(r, g, b);
		}

		//-------------------------------------------------------------------------//
		// Attach the Viewport to the given camera. To do this, we update our camera
		// reference member, and register an update event with the camera.
		public void AttachToCamera(Camera c)
		{
			if (c == m_Camera) return;

			// Unregister with the current camera
			if (m_Camera != null)
				m_Camera.CameraUpdate -= m_CH;

			m_Camera = c;
			m_Camera.CameraUpdate += m_CH;

			Resize();
		}

		//-------------------------------------------------------------------------//
		// When a Camera Update happens, we want to call resize so the new Camera 
		// details can be taken into consideration.
		private void CameraUpdateHandler(object sender, CameraEventArgs e)
		{
			Resize();
		}

		//-------------------------------------------------------------------------//
		// Translate window coordinates to device coordinates (pixel space
		public bool HardwareToDevice(int hX, int hY, out int dX, out int dY)
		{
			if (m_Width<0 || m_Height<0) 
			{
				dX = 0;
				dY = 0;
				return false;
			}

			dX = hX;
			dY = m_Height - hY;

			return true;
		}


		//-------------------------------------------------------------------------//
		public bool WorldToObject(SceneNode obj, Vector3 vw, out Vector3 vo)
		{
			InitForUnProject();
			BuildMatrix(obj);
			UnProject(vw, out vo);
			RestoreFromUnProject();

			return true;
		}

		//-------------------------------------------------------------------------//
		public bool ObjectToWorld(SceneNode obj, Vector3 vo, out Vector3 vw)
		{
			InitForUnProject();
			BuildMatrix(obj);
			Project(vo, out vw);
			RestoreFromUnProject();

			return true;
		}

		//-------------------------------------------------------------------------//
		// Translate device coords to world coordinates
		public bool DeviceToHardware(int dx, int dy, out int hx, out int hy)
		{
			if (m_Width<0 || m_Height<0) 
			{
				hx = 0;
				hy = 0;
				return false;
			}

			hx = dx;
			hy = m_Height - dy;

			return true;
		}

		//-------------------------------------------------------------------------//
		// Create and Initialize graphics drawing context, all we do here is save
		// the arguments to our private data members, inheriting classes must implement
		// the Create() method, and may assume that m_WindowContext, height, width and
		// oar have all been set.
		public bool CreateGraphicsDevice(IntPtr hOutputWindow, int width, int height)
		{
			m_WindowContext = hOutputWindow;

			m_Height  = height;
			m_Width   = width;
		
			return Create();
		}

		//-------------------------------------------------------------------------//
		// See comments for CreateGraphicsDevice
		public void ResizeGraphicsDevice(int width, int height)
		{
			m_Height = height;
			m_Width  = width;

			// Call device specific implementation
			Resize();
		}

		//-------------------------------------------------------------------------//
		public void InitializeDeviceForDrawing()
		{
			// Call device specific implementation
			Initialize();
		}

		//-------------------------------------------------------------------------//
		public void DrawToDevice()
		{
			// Call device specific implementation
			Draw();
		}

		//-------------------------------------------------------------------------//
		public void RestoreDeviceAfterDrawing()
		{
			// Call device specific implementation
			Restore();
		}

		//-------------------------------------------------------------------------//
		// Implementing classes must implement these following functions
		abstract protected bool Create();
		abstract protected void Resize();
		abstract protected void Initialize();
		abstract protected void Draw();
		abstract protected void Restore();

		abstract protected void InitForUnProject();
		abstract protected void BuildMatrix(SceneNode obj);
		abstract protected void Project(Vector3 vo, out Vector3 vw);
		abstract protected void UnProject(Vector3 vw, out Vector3 vo);
		abstract protected void RestoreFromUnProject();

		//-------------------------------------------------------------------------//
		// Empty default implementations. By default we do nothing for any of these
		// and let implementing classes define as many as they wish to provide some
		// behavior for.
		virtual public void OnMouseMove(int mx, int my, uint nDown)  {}
		virtual public void OnMouseDown(int mx, int my, uint nDown)  {}
		virtual public void OnMouseUp(int mx, int my, uint nDown)    {}


	}
}
