using System;
using System.Collections;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

using Tao.OpenGl;
using Tao.Platform.Windows;


namespace JPGL.OpenGL
{
	/// <summary>
	/// Summary description for OpenGLGraphicsSystem.
	/// </summary>
	public struct OpenGLContext 
	{
		public IntPtr dc, rc;
	}

	public class GraphicsSystem
	{
		//-----------------------------------// Texture array
		public static int[] Textures = new int[64];
		//-----------------------------------//

		//-----------------------------------// 
		private static int m_NextTexture = -1;// Next texture index to queue
		private static int m_LastTexture = -1;// Last texture we actually loaded
		//-----------------------------------//  into OpenGL.

		//-----------------------------------// Textures waiting to be loaded.
		private static Queue m_QueuedTextures = new Queue(2); 
		//-----------------------------------//

		//-----------------------------------// OpenGL contexts
		private static ArrayList m_OpenGLContexts = new ArrayList(2);
		//-----------------------------------//

		//-----------------------------------//
		private static IntPtr m_ActiveRC;
		private static IntPtr m_ActiveDC;
		//-----------------------------------//

		//-----------------------------------//
		private static bool   m_TakeScreenshot = false;
		private static int[]  m_ScreenshotSize = new int[2];
		private static string m_ScreenshotFile = "";
		//-----------------------------------//

		//-----------------------------------// Extensions
		public static bool PointSpritesSupported        = false;
		public static bool VertexBufferObjectsSupported = false;
		//-----------------------------------//

		//-----------------------------------//
		private static int  m_FontBase;
		private static uint m_FontTexture;
		//-----------------------------------//


		//-------------------------------------------------------------------------//
		public static JPGL.Common.Texture LoadTexture(string fileName)
		{
			string fileNameWPath;
			if (System.IO.File.Exists(fileName))
				fileNameWPath = fileName;
			else if (System.IO.File.Exists("../../../Textures/" + fileName))
				fileNameWPath = "../../../Textures/" + fileName;
			else if (System.IO.File.Exists("Textures/" + fileName))
				fileNameWPath = "Textures/" + fileName;
			else
				return new JPGL.Common.Texture(-1);

			m_QueuedTextures.Enqueue(fileNameWPath);

			m_NextTexture++;
			return new JPGL.Common.Texture(m_NextTexture);
		}

		//-------------------------------------------------------------------------//
		public static void UnloadTexture(JPGL.Common.Texture t)
		{
			uint tid = (uint)Textures[t.Index];
			Gl.glDeleteTextures(1, ref tid);
		}

		//-------------------------------------------------------------------------//
		// Contexts have to be registered so they can have textures loaded on them
		public static void RegisterContext(IntPtr dc, IntPtr rc)
		{
			if (m_OpenGLContexts.Count > 0)
			{
				OpenGLContext oglcold = (OpenGLContext)m_OpenGLContexts[0];
				Wgl.wglShareLists(oglcold.rc, rc);
			}
			else// 1st Registering context gets to do extension checking
			{
				// Check for and load point sprite extension
				if (GlExtensionLoader.LoadExtension("GL_ARB_point_parameters"))
					PointSpritesSupported = true;
		
				// Check for and load VBO extension
				if (GlExtensionLoader.LoadExtension("GL_ARB_vertex_buffer_object"))
					VertexBufferObjectsSupported = true;

			}

			// Register context to share lists
			OpenGLContext oglc;
			oglc.dc = dc;
			oglc.rc = rc;
			m_OpenGLContexts.Add(oglc);
		}

		//-------------------------------------------------------------------------//
		private static void LoadQueuedTextures()
		{
			while (m_QueuedTextures.Count > 0)
			{
				m_LastTexture++;

				string fileName = (string)m_QueuedTextures.Dequeue();
				int    index    = m_LastTexture;

				// Get the bitmap data
				Bitmap image = new Bitmap(fileName);
				image.RotateFlip(RotateFlipType.RotateNoneFlipY);
				Rectangle rectangle = new Rectangle(0, 0, image.Width, image.Height);

				Gl.glGenTextures(1, out Textures[index]);
				Gl.glBindTexture(Gl.GL_TEXTURE_2D, Textures[index]);
				
				if (image.PixelFormat == PixelFormat.Format32bppArgb) // PNG or such, with alpha
				{
					BitmapData bitmapData = image.LockBits(
						rectangle, 
						ImageLockMode.ReadOnly, 
						PixelFormat.Format32bppArgb);

					Glu.gluBuild2DMipmaps(
						Gl.GL_TEXTURE_2D, 4, 
						image.Width, image.Height, 
						Gl.GL_BGRA, Gl.GL_UNSIGNED_BYTE, 
						bitmapData.Scan0);
		
					image.UnlockBits(bitmapData);
				}
				else // Just assume no alpha
				{
					BitmapData bitmapData = image.LockBits(
						rectangle, 
						ImageLockMode.ReadOnly, 
						PixelFormat.Format24bppRgb);

					Glu.gluBuild2DMipmaps(
						Gl.GL_TEXTURE_2D, 3, 
						image.Width, image.Height, 
						Gl.GL_BGR, Gl.GL_UNSIGNED_BYTE, 
						bitmapData.Scan0);

					image.UnlockBits(bitmapData);
				}
		
				image.Dispose();
			}
		}

		//-------------------------------------------------------------------------//
		public static void Terminate(IntPtr renderingContext)
		{
			// Deletes the given RenderingContext. If the context
			// is the active context, the wgl funciton will make
			// it not current before deleting it, so we do not need
			// to call wglMakeCurrent(NULL, NULL).
			Wgl.wglDeleteContext(renderingContext);
		}

		//-------------------------------------------------------------------------//
		public static bool SetActiveDevice(IntPtr deviceContext, IntPtr renderingContext)
		{
			// Deactivating a context is now disabled, you can only switch to another
			// context. Neither can you switch to the already active context. We are
			// trying to minimize context switches here, and keep a context active
			// at all times so our extremely context-switch heavy project/unproject
			// functionality in OpenGLViewport has a context to use.
			if (deviceContext == IntPtr.Zero || renderingContext == IntPtr.Zero) return true;
			if (deviceContext == m_ActiveDC && renderingContext == m_ActiveRC)
			{
				LoadQueuedTextures();
				return true;
			}

			// Each thread gets one rendering pipe to talk to the graphics device with.
			// This function tells the system that the given Rendering Context will be the
			// one using the pipe. All calls to opengl functions will render to the pipe
			// through this rendering context (and device context for that matter) until
			// we release the pipe (by calling this function with IntPtr.Zero values).
			if(!Wgl.wglMakeCurrent(deviceContext, renderingContext))
			{
				return false;
			}
			m_ActiveRC = renderingContext;
			m_ActiveDC = deviceContext;

			// There is some extra work to do now for textures. Since we need to load 
			// textures when a context is active, yet LoadTextures has presumably been 
			// called when no context is active, we need to actually load any textures
			// that have yet to be loaded.
			LoadQueuedTextures();

			// Do we need to take a screenshot? This is done a little poorly, as
			// it will only take a screenshot of the first registered context, but
			// the caller must specify the size of the shot...
			if (m_TakeScreenshot)
			{
				SaveScreenshot();
				m_TakeScreenshot = false;
			}
		
			return true;	
		}

		//-------------------------------------------------------------------------//
		public static bool InitPixelFormat(IntPtr deviceContext)
		{
			// Set up pixel formal
			Gdi.PIXELFORMATDESCRIPTOR pfd = new Gdi.PIXELFORMATDESCRIPTOR();    // pfd Tells Windows How We Want Things To Be
			pfd.nSize = (short) Marshal.SizeOf(pfd);                            // Size Of This Pixel Format Descriptor
			pfd.nVersion = 1;                                                   // Version Number
			pfd.dwFlags = Gdi.PFD_DRAW_TO_WINDOW |                              // Format Must Support Window
				Gdi.PFD_SUPPORT_OPENGL |                                        // Format Must Support OpenGL
				Gdi.PFD_DOUBLEBUFFER;                                           // Format Must Support Double Buffering
			pfd.iPixelType = (byte) Gdi.PFD_TYPE_RGBA;                          // Request An RGBA Format
			pfd.cColorBits = (byte) 32;                                         // Select Our Color Depth
			pfd.cRedBits = 0;                                                   // Color Bits Ignored
			pfd.cRedShift = 0;
			pfd.cGreenBits = 0;
			pfd.cGreenShift = 0;
			pfd.cBlueBits = 0;
			pfd.cBlueShift = 0;
			pfd.cAlphaBits = 0;                                                 // No Alpha Buffer
			pfd.cAlphaShift = 0;                                                // Shift Bit Ignored
			pfd.cAccumBits = 0;                                                 // No Accumulation Buffer
			pfd.cAccumRedBits = 0;                                              // Accumulation Bits Ignored
			pfd.cAccumGreenBits = 0;
			pfd.cAccumBlueBits = 0;
			pfd.cAccumAlphaBits = 0;
			pfd.cDepthBits = 16;                                                // 16Bit Z-Buffer (Depth Buffer)
			pfd.cStencilBits = 0;                                               // No Stencil Buffer
			pfd.cAuxBuffers = 0;                                                // No Auxiliary Buffer
			pfd.iLayerType = (byte) Gdi.PFD_MAIN_PLANE;                         // Main Drawing Layer
			pfd.bReserved = 0;                                                  // Reserved
			pfd.dwLayerMask = 0;                                                // Layer Masks Ignored
			pfd.dwVisibleMask = 0;
			pfd.dwDamageMask = 0;

			// Tell the graphics hardware how to display pixels.
			int pixelFormat = Gdi.ChoosePixelFormat(deviceContext, ref pfd);
			if (pixelFormat == 0) return false;

			if(!Gdi.SetPixelFormat(deviceContext, pixelFormat, ref pfd)) return false;

			return true;
		}

		//-------------------------------------------------------------------------//
		public static void TakeScreenshot(int width, int height, string filename)
		{
			m_TakeScreenshot = true;
			m_ScreenshotSize[0] = width;
			m_ScreenshotSize[1] = height;
			m_ScreenshotFile = filename;
		}

		//-------------------------------------------------------------------------//
		private static void SaveScreenshot()
		{
			System.Drawing.Bitmap b = new System.Drawing.Bitmap(m_ScreenshotSize[0], m_ScreenshotSize[1], System.Drawing.Imaging.PixelFormat.Format24bppRgb);
			System.Drawing.Rectangle rect = new System.Drawing.Rectangle(0, 0, m_ScreenshotSize[0], m_ScreenshotSize[1]); 
			System.Drawing.Imaging.BitmapData bitmapData = b.LockBits(rect, System.Drawing.Imaging.ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format24bppRgb);
    
			// read the pixels from the GL buffer
			Gl.glReadPixels(0, 0, m_ScreenshotSize[0], m_ScreenshotSize[1], Gl.GL_BGR, Gl.GL_UNSIGNED_BYTE, bitmapData.Scan0); 
     
			b.UnlockBits(bitmapData); 
			b.RotateFlip(System.Drawing.RotateFlipType.RotateNoneFlipY);
			b.Save(m_ScreenshotFile, System.Drawing.Imaging.ImageFormat.Jpeg);
		}

		//-------------------------------------------------------------------------//
		public static int CreateNewCache()
		{
			int listid = Gl.glGenLists(1);
			Gl.glNewList(listid, Gl.GL_COMPILE_AND_EXECUTE);
			return listid;
		}

		//-------------------------------------------------------------------------//
		public static void EndNewCache()
		{
			Gl.glEndList();
		}

		//-------------------------------------------------------------------------//
		public static void DeleteCache(int id)
		{
			Gl.glDeleteLists(id, 1);
		}

		//-------------------------------------------------------------------------//
		public static void DisplayCache(int id)
		{
			Gl.glCallList(id);
		}

		//-------------------------------------------------------------------------//
		public static void StartBackgroundDraw()
		{
			Gl.glDepthMask(false);
		}

		//-------------------------------------------------------------------------//
		public static void EndBackgroundDraw()
		{
			Gl.glDepthMask(true);
		}

		//-------------------------------------------------------------------------//
		public static void BuildFonts(string filename)
		{
			#region Load font Texture (m_FontTexture)
			string fileNameWPath;
			if (System.IO.File.Exists(filename))
				fileNameWPath = filename;
			else if (System.IO.File.Exists("../../../Textures/" + filename))
				fileNameWPath = "../../../Textures/" + filename;
			else if (System.IO.File.Exists("Textures/" + filename))
				fileNameWPath = "Textures/" + filename;
			else
				return;
			Bitmap image = new Bitmap(fileNameWPath);
			image.RotateFlip(RotateFlipType.RotateNoneFlipY);
			Rectangle rectangle = new Rectangle(0, 0, image.Width, image.Height);

			Gl.glGenTextures(1, out m_FontTexture);
			Gl.glBindTexture(Gl.GL_TEXTURE_2D, m_FontTexture);
				
			if (image.PixelFormat == PixelFormat.Format32bppArgb) // PNG or such, with alpha
			{
				BitmapData bitmapData = image.LockBits(
					rectangle, 
					ImageLockMode.ReadOnly, 
					PixelFormat.Format32bppArgb);

				Glu.gluBuild2DMipmaps(
					Gl.GL_TEXTURE_2D, 4, 
					image.Width, image.Height, 
					Gl.GL_BGRA, Gl.GL_UNSIGNED_BYTE, 
					bitmapData.Scan0);
		
				image.UnlockBits(bitmapData);
			}
			else // Just assume no alpha
			{
				BitmapData bitmapData = image.LockBits(
					rectangle, 
					ImageLockMode.ReadOnly, 
					PixelFormat.Format24bppRgb);

				Glu.gluBuild2DMipmaps(
					Gl.GL_TEXTURE_2D, 3, 
					image.Width, image.Height, 
					Gl.GL_BGR, Gl.GL_UNSIGNED_BYTE, 
					bitmapData.Scan0);

				image.UnlockBits(bitmapData);
			}
		
			image.Dispose();
			#endregion

			float cx,cy;
			m_FontBase = Gl.glGenLists(256);
			Gl.glBindTexture(Gl.GL_TEXTURE_2D, m_FontTexture);
			for (int loop=0; loop<256; loop++)
			{
				cx = ((float)(loop%16))/16.0f;
				cy = ((float)(loop/16))/16.0f;

				Gl.glNewList(m_FontBase+loop, Gl.GL_COMPILE);
				Gl.glBegin(Gl.GL_QUADS);
				Gl.glTexCoord2f(cx, 1.0f-cy-0.0625f);
				Gl.glVertex2i(0, 0);
				Gl.glTexCoord2f(cx+0.0625f, 1.0f-cy-0.0625f);
				Gl.glVertex2i(32, 0);
				Gl.glTexCoord2f(cx+0.0625f, 1.0f-cy);
				Gl.glVertex2i(32, 32);
				Gl.glTexCoord2f(cx, 1.0f-cy);
				Gl.glVertex2i(0, 32);
				Gl.glEnd();
				Gl.glTranslated(20, 0, 0);
				Gl.glEndList();
			}
		}

		//-------------------------------------------------------------------------//
		public static void KillFonts()
		{
			Gl.glDeleteLists(m_FontBase, 256);
			Gl.glDeleteTextures(1, ref m_FontTexture);
		}

		//-------------------------------------------------------------------------//
		// This assumes depth testing is disabled and we are in ortho mode!
		public static void Print(
			int x, int y, float sx, float sy,
			string str, int charset,
			float r, float g, float b)
		{
			if (charset > 1) charset = 1;


			Gl.glDisable(Gl.GL_LIGHTING);
			Gl.glEnable(Gl.GL_TEXTURE_2D);
			Gl.glBindTexture(Gl.GL_TEXTURE_2D, m_FontTexture);
			Gl.glTexEnvi(Gl.GL_TEXTURE_ENV, Gl.GL_TEXTURE_ENV_MODE, Gl.GL_MODULATE);
			Gl.glColor3f(r, g, b);
			
			Gl.glEnable(Gl.GL_BLEND);
			Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA);

			Gl.glMatrixMode(Gl.GL_TEXTURE);
			Gl.glLoadIdentity();
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glPushMatrix();
			Gl.glLoadIdentity();
			Gl.glScalef(sx, sy, 1.0f);
			Gl.glTranslated(x, y, 0);
			Gl.glListBase(m_FontBase - 32 + (128 * charset));
			Gl.glCallLists(str.Length, Gl.GL_UNSIGNED_BYTE, str);
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glPopMatrix();

			Gl.glDisable(Gl.GL_BLEND);
			Gl.glDisable(Gl.GL_TEXTURE_2D);
		}
	}
}
