using System;
using System.Collections;
using System.Drawing;
using System.Runtime.InteropServices;

using JPGL.Common;
using JPGL.Common.Lighting;
using JPGL.Common.Primitives;
using JPGL.Math;
using Tao.OpenGl;


namespace JPGL.OpenGL
{
	/// <summary>
	/// Summary description for OpenGLPrimitiveRenderer.
	/// </summary>
	public class OpenGLPrimitiveRenderer : PrimitiveRenderer
	{
		private static Hashtable m_ListLookup  = new Hashtable();  // Display list lookup table
		private static Hashtable m_LightLookup = new Hashtable(8); // Light lookup table
		private static int       m_LightCount  = 0;                // # of lights in scene
		private static int[]     m_glLightTable = { Gl.GL_LIGHT0, Gl.GL_LIGHT1, Gl.GL_LIGHT2, Gl.GL_LIGHT3, Gl.GL_LIGHT4, Gl.GL_LIGHT5, Gl.GL_LIGHT6, Gl.GL_LIGHT7 };

		#region VBO Lookup Table and Struct
		private static Hashtable m_VBOLookup = new Hashtable();  // VBO lookup table
		private struct VBO
		{
			public uint vId, tId, nId;
		}
		#endregion

		//-------------------------------------------------------------------------//
		protected override void ApplyTransformation(Transform t)
		{
			// Push the modelview matrix
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glPushMatrix();
			
			// Translate to the position offset with the pivot
			Gl.glTranslatef(t.T.X+t.P.X, t.T.Y+t.P.Y, t.T.Z+t.P.Z);

			// Rotate in x, y and z
			Gl.glMultMatrixf(t.RQ.GetMatrix());

			// Scale
			Gl.glScalef(t.S.X, t.S.Y, t.S.Z);

			// Translate away from pivot to control point
			Gl.glTranslatef(-1.0f*t.P.X, -1.0f*t.P.Y, -1.0f*t.P.Z);			
		}

		//-------------------------------------------------------------------------//
		protected override void UndoLastTransform()
		{
			// Pop the ModelView matrix
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glPopMatrix();
		}

		//-------------------------------------------------------------------------//
		protected override void SetupLight(Light l)
		{
			// Check the lookup table for the light id
			int lightId = -1;
			if (m_LightLookup.ContainsKey(l))
			{
				lightId = (int)m_LightLookup[l];
			}
			else
			{
				if (m_LightCount+1 >= 8) return;
				lightId = m_glLightTable[m_LightCount];
				m_LightLookup.Add(l, lightId);
				m_LightCount++;
			}

			// If the light is off, disable it
			if (!l.On)
			{
				Gl.glDisable(lightId);
				return;
			}

			// Set Colors
			float[] alcolor = {l.AmbientColor.R / 255.0f, l.AmbientColor.G / 255.0f, l.AmbientColor.B / 255.0f, l.AmbientColor.A / 255.0f};
			Gl.glLightfv(lightId, Gl.GL_AMBIENT, alcolor);

			float[] dlcolor = {l.DiffuseColor.R / 255.0f, l.DiffuseColor.G / 255.0f, l.DiffuseColor.B / 255.0f, l.DiffuseColor.A / 255.0f};
			Gl.glLightfv(lightId, Gl.GL_DIFFUSE, dlcolor);

			float[] slcolor = {l.SpecularColor.R / 255.0f, l.SpecularColor.G / 255.0f, l.SpecularColor.B / 255.0f, l.SpecularColor.A / 255.0f};
			Gl.glLightfv(lightId, Gl.GL_SPECULAR, slcolor);

			// Set attenuation
			Gl.glLightf(lightId, Gl.GL_CONSTANT_ATTENUATION,  l.Attenuation0);
			Gl.glLightf(lightId, Gl.GL_LINEAR_ATTENUATION,    l.Attenuation1);
			Gl.glLightf(lightId, Gl.GL_QUADRATIC_ATTENUATION, l.Attenuation2);

			// Point light has position only
			if (l.LightType == Light.LIGHT_POINT)
			{
				float[] lpos = {l.Position.X, l.Position.Y, l.Position.Z, 1.0f};
				Gl.glLightfv(lightId, Gl.GL_POSITION, lpos);
			}

			// Directional Light has direction only
			if (l.LightType == Light.LIGHT_DIRECTIONAL)
			{			
				float[] lpos = {l.Direction.X, l.Direction.Y, l.Direction.Z, 0.0f};
				Gl.glLightfv(lightId, Gl.GL_POSITION, lpos);
			}
			
			// Spot Light has dir, pos, theta and phi
			if (l.LightType == Light.LIGHT_SPOT)
			{
				float[] lpos = {l.Position.X, l.Position.Y, l.Position.Z, 1.0f};
				float[] ldir = {l.Direction.X, l.Direction.Y, l.Direction.Z};

				Gl.glLightfv(lightId, Gl.GL_POSITION, lpos);
				Gl.glLightfv(lightId, Gl.GL_SPOT_DIRECTION, ldir);
				Gl.glLightf(lightId, Gl.GL_SPOT_CUTOFF, (l.Theta * 180.0f) / (float)System.Math.PI);
				Gl.glLightf(lightId, Gl.GL_SPOT_EXPONENT, l.Falloff);
			}

			// Turn the light on
			Gl.glEnable(lightId);
		}

		//-------------------------------------------------------------------------//
		protected override void PrepareForPrimitive(Geometry g, DrawablePrimitive p)
		{
			// Point size
			Gl.glPointSize(p.PointSize);

			// Shading
			Gl.glShadeModel((p.Shaded ? SMOOTH : FLAT));

			// Fill mode
			Gl.glPolygonMode(Gl.GL_FRONT_AND_BACK, (p.Filled ? FILL : (p.Wireframe ? LINE : POINT)));

			// Texturing
			if (p.TextureOn)
			{
				Gl.glEnable(Gl.GL_TEXTURE_2D);
				Gl.glBindTexture(Gl.GL_TEXTURE_2D, GraphicsSystem.Textures[p.Texture.Index]);

				Gl.glTexEnvi(Gl.GL_TEXTURE_ENV, Gl.GL_TEXTURE_ENV_MODE, Gl.GL_MODULATE);
				Gl.glColor3f(1.0f, 1.0f, 1.0f); // To offset texture modulation

				#region Texture Settings
				// Minimization Filter
				int minfilter = -1;
				if      (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_NEAREST)
					minfilter = Gl.GL_NEAREST;
				else if (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_LINEAR)
					minfilter = Gl.GL_LINEAR;
				else if (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_NEAREST_MIPMAP_NEAREST)
					minfilter = Gl.GL_NEAREST_MIPMAP_NEAREST;
				else if (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_NEAREST_MIPMAP_LINEAR)
					minfilter = Gl.GL_NEAREST_MIPMAP_LINEAR;
				else if (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_LINEAR_MIPMAP_LINEAR)
					minfilter = Gl.GL_LINEAR_MIPMAP_LINEAR;
				else //if (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_LINEAR_MIPMAP_NEAREST)
					minfilter = Gl.GL_LINEAR_MIPMAP_NEAREST;

				// Magnification Filter
				int magfilter = -1;
				if (p.Texture.MagFilter == Texture.FilterType.FILTER_TYPE_NEAREST)
					magfilter = Gl.GL_NEAREST;
				else //if (p.Texture.MinFilter == Texture.FilterType.FILTER_TYPE_LINEAR)
					magfilter = Gl.GL_LINEAR;
				
				Gl.glTexParameteri(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MAG_FILTER, minfilter);
				Gl.glTexParameteri(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MIN_FILTER, magfilter);

				// Addressing U
				int addressingu = -1;
				if      (p.Texture.AddressingU == Texture.AddressingMethod.ADDRESSING_METHOD_CLAMP)
					addressingu = Gl.GL_CLAMP;
				else if (p.Texture.AddressingU == Texture.AddressingMethod.ADDRESSING_METHOD_CLAMP_TO_BORDER)
					addressingu = Gl.GL_CLAMP_TO_BORDER;
				else if (p.Texture.AddressingU == Texture.AddressingMethod.ADDRESSING_METHOD_CLAMP_TO_EDGE)
					addressingu = Gl.GL_CLAMP_TO_EDGE;
				else if (p.Texture.AddressingU == Texture.AddressingMethod.ADDRESSING_METHOD_MIRRORED_REPEAT)
					addressingu = Gl.GL_MIRRORED_REPEAT;
				else if (p.Texture.AddressingU == Texture.AddressingMethod.ADDRESSING_METHOD_REPEAT)
					addressingu = Gl.GL_REPEAT;

				// Addressing V
				int addressingv = -1;
				if      (p.Texture.AddressingV == Texture.AddressingMethod.ADDRESSING_METHOD_CLAMP)
					addressingv = Gl.GL_CLAMP;
				else if (p.Texture.AddressingV == Texture.AddressingMethod.ADDRESSING_METHOD_CLAMP_TO_BORDER)
					addressingv = Gl.GL_CLAMP_TO_BORDER;
				else if (p.Texture.AddressingV == Texture.AddressingMethod.ADDRESSING_METHOD_CLAMP_TO_EDGE)
					addressingv = Gl.GL_CLAMP_TO_EDGE;
				else if (p.Texture.AddressingV == Texture.AddressingMethod.ADDRESSING_METHOD_MIRRORED_REPEAT)
					addressingv = Gl.GL_MIRRORED_REPEAT;
				else if (p.Texture.AddressingV == Texture.AddressingMethod.ADDRESSING_METHOD_REPEAT)
					addressingv = Gl.GL_REPEAT;

				Gl.glTexParameteri(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_WRAP_S, addressingu);
				Gl.glTexParameteri(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_WRAP_T, addressingv);

				// Texture transformation, we are going to do something bad here
				// and destroy whatever is in the texture matrix. Very very bad.
				Gl.glMatrixMode(Gl.GL_TEXTURE);
				Gl.glLoadIdentity();
				Gl.glTranslatef(p.Texture.TransX+0.5f, p.Texture.TransY+0.5f, 0);
				Gl.glRotatef(p.Texture.Rotation, 0, 0, 1);
				Gl.glScalef(p.Texture.ScaleX, p.Texture.ScaleY, 1);
				Gl.glTranslatef(-0.5f, -0.5f, 0);

				// Border color
				float[] bcolor = {p.Texture.BorderColor.R / 255.0f, p.Texture.BorderColor.G / 255.0f, p.Texture.BorderColor.B / 255.0f, p.Texture.BorderColor.A / 255.0f};
				Gl.glTexParameterfv(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_BORDER_COLOR, bcolor);
				#endregion
			}
			else
				Gl.glDisable(Gl.GL_TEXTURE_2D);

			// Alpha blending
			if (p.Blended)
			{
				Gl.glEnable(Gl.GL_BLEND);
				Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA);
				Gl.glDisable(Gl.GL_DEPTH_TEST);
			}
			else
			{
				Gl.glDisable(Gl.GL_BLEND);
				Gl.glEnable(Gl.GL_DEPTH_TEST);
			}

			// Illumination
			if (p.Lighted)
			{
				// Turn lighting on
				Gl.glEnable(Gl.GL_LIGHTING);

				// Get the material
				Material m = p.Material;

				#region Material Settings
				float[] acolor = {m.Ambient.R / 255.0f, m.Ambient.G / 255.0f, m.Ambient.B / 255.0f, m.Ambient.A / 255.0f};
				Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_AMBIENT, acolor);

				float[] scolor = {m.Specular.R / 255.0f, m.Specular.G / 255.0f, m.Specular.B / 255.0f, m.Specular.A / 255.0f};
				Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_SPECULAR, scolor);

				float[] dcolor = {m.Diffuse.R / 255.0f, m.Diffuse.G / 255.0f, m.Diffuse.B / 255.0f, m.Diffuse.A / 255.0f};
				Gl.glMaterialfv(Gl.GL_FRONT_AND_BACK, Gl.GL_DIFFUSE, dcolor);

				Gl.glMaterialf(Gl.GL_FRONT_AND_BACK, Gl.GL_SHININESS, m.Shininess);

				Gl.glEnable(Gl.GL_COLOR_MATERIAL);
				Gl.glColorMaterial(Gl.GL_FRONT_AND_BACK, Gl.GL_AMBIENT_AND_DIFFUSE);
				#endregion
			}
			else
			{
				// Turn lighting off
				Gl.glDisable(Gl.GL_LIGHTING);
				Gl.glDisable(Gl.GL_COLOR_MATERIAL);
			}
		}

		//-------------------------------------------------------------------------//
		// Protected interface that must be implemented.
		protected override void DrawGeometry(Geometry g, float[] c, bool textured)
		{
			// If the Geometry is cached and dirty, delete the display list
			if (g.AutoCache && 
				textured && 
				g.GeometryID != -1 && 
				g.Dirty && 
				m_ListLookup.ContainsKey(g.GeometryID))
			{
				Gl.glDeleteLists((int)m_ListLookup[g.GeometryID], 1);
				m_ListLookup.Remove(g.GeometryID);
			}

			// Color array is not included in the display list
			if (!textured && c != null)
			{
				Gl.glEnableClientState(Gl.GL_COLOR_ARRAY);
				Gl.glColorPointer(4, Gl.GL_FLOAT, 0, c);
			}
			else
				Gl.glDisableClientState(Gl.GL_COLOR_ARRAY);

			// If a display list is available, use it
			if (g.AutoCache && 
				textured && 
				g.GeometryID != -1 && 
				m_ListLookup.ContainsKey(g.GeometryID))
			{
				Gl.glCallList((int)m_ListLookup[g.GeometryID]);
				return;
			}

			// Create a display list
			int listid = -1;
			if (g.AutoCache && 
				textured && 
				g.GeometryID != -1 && 
				!m_ListLookup.ContainsKey(g.GeometryID))
			{
				listid = Gl.glGenLists(1);
				Gl.glNewList(listid, Gl.GL_COMPILE_AND_EXECUTE);
			}

			// Tell opengl what types of arrays we will be giving it
			Gl.glEnableClientState(Gl.GL_VERTEX_ARRAY);
			Gl.glEnableClientState(Gl.GL_NORMAL_ARRAY);
			Gl.glEnableClientState(Gl.GL_TEXTURE_COORD_ARRAY);

			// Give opengl pointers to our arrays
//			if (GraphicsSystem.VertexBufferObjectsSupported)
//			{
//				DrawGeometryWithVBO(g, c, textured);
//			}
//			else
//			{
				Gl.glVertexPointer(3, Gl.GL_FLOAT, 0, g.Vertices);
				Gl.glNormalPointer(Gl.GL_FLOAT, 0, g.Normals);
				Gl.glTexCoordPointer(2, Gl.GL_FLOAT, 0, g.Texels);	
	//		}
			
			// Draw the Primitive
			Gl.glDrawArrays(GetDeviceDrawmode(g.DrawMode), 0, g.VertexCount);

			// Disable any client states we may have enabled
			Gl.glDisableClientState(Gl.GL_COLOR_ARRAY);
			Gl.glDisableClientState(Gl.GL_NORMAL_ARRAY);
			Gl.glDisableClientState(Gl.GL_VERTEX_ARRAY);
			Gl.glDisableClientState(Gl.GL_TEXTURE_COORD_ARRAY);

			// If we were making a display list, end and save it
			if (g.AutoCache && 
				textured && 
				g.GeometryID != -1 && 
				!m_ListLookup.ContainsKey(g.GeometryID))
			{
				Gl.glEndList();
				m_ListLookup.Add(g.GeometryID, listid);
				g.Dirty = false;
			}
		}

		//-------------------------------------------------------------------------//
		// Draw the geometry, but using VBOs
		private void DrawGeometryWithVBO(Geometry g, float[] c, bool textured)
		{
			// Create VBO if necessary
			VBO vboID;
			if (!m_VBOLookup.ContainsKey(g.GeometryID))
			{
				// Generate And Bind The Vertex Buffer
				Gl.glGenBuffersARB(1, out vboID.vId);
				Gl.glBindBufferARB(Gl.GL_ARRAY_BUFFER_ARB, vboID.vId);
				Gl.glBufferDataARB(
					Gl.GL_ARRAY_BUFFER_ARB, 
					g.VertexCount*3*Marshal.SizeOf(new float()), 
					g.Vertices, 
					Gl.GL_STATIC_DRAW_ARB);

				// Generate And Bind The Texture Coordinate Buffer
				Gl.glGenBuffersARB(1, out vboID.tId);
				Gl.glBindBufferARB(Gl.GL_ARRAY_BUFFER_ARB, vboID.tId);
				Gl.glBufferDataARB(
					Gl.GL_ARRAY_BUFFER_ARB, 
					g.VertexCount*2*Marshal.SizeOf(new float()), 
					g.Texels, 
					Gl.GL_STATIC_DRAW_ARB);

				// Generate And Bind The Normal Buffer
				Gl.glGenBuffersARB(1, out vboID.nId);
				Gl.glBindBufferARB(Gl.GL_ARRAY_BUFFER_ARB, vboID.nId);
				Gl.glBufferDataARB(
					Gl.GL_ARRAY_BUFFER_ARB, 
					g.VertexCount*3*Marshal.SizeOf(new float()), 
					g.Normals, 
					Gl.GL_STATIC_DRAW_ARB);

				m_VBOLookup.Add(g.GeometryID, vboID);
			}
			else
			{
				vboID = (VBO)m_VBOLookup[g.GeometryID];
			}

			Gl.glBindBufferARB(Gl.GL_ARRAY_BUFFER_ARB, vboID.vId);
			Gl.glVertexPointer(3, Gl.GL_FLOAT, 0, System.IntPtr.Zero);
			Gl.glBindBufferARB(Gl.GL_ARRAY_BUFFER_ARB, vboID.nId);
			Gl.glNormalPointer(Gl.GL_FLOAT, 0, System.IntPtr.Zero);
			Gl.glBindBufferARB(Gl.GL_ARRAY_BUFFER_ARB, vboID.tId);
			Gl.glTexCoordPointer(2, Gl.GL_FLOAT, 0, System.IntPtr.Zero);
		}

		//-------------------------------------------------------------------------//
		// Draw a geometry that has Indices for faces, rather than simply vertices.
		// Use of glInterleavedArrays prevents caching in a display list.
		protected override void DrawIndexedGeometry(IndexedGeometry g)
		{
			// Allocate an unmanaged handle to access our managed array.
			GCHandle gch = GCHandle.Alloc(g.VertexData, GCHandleType.Pinned);
			IntPtr vertexPtr = gch.AddrOfPinnedObject();

			// Tell OpenGL about the array and draw it
			Gl.glInterleavedArrays(Gl.GL_T2F_C4F_N3F_V3F, 0, vertexPtr);
			Gl.glDrawElements(GetDeviceDrawmode(g.DrawMode), g.Indices.Length, Gl.GL_UNSIGNED_INT, g.Indices);

			// Free the unmanaged handle
			gch.Free();
		}

		//-------------------------------------------------------------------------//
		public override void RenderParticleSystem(ParticleSystem s)
		{
			if (!GraphicsSystem.PointSpritesSupported) return;

			ApplyTransformation(s.T);

			Gl.glEnable(Gl.GL_TEXTURE_2D);
			Gl.glBindTexture(Gl.GL_TEXTURE_2D, GraphicsSystem.Textures[ParticleSystem.Texture.Index]);
			
			Gl.glTexEnvi(Gl.GL_TEXTURE_ENV, Gl.GL_TEXTURE_ENV_MODE, Gl.GL_MODULATE);
			Gl.glColor3f(0, 0, 0);

			Gl.glTexParameteri(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MIN_FILTER, Gl.GL_LINEAR);
			Gl.glTexParameteri(Gl.GL_TEXTURE_2D, Gl.GL_TEXTURE_MAG_FILTER, Gl.GL_LINEAR);

			Gl.glEnable(Gl.GL_DEPTH_TEST);
			Gl.glDepthMask(false);

			Gl.glEnable(Gl.GL_BLEND);
			Gl.glBlendFunc(Gl.GL_SRC_ALPHA, Gl.GL_ONE);

			// This is how will our point sprite's size will be modified by 
			// distance from the viewer
			Gl.glPointParameterfvARB(Gl.GL_POINT_DISTANCE_ATTENUATION_ARB, new float[]{ 1.0f, 0.0f, 0.01f });

			// Query for the max point size supported by the hardware
			float maxSize; Gl.glGetFloatv(Gl.GL_POINT_SIZE_MAX_ARB, out maxSize);

			// Clamp size to 100.0f or the sprites could get a little too big on some  
			// of the newer graphic cards.
			if( maxSize > 100.0f )
				maxSize = 100.0f;

			Gl.glPointSize(maxSize);

			// The alpha of a point is calculated to allow the fading of points 
			// instead of shrinking them past a defined threshold size. The threshold 
			// is defined by GL_POINT_FADE_THRESHOLD_SIZE_ARB and is not clamped to 
			// the minimum and maximum point sizes.
			Gl.glPointParameterfARB(Gl.GL_POINT_FADE_THRESHOLD_SIZE_ARB, 60.0f);

			Gl.glPointParameterfARB(Gl.GL_POINT_SIZE_MIN_ARB, 1.0f);
			Gl.glPointParameterfARB(Gl.GL_POINT_SIZE_MAX_ARB, maxSize);

			// Specify point sprite texture coordinate replacement mode for each 
			// texture unit
			Gl.glTexEnvf(Gl.GL_POINT_SPRITE_ARB, Gl.GL_COORD_REPLACE_ARB, Gl.GL_TRUE);

			Gl.glEnable(Gl.GL_POINT_SPRITE_ARB);
			
			Gl.glBegin( Gl.GL_POINTS );
			{
				foreach (ParticleSystem.Particle p in s.Particles)
				{
					if (!p.alive) continue;

					Gl.glColor4f(p.r, p.g, p.b, 1.0f);
					Gl.glVertex3f(p.x, p.y, p.z);
				}
			}
			Gl.glEnd();

			Gl.glDisable(Gl.GL_POINT_SPRITE_ARB);

			Gl.glDisable(Gl.GL_DEPTH_TEST);
			Gl.glDepthMask(true);

			UndoLastTransform();
		}

		//-------------------------------------------------------------------------//
		private int GetDeviceDrawmode(int mode)
		{
			if (mode == 0) return Gl.GL_POINTS;
			if (mode == 1) return Gl.GL_LINES;
			if (mode == 2) return Gl.GL_LINE_STRIP;
			if (mode == 3) return Gl.GL_LINE_LOOP;
			if (mode == 4) return Gl.GL_TRIANGLES;
			if (mode == 5) return Gl.GL_TRIANGLE_STRIP;
			if (mode == 6) return Gl.GL_TRIANGLE_FAN;
			if (mode == 7) return Gl.GL_QUADS;
			if (mode == 8) return Gl.GL_QUAD_STRIP;
			if (mode == 9) return Gl.GL_POLYGON;

			return -1;
		}

		//-------------------------------------------------------------------------//
		// Shading modes.
		public override int FLAT   { get { return Gl.GL_FLAT; } }
		public override int SMOOTH { get { return Gl.GL_SMOOTH; } }

		//-------------------------------------------------------------------------//
		// Shading modes.
		public override int POINT { get { return Gl.GL_POINT; } } 
		public override int LINE  { get { return Gl.GL_LINE; } }
		public override int FILL  { get { return Gl.GL_FILL; } }
	}
}
