using System;
using JPGL.Common;
using JPGL.Common.Primitives;
using JPGL.Common.Viewports;
using JPGL.Math;

using Tao.OpenGl;
using Tao.Platform.Windows;


namespace JPGL.OpenGL
{
	/// <summary>
	/// OpenGL specific implementation of a Viewport.
	/// </summary>
	public abstract class OpenGLViewport : Viewport
	{
		//-----------------------------------//
		private IntPtr m_DeviceContext;      // OpenGL DeviceContext
		private IntPtr m_RenderingContext;   // OpenGL RenderingContext
		//-----------------------------------//

		//-----------------------------------// OpenGL Primitive Renderer
		// Note that all opengl viewports share a common opengl primitive renderer...
		private static PrimitiveRenderer     m_OpenGLPrimitiveRenderer;
		protected override PrimitiveRenderer PrimRenderer
		{
			get
			{
				return m_OpenGLPrimitiveRenderer;
			}
		}                                    // 
		//-----------------------------------//

		
		//-------------------------------------------------------------------------//
		public OpenGLViewport() : base()
		{
			m_DeviceContext    = IntPtr.Zero;
			m_RenderingContext = IntPtr.Zero;

			m_OpenGLPrimitiveRenderer = new OpenGLPrimitiveRenderer();
		}

		//-------------------------------------------------------------------------//
		~OpenGLViewport()
		{
			if (m_DeviceContext != IntPtr.Zero)
			{
				// Delete the Rendering Context from the graphics hardware
				GraphicsSystem.Terminate(m_RenderingContext);
				m_RenderingContext = IntPtr.Zero;
				m_DeviceContext    = IntPtr.Zero;
			}
		}

		//-------------------------------------------------------------------------//
		protected sealed override bool Create()
		{
			// Get device context of the window we are connected to
			m_DeviceContext = User.GetDC(m_WindowContext);
			if (m_DeviceContext == IntPtr.Zero) return false;

			// Initialize the Pixel Format
			if (!GraphicsSystem.InitPixelFormat(m_DeviceContext)) return false;

			// Get an OpenGL Rendering Context
			m_RenderingContext = Wgl.wglCreateContext(m_DeviceContext);         
			if (m_RenderingContext == IntPtr.Zero) return false;

			// Activate context so we can initialize OpenGL details
			GraphicsSystem.SetActiveDevice(m_DeviceContext, m_RenderingContext);

			// Register with the graphics system
			GraphicsSystem.RegisterContext(m_DeviceContext, m_RenderingContext);

			Gl.glClearDepth(1.0f);										// Depth Buffer Setup
			Gl.glEnable(Gl.GL_DEPTH_TEST);								// Enables Depth Testing
			Gl.glDepthFunc(Gl.GL_LEQUAL);								// The Type Of Depth Testing To Do
			Gl.glHint(Gl.GL_PERSPECTIVE_CORRECTION_HINT, Gl.GL_NICEST);	// Really Nice Perspective Calculations
			Gl.glCullFace(Gl.GL_BACK);

			Gl.glLightModelf(Gl.GL_LIGHT_MODEL_LOCAL_VIEWER, 1.0f);
			float[] acolor = {0.06f, 0.06f, 0.06f, 0.0f};
			Gl.glLightModelfv(Gl.GL_LIGHT_MODEL_AMBIENT, acolor);
			Gl.glEnable(Gl.GL_NORMALIZE);

			// Deactivate our rendering context
			GraphicsSystem.SetActiveDevice(IntPtr.Zero, IntPtr.Zero);

			// Set up the matrices
			Resize();

			return true;
		}

		//-------------------------------------------------------------------------//
		protected sealed override void Resize()
		{
			// Activate our rendering context so we can resize
			GraphicsSystem.SetActiveDevice(m_DeviceContext, m_RenderingContext);

			// Tell OpenGL how to take its normalized device space (-1->1) and
			// draw it to the screen. This was not done in device initialization
			// because the wgl creation routine takes care of it. Since we do 
			// this, we do not need to recreate the device.
			Gl.glViewport(0, 0, m_Width, m_Height);

			// The PROJECTION matrix
			Gl.glMatrixMode(Gl.GL_PROJECTION);
			Gl.glLoadIdentity();
			Glu.gluPerspective(m_Camera.FOV, (float)m_Width / (float)m_Height, m_Camera.Near, m_Camera.Far);

			// The MODELVIEW matrix 
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glLoadIdentity();
			Glu.gluLookAt(
				m_Camera.EyePosition.X, m_Camera.EyePosition.Y, m_Camera.EyePosition.Z,
				m_Camera.LookAtPosition.X, m_Camera.LookAtPosition.Y, m_Camera.LookAtPosition.Z,
				m_Camera.EyeUpDirection.X, m_Camera.EyeUpDirection.Y, m_Camera.EyeUpDirection.Z);

			// Deactivate our rendering context
			GraphicsSystem.SetActiveDevice(IntPtr.Zero, IntPtr.Zero);
		}

		//-------------------------------------------------------------------------//
		protected void EnterOrthoMode()
		{
			Gl.glMatrixMode(Gl.GL_PROJECTION);
			Gl.glLoadIdentity();
			Glu.gluOrtho2D(0, m_Width, 0, m_Height);

			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glLoadIdentity();

			Gl.glDisable(Gl.GL_DEPTH_TEST);
		}

		//-------------------------------------------------------------------------//
		protected void ExitOrthoMode()
		{
			Gl.glEnable(Gl.GL_DEPTH_TEST);
			Resize();
		}

		//-------------------------------------------------------------------------//
		protected sealed override void Initialize()
		{
			// Grab the rendering pipe
			GraphicsSystem.SetActiveDevice(m_DeviceContext, m_RenderingContext);

			// Clear the buffers
			Gl.glClear(Gl.GL_COLOR_BUFFER_BIT | Gl.GL_DEPTH_BUFFER_BIT);

			// Set Background Color
			Gl.glClearColor(m_BackgroundColor.R/255.0f, m_BackgroundColor.G/255.0f, m_BackgroundColor.B/255.0f, m_BackgroundColor.A/255.0f);
		}

		//-------------------------------------------------------------------------//
		protected sealed override void Restore()
		{
			// We need to flush the rendering pipe and swap the back
			// buffer that we have been drawing to with the screen buffer.
			Gl.glFlush();
			Gdi.SwapBuffers(m_DeviceContext);
			
			// Release the Rendering Pipe
			GraphicsSystem.SetActiveDevice(IntPtr.Zero, IntPtr.Zero);
		}

		//-------------------------------------------------------------------------//
		// Our unprojection/projection is interested in the conversion between
		// object and world spaces. Since gluUnProject/gluProject convert between
		// object and device spaces, we will setup the projection and modelview
		// matrices to both be the identity matrix before we build the matrix
		// for our object. 
		protected sealed override void InitForUnProject()
		{
			GraphicsSystem.SetActiveDevice(m_DeviceContext, m_RenderingContext);

			Gl.glMatrixMode(Gl.GL_PROJECTION);
			Gl.glPushMatrix();
			Gl.glLoadIdentity();

			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glPushMatrix();
			Gl.glLoadIdentity();
		}

		//-------------------------------------------------------------------------//
		// Build the modelview matrix for our object. It is assumed that going into
		// this function, the matrix mode is GL_MODELVIEW, and the it is safe to 
		// modify whatever is on the top of the modelview matrix stack. Note that
		// this matrix is build without pushing at each step. This is done because
		// this matrix is simply thrown away after use, so the extra overhead of
		// pushing and poping is pointless.
		protected sealed override void BuildMatrix(SceneNode obj)
		{
			// Recursively build matrix starting from top of the tree.
			if (obj.Parent != null)
				BuildMatrix(obj.Parent);

			// Translate to the position offset with the pivot
			Gl.glTranslatef(obj.T.T.X+obj.T.P.X, obj.T.T.Y+obj.T.P.Y, obj.T.T.Z+obj.T.P.Z);

			// Rotate in x, y and z
			Gl.glMultMatrixf(obj.T.RQ.GetMatrix());

			// Scale
			Gl.glScalef(obj.T.S.X, obj.T.S.Y, obj.T.S.Z);

			// Translate away from pivot to control point
			Gl.glTranslatef(-1.0f*obj.T.P.X, -1.0f*obj.T.P.Y, -1.0f*obj.T.P.Z);

		}

		//-------------------------------------------------------------------------//
		protected sealed override void Project(Vector3 vo, out Vector3 vw)
		{
			// Get the modelview matrix
			double[] mvM = new double[16];
			Gl.glGetDoublev(Gl.GL_MODELVIEW_MATRIX, mvM);

			// Get the identity matrix for Projection
			double[] pjM = new double[16];
			Gl.glGetDoublev(Gl.GL_PROJECTION_MATRIX, pjM);

			// Get what is effectively the identity for the viewport transformation
			// which takes normalized device coordinates and maps them into hardware
			// space.
			int[] v = new int[4];
			v[0] = -1; v[1] = -1; v[2] = 2; v[3] = 2;

			// Project the point
			//double dwx, dwy, dwz;
			//Glu.gluProject(vo.X, vo.Y, vo.Z, mvM, pjM, v, out dwx, out dwy, out dwz);
			float[] mview = new float[16];
			Gl.glGetFloatv(Gl.GL_MODELVIEW_MATRIX, mview);

			vw = vo * new Matrix(
				mview[0], mview[1], mview[2], mview[3],
				mview[4], mview[5], mview[6], mview[7],
				mview[8], mview[9], mview[10], mview[11],
				mview[12], mview[13], mview[14], mview[15]);

			//vw = new Vector3((float)dwx, (float)dwy, (float)dwz);
		}

		//-------------------------------------------------------------------------//
		protected sealed override void UnProject(Vector3 vw, out Vector3 vo)
		{
			// Get the modelview matrix
			double[] mvM = new double[16];
			Gl.glGetDoublev(Gl.GL_MODELVIEW_MATRIX, mvM);

			// Get the identity matrix for Projection
			double[] pjM = new double[16];
			Gl.glGetDoublev(Gl.GL_PROJECTION_MATRIX, pjM);

			// Get what is effectively the identity for the viewport transformation
			// which takes normalized device coordinates and maps them into hardware
			// space.
			int[] v = new int[4];
			v[0] = -1; v[1] = -1; v[2] = 2; v[3] = 2;


			// Unproject the point
			double dox, doy, doz;
			Glu.gluUnProject(vw.X, vw.Y, vw.Z, mvM, pjM, v, out dox, out doy, out doz);

			vo = new Vector3((float)dox, (float)doy, (float)doz);
		}
	
		//-------------------------------------------------------------------------//
		// Pop the top of the modelview and projection matrices.
		protected sealed override void RestoreFromUnProject()
		{
			Gl.glMatrixMode(Gl.GL_MODELVIEW);
			Gl.glPopMatrix();

			Gl.glMatrixMode(Gl.GL_PROJECTION);
			Gl.glPopMatrix();

			GraphicsSystem.SetActiveDevice(IntPtr.Zero, IntPtr.Zero);
		}

	}
}
