using System;
using System.ComponentModel;
using System.Windows.Forms;
using JPGL.Common.Viewports;


namespace JPGL.SWF
{
	/// <summary>
	/// Summary description for UIWindow.
	/// </summary>
	public class UIWindow : System.Windows.Forms.UserControl
	{
		//-----------------------------------//
		private Viewport m_Viewport;        // The viewport attached to us

		//-----------------------------------// Constants for mouse buttons
		public const byte MB_LEFT   = 1;     //
		public const byte MB_RIGHT  = 2;     //
		public const byte MB_MIDDLE = 4;     //
		//-----------------------------------// 

		//-----------------------------------// 
		private float m_WidthPercent;        // Percent of the window the 
		public float WidthPercent            // window width for view
		{
			set 
			{ 
				m_WidthPercent = value; 
				UIWindow_Resize(null, null);
			}
		}
		//-----------------------------------// 
		
		//-----------------------------------// 
		private bool m_ForceSquare;          // Force square views?
		public bool ForceSquare              //
		{
			set 
			{
				m_ForceSquare = value;
				UIWindow_Resize(null, null);
			}
		}
		//-----------------------------------// 


		private System.Windows.Forms.Panel Panel_Border;
		private System.Windows.Forms.Panel Panel_GW;
		private System.ComponentModel.IContainer components;

		//-------------------------------------------------------------------------//
		public UIWindow()
		{
			// This call is required by the Windows.Forms Form Designer.
			InitializeComponent();

			// Register the resize handler
			this.Resize += new EventHandler(UIWindow_Resize);

			// Register the mouse event handlers
			this.Panel_GW.MouseMove += new System.Windows.Forms.MouseEventHandler(this.OnMouseMove);
			this.Panel_GW.MouseUp   += new System.Windows.Forms.MouseEventHandler(this.OnMouseUp);
			this.Panel_GW.MouseDown += new System.Windows.Forms.MouseEventHandler(this.OnMouseDown);

			m_ForceSquare  = false;
			m_WidthPercent = 100.0f;
		}

		//-------------------------------------------------------------------------//
		// Tell the viewport to draw itself.
		public void DrawViewport()
		{
			if (m_Viewport == null) return;

			m_Viewport.InitializeDeviceForDrawing();
			m_Viewport.DrawToDevice();
			m_Viewport.RestoreDeviceAfterDrawing();
		}

		//-------------------------------------------------------------------------//
		// Initialize the given viewport and attach it to this UI Window.
		public bool AttachViewport(Viewport viewport)
		{
			if (!viewport.CreateGraphicsDevice(this.Panel_GW.Handle, this.Panel_GW.Width, this.Panel_GW.Height)) return false;

			m_Viewport = viewport;

			return true;
		}

		//-------------------------------------------------------------------------//
		// Alert the viewport with the new size.
		public void UIWindow_Resize(object sender, System.EventArgs e)
		{
			if (m_Viewport == null) return;

			this.Panel_GW.Size = new System.Drawing.Size(Panel_Border.Width-2, Panel_Border.Height-2);

			// Size as percent of window width, if necessary
			float ratio = m_WidthPercent / 100.0f;
			if (ratio < 1.0f)
			{
				Panel_GW.Width = (int)(Panel_GW.Width * ratio);
			}

			// Should the viewport be forced to a square?
			if (m_ForceSquare)
			{
				if (this.Panel_GW.Width > this.Panel_GW.Height)
					this.Panel_GW.Width = this.Panel_GW.Height;
				else if (this.Panel_GW.Width < this.Panel_GW.Height)
					this.Panel_GW.Height = this.Panel_GW.Width;
			}

			m_Viewport.ResizeGraphicsDevice(this.Panel_GW.Width, this.Panel_GW.Height);
		}

		//-------------------------------------------------------------------------//
		// Mouse Event Handlers - Translate MouseEventArgs to a form the independant
		// of System.Windows.Forms, and call the viewport's handlers.
		private void OnMouseMove(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (m_Viewport != null) 
				m_Viewport.OnMouseMove(e.X, e.Y, MouseButtonsToFlags(e.Button));
		}
		private void OnMouseDown(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (m_Viewport != null)
				m_Viewport.OnMouseDown(e.X, e.Y, MouseButtonsToFlags(e.Button));
		}
		private void OnMouseUp(object sender, System.Windows.Forms.MouseEventArgs e)
		{
			if (m_Viewport != null)
				m_Viewport.OnMouseUp(e.X, e.Y, MouseButtonsToFlags(e.Button));
		}

		//-------------------------------------------------------------------------//
		// Given a System.Windows.Forms MouseButtons object, translate it to our
		// nFlags mouse button form - returns mouse button generating the mouse event.
		private uint MouseButtonsToFlags(MouseButtons m)
		{
			uint nFlags = 0;

			if (m == MouseButtons.Left)   nFlags = nFlags | MB_LEFT;
			if (m == MouseButtons.Right)  nFlags = nFlags | MB_RIGHT;
			if (m == MouseButtons.Middle) nFlags = nFlags | MB_MIDDLE;

			return nFlags;
		}

		/// <summary> 
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if(components != null)
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Component Designer generated code
		/// <summary> 
		/// Required method for Designer support - do not modify 
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.Panel_Border = new System.Windows.Forms.Panel();
			this.Panel_GW = new System.Windows.Forms.Panel();
			this.Panel_Border.SuspendLayout();
			this.SuspendLayout();
			// 
			// Panel_Border
			// 
			this.Panel_Border.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.Panel_Border.BorderStyle = System.Windows.Forms.BorderStyle.FixedSingle;
			this.Panel_Border.Controls.Add(this.Panel_GW);
			this.Panel_Border.ForeColor = System.Drawing.Color.Black;
			this.Panel_Border.Location = new System.Drawing.Point(0, 0);
			this.Panel_Border.Name = "Panel_Border";
			this.Panel_Border.Size = new System.Drawing.Size(431, 408);
			this.Panel_Border.TabIndex = 0;
			// 
			// Panel_GW
			// 
			this.Panel_GW.Anchor = ((System.Windows.Forms.AnchorStyles)((((System.Windows.Forms.AnchorStyles.Top | System.Windows.Forms.AnchorStyles.Bottom) 
				| System.Windows.Forms.AnchorStyles.Left) 
				| System.Windows.Forms.AnchorStyles.Right)));
			this.Panel_GW.Location = new System.Drawing.Point(0, 0);
			this.Panel_GW.Name = "Panel_GW";
			this.Panel_GW.Size = new System.Drawing.Size(429, 406);
			this.Panel_GW.TabIndex = 0;
			// 
			// UIWindow
			// 
			this.Controls.Add(this.Panel_Border);
			this.Name = "UIWindow";
			this.Size = new System.Drawing.Size(432, 408);
			this.Panel_Border.ResumeLayout(false);
			this.ResumeLayout(false);

		}
		#endregion
	}
}
